// input_noc.cpp: derived from particle_simulation.cpp by K. Moazzemi et al.

#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "systemc.h"

#include "config.h"

//////////////////////////////////////////////////////////////////////////
// Data structures
//////////////////////////////////////////////////////////////////////////

typedef struct Particle_t
{
  double 	x, y;	// position
  double 	vx, vy;	// velocity
  double	ax, ay;	// acceleration
  unsigned	color;	// color in hex RGB byte format
  int		id;	// ID (if INVALID, no particle)

  inline friend std::ostream& operator<<(
	std::ostream& os, const Particle_t& co)
  { os << co.x << ' ' << co.y << '\n';
    return os;
  }
} particle_t;

unsigned const BLACK = 0x000000; // colorful code ;-)
unsigned const RED   = 0xff0000;
unsigned const GREEN = 0x00ff00;
unsigned const BLUE  = 0x0000ff;
unsigned const WHITE = 0xffffff;
unsigned const PURPLE= 0xff00ff;

const int INVALID = -1;	// invalid particle ID (no particle)
const particle_t pINVALID = {0,0,0,0,0,0,WHITE,INVALID};

enum {NORTH, EAST, SOUTH, WEST}; // transfer directions

//////////////////////////////////////////////////////////////////////////
// Channels
//////////////////////////////////////////////////////////////////////////

//
// user-defined typed double-handshake channel (with transceiver interface)
//
class i_particle_transceiver: virtual public sc_interface
{
    public:	// (mimic a subset of sc_fifo interface)
	virtual std::vector<particle_t> read() = 0;
	virtual void write(const std::vector<particle_t>) = 0;
};

class c_particle_double_handshake_channel:
	public sc_channel, public i_particle_transceiver
{
    public:
	c_particle_double_handshake_channel(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_particle_double_handshake_channel()
	{ }

	#ifdef RISC
	#pragma RISC read unconditional-wait event
	#endif
	std::vector<particle_t> read()
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    std::vector<particle_t> temp = tmpd;

	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	    return temp;
	}

	#ifdef RISC
	#pragma RISC write unconditional-wait event
	#endif
	void write(const std::vector<particle_t> X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}

    private:
	sc_event req;
	sc_event ack;
	bool v, w;
  std::vector<particle_t> tmpd;
};

//////////////////////////////////////////////////////////////////////////
// Testbench Module
//////////////////////////////////////////////////////////////////////////

SC_MODULE(Testbench)
{

#if NUM_DUT == 1
    sc_port<i_particle_transceiver> CoordsNorth;
    sc_port<i_particle_transceiver> CoordsEast;
    sc_port<i_particle_transceiver> CoordsSouth;
    sc_port<i_particle_transceiver> CoordsWest;
    sc_port<i_particle_transceiver> Coords_1;
#endif
#if NUM_DUT == 4
    sc_port<i_particle_transceiver> Coords_1;
    sc_port<i_particle_transceiver> Coords_2;
    sc_port<i_particle_transceiver> Coords_3;
    sc_port<i_particle_transceiver> Coords_4;
#endif
#if NUM_DUT == 9
    sc_port<i_particle_transceiver> Coords_1;
    sc_port<i_particle_transceiver> Coords_2;
    sc_port<i_particle_transceiver> Coords_3;
    sc_port<i_particle_transceiver> Coords_4;
    sc_port<i_particle_transceiver> Coords_5;
    sc_port<i_particle_transceiver> Coords_6;
    sc_port<i_particle_transceiver> Coords_7;
    sc_port<i_particle_transceiver> Coords_8;
    sc_port<i_particle_transceiver> Coords_9;
#endif

    const double Size;	// size of the entire particle space
    particle_t particles_pool[NUM_PARTICLES];
    int LostParticles;
    FILE *LogFile;

    Testbench(sc_module_name name_): sc_module(name_),
	Size(sqrt(density * NUM_PARTICLES))	// keep density constant
    {
	LostParticles = 0;
	LogFile = 0;
	SC_HAS_PROCESS(Testbench);
	//SC_THREAD(main);
    }

    void init_particles(void);
    void save(void);

    void main(void)
    {
	int		i, c, r;
	std::vector<particle_t> p;

	init_particles();

	LogFile = fopen("newcoords.txt", "w");
	if (!LogFile)
	{
	    #if VERBOSITY >= 1
	    printf("Cannot open log file for writing\n");
	    #endif
	    sc_stop();
	}

	#if VERBOSITY >= 3
	printf("    TESTBENCH saved particles in step %d\n", 0);
	#endif

	std::vector<particle_t> to_coords_1;
	std::vector<particle_t> to_coords_2;
	std::vector<particle_t> to_coords_3;
	std::vector<particle_t> to_coords_4;
	std::vector<particle_t> to_coords_5;
	std::vector<particle_t> to_coords_6;
	std::vector<particle_t> to_coords_7;
	std::vector<particle_t> to_coords_8;
	std::vector<particle_t> to_coords_9;

	// Stimulus job
	for(i=0; i<NUM_PARTICLES; i++)
	{
	    particle_t p = particles_pool[i];
	    c = (int)(p.x / (Size/NUM_COLS));
	    r = (int)(p.y / (Size/NUM_ROWS));
	    r = NUM_ROWS - r - 1;	// tiles are numbered top-down
	    switch(r*NUM_ROWS+c)
	    {
	    #if NUM_DUT == 1
		case 0: to_coords_1.push_back(p); break; // send to tile 1
	    #endif
	    #if NUM_DUT == 4
		case 0: to_coords_1.push_back(p); break; // send to tile 1
		case 1: to_coords_2.push_back(p); break; // send to tile 1
		case 2: to_coords_3.push_back(p); break; // send to tile 1
		case 3: to_coords_4.push_back(p); break; // send to tile 1
	    #endif
	    #if NUM_DUT == 9
		case 0: to_coords_1.push_back(p); break; // send to tile 1
		case 1: to_coords_2.push_back(p); break; // send to tile 1
		case 2: to_coords_3.push_back(p); break; // send to tile 1
		case 3: to_coords_4.push_back(p); break; // send to tile 1
		case 4: to_coords_5.push_back(p); break; // send to tile 1
		case 5: to_coords_6.push_back(p); break; // send to tile 1
		case 6: to_coords_7.push_back(p); break; // send to tile 1
		case 7: to_coords_8.push_back(p); break; // send to tile 1
		case 8: to_coords_9.push_back(p); break; // send to tile 1
	    #endif
		default: assert(false);
	    }
	}

	to_coords_1.push_back(pINVALID);
	to_coords_2.push_back(pINVALID);
	to_coords_3.push_back(pINVALID);
	to_coords_4.push_back(pINVALID);
	to_coords_5.push_back(pINVALID);
	to_coords_6.push_back(pINVALID);
	to_coords_7.push_back(pINVALID);
	to_coords_8.push_back(pINVALID);
	to_coords_9.push_back(pINVALID);

#if NUM_DUT == 1
	Coords_1->write(to_coords_1);	// "done" message
#endif
#if NUM_DUT == 4
	Coords_1->write(to_coords_1);	// "done" message
	Coords_2->write(to_coords_2);	// "done" message
	Coords_3->write(to_coords_3);	// "done" message
	Coords_4->write(to_coords_4);	// "done" message
#endif
#if NUM_DUT == 9
	Coords_1->write(to_coords_1);	// "done" message
	Coords_2->write(to_coords_2);	// "done" message
	Coords_3->write(to_coords_3);	// "done" message
	Coords_4->write(to_coords_4);	// "done" message
	Coords_5->write(to_coords_5);	// "done" message
	Coords_6->write(to_coords_6);	// "done" message
	Coords_7->write(to_coords_7);	// "done" message
	Coords_8->write(to_coords_8);	// "done" message
	Coords_9->write(to_coords_9);	// "done" message
#endif
	#if VERBOSITY >= 3
	printf("    TESTBENCH sent all particles\n");
	#endif

	// Monitor Job
	for(int step = 0; step < NSTEPS; step++)
	{
	    i = 0;
	    p = Coords_1->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    #if NUM_DUT > 1
	    p = Coords_2->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    p = Coords_3->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    p = Coords_4->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    #endif
	    #if NUM_DUT > 4
	    p = Coords_5->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    p = Coords_6->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    p = Coords_7->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    p = Coords_8->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    p = Coords_9->read();
	    for(std::vector<particle_t>::iterator iter = p.begin(); iter != p.end(); iter++) {
		particle_t p = *iter;
	        if (p.id == INVALID) break;
		    particles_pool[i++] = p;
	    }
	    #endif
	    #if VERBOSITY >= 3
	    printf("    TESTBENCH received %d particles in step %d\n", i, step);
	    #endif
	    assert(i <= NUM_PARTICLES);
	    for (/* i */; i<NUM_PARTICLES; i++)
	    {	// in case any particles got lost, mark those invalid
		particles_pool[i] = pINVALID;
	    }

	    //save();		// save another frame
	    #if VERBOSITY >= 3
	    printf("    TESTBENCH saved particles in step %d\n", step);
	    #endif
	}
#if VERBOSITY >= 1
	if (LostParticles)
	    printf("%d particles lost! (total %d, tile capacity %d)\n",
			LostParticles, NUM_PARTICLES, MAX_PARTICLES);
#endif
#if VERBOSITY >= 2
	printf("  TESTBENCH done, stopping simulation.\n");
#endif
	fclose(LogFile);
	LogFile = 0;
	sc_stop();
    }
};

void Testbench::init_particles(void)
{
    particle_t *p = particles_pool;
    int n = NUM_PARTICLES;

    srand48(42);	// repeatable random numbers
    int sx = (int)ceil(sqrt((double)n));
    int sy = (n+sx-1)/sx;

    std::vector<int> shuffle(n);
    for(int i=0; i<n; i++)
	shuffle[i] = i;

    for(int i=0; i<n; i++)
    {
	//
	// make sure particles are not spatially sorted
	//
	int j = lrand48()%(n-i);
	int k = shuffle[j];
	shuffle[j] = shuffle[n-i-1];
	//
	// distribute particles evenly to ensure proper spacing
	//
	p[i].x = Size*(1.+(k%sx))/(1+sx);
	p[i].y = Size*(1.+(k/sx))/(1+sy);
	//
	// assign random velocities within a bound
	//
	p[i].vx = drand48()*2-1;
	p[i].vy = drand48()*2-1;
	#ifdef FALLING_INIT_ORIENTATION
	p[i].vx +=  2;	// (adjusted by RD)
	p[i].vy += -4;
	#endif
	#ifdef CENTER_INIT_ORIENTATION
	p[i].vx += 8 * ((Size/2)-p[i].x)/Size;	// (fixed by RD)
	p[i].vy += 8 * ((Size/2)-p[i].y)/Size;
	#endif
	p[i].ax = 0;
	p[i].ay = 0;
	p[i].color = BLACK;
	p[i].id = i;
    }
}

void Testbench::save(void)
{
    particle_t *p = particles_pool;
    int n = NUM_PARTICLES;
    static bool first = true;
    int	count = 0;
    assert(LogFile);

    if (first)	// first line: number particles, size of space
    {
	fprintf(LogFile, "%d %g\n", n, Size);
	first = false;
    }
    for(int i=0; i<n; i++)
    {		// following lines: position, color, and id
	if (p[i].id != INVALID)
	{
	    fprintf(LogFile, "%g %g %x %d\n",
			p[i].x, p[i].y, p[i].color, p[i].id);
	    count++;
	}
    }
    assert(NUM_PARTICLES <= MAX_PARTICLES * NUM_DUT);
    assert(count <= MAX_PARTICLES * NUM_DUT);
    assert(count <= NUM_PARTICLES);
    assert(MAX_PARTICLES < NUM_PARTICLES	// if one tile can handle all,
		|| count == NUM_PARTICLES);	// then none must get lost!

    // if particles got lost due to the limited tile storage,
    // fill this frame with as many missing particles
    // (so that lost particles are nicely visualized)
    LostParticles = n - count;
    for (int i=0; i<LostParticles; i++)
    {
	fprintf(LogFile, "%g %g %x 0\n",
		i*(Size/NUM_PARTICLES),		// row of lost particles
		-.01 * (Size/NUM_PARTICLES),	// 1% below the x-axis
		PURPLE);			// in purple color
    }
}

//////////////////////////////////////////////////////////////////////////
// design under test units
//////////////////////////////////////////////////////////////////////////

SC_MODULE(DUT)
{
    sc_port<i_particle_transceiver> CoordsControl;
    sc_port<i_particle_transceiver> CoordsNorth;
    sc_port<i_particle_transceiver> CoordsEast;
    sc_port<i_particle_transceiver> CoordsSouth;
    sc_port<i_particle_transceiver> CoordsWest;

    void main(void);
    void apply_force(int p, int n);
    void move(int pIndex);
    void advanced_move(int pIndex);
    void transfer_escaping_particles(void);
    int  add_immigrants(void);
    void recycle_memory(void);

    DUT(sc_module_name name_, int Row_ = 1, int Col_ = 1):
	sc_module(name_),
	Row(Row_), Col(Col_),
	TileSize(sqrt(density * NUM_PARTICLES) / sqrt(NUM_DUT)),
	LeftBound((Col-1)*TileSize),
	RightBound(LeftBound + TileSize),
	BottomBound((NUM_ROWS-Row)*TileSize),	// top-down order
	TopBound(BottomBound + TileSize)
    {
	SC_HAS_PROCESS(DUT);
	SC_THREAD(main);
    }

    private:
	const int Row, Col;	// tile number (TILE_NO(Row,Col))
	const double TileSize;	// size of this unit's particle space
	const double LeftBound;	// space controlled by this tile
	const double RightBound;
	const double BottomBound;
	const double TopBound;
	particle_t particles[MAX_PARTICLES];
	int nParticlesInUse;	// current memory usage of MAX_PARTICLES
				// (some slots may be empty)
	std::vector<std::vector<particle_t> > fugitives;
	std::vector<particle_t> immigrants;
	std::vector<int> empty_cells;
};

void DUT::main(void)
{
	fugitives.push_back(std::vector<particle_t>());
	assert(0 == NORTH);
	fugitives.push_back(std::vector<particle_t>());
	assert(1 == EAST);
	fugitives.push_back(std::vector<particle_t>());
	assert(2 == SOUTH);
	fugitives.push_back(std::vector<particle_t>());
	assert(3 == WEST);

	// DUT job:
	// receive initial particles from stimulus,
	// then run particle simulation (repeated steps)
	// and report status regularly to the monitor
	#ifdef DEBUG
	printf("  DUT%d[r%d,c%d] starts.\n", TILE_NO(Row,Col), Row, Col);
	#endif
	nParticlesInUse = 0;
	std::vector<particle_t> pp;
	pp = CoordsControl->read();	// read particles from testbench

	#if VERBOSITY >= 2
	printf("  DUT%d[r%d,c%d] received %d particles.\n",
		TILE_NO(Row,Col), Row, Col, nParticlesInUse);
	#endif

	for(int step = 0; step < NSTEPS; step++)
	{
	    #if VERBOSITY >= 3
	    printf("    DUT%d[r%d,c%d] starting step %d\n",
		TILE_NO(Row,Col), Row, Col, step);
	    #endif

	#ifdef VECTORIZE
		#pragma simd assert
	#endif
	for(int i=0; i<nParticlesInUse; i++)
	{
		if (particles[i].id == INVALID)
			continue;

		particles[i].ax = GRAVITY_X;
		particles[i].ay = GRAVITY_Y;

		double pix = particles[i].x;
		double piy = particles[i].y;
		double piax = particles[i].ax;
		double piay = particles[i].ay;

		for (int j=0; j<nParticlesInUse; j++)
		{
			if (particles[j].id != INVALID)
			{
				double dx = particles[j].x - pix;
				double dy = particles[j].y - piy;

				double r2 = dx * dx + dy * dy;
				if( r2 > cutoff*cutoff )
				{
					//return;
				} else {

					r2 = std::max( r2, min_r*min_r );
					double r = sqrt( r2 );
					//
					// very simple short-range repulsive force
					//
					double coef = ( 1 - cutoff / r ) / r2 / mass;
					particles[i].ax += coef * dx;
					particles[i].ay += coef * dy;

					if (piax*piax + piay*piay > 10000000)
					{
						particles[i].ax = 0;
						particles[i].ay = 0;
					}
				}
			}
		}
	}

	    if (NUM_DUT > 1)
	    {
		transfer_escaping_particles();
	    }
	    if(immigrants.size() > 0)
	    {
		int lost = add_immigrants();
		if (lost)
		{
		    #if VERBOSITY >= 1
		    printf("DUT%d[r%d,c%d] out of memory in step %d!"
			" Losing %d of %u incoming particles.\n",
			TILE_NO(Row,Col), Row, Col, step, lost,
			(unsigned)(immigrants.size()));
		    #endif
		}
		immigrants.erase (immigrants.begin(),immigrants.end());
	    }

	    //
	    // communicate with testbench to save the particles
	    //
	    if ((SAVE_FILE == 1) && ((step%SAVEFREQ) == 0))
	    {
		std::vector<particle_t> pp;
		for(int i = 0; i<nParticlesInUse; i++)
		{
		    if (particles[i].id != INVALID)
			pp.push_back(particles[i]);
		}
		pp.push_back(pINVALID);
		CoordsControl->write(pp);

		#if VERBOSITY >= 3
		printf("    DUT%d[r%d,c%d] sent new coordinates in step %d.\n",
			TILE_NO(Row,Col), Row, Col, step);
		#endif
	    }
	}

	int count = 0;
	for(int i = 0; i< MAX_PARTICLES; i++){
	    if (particles[i].id != INVALID)
		count++;
	    assert(i < nParticlesInUse || particles[i].id == INVALID);
	}
	#if VERBOSITY >= 2
	printf("  DUT%d[r%d,c%d] ends with %d particles.\n",
		TILE_NO(Row,Col), Row, Col, count);
	#endif
	assert(nParticlesInUse >= count);
	assert(nParticlesInUse <= MAX_PARTICLES);
}

//
// interact two particles
//
#ifdef VECTORIZE
#pragma omp declare simd
#endif
void DUT::apply_force(int p, int n)
{
    double dx = particles[n].x - particles[p].x;
    double dy = particles[n].y - particles[p].y;
    double r2 = dx * dx + dy * dy;
    if( r2 > cutoff*cutoff )
	return;

	r2 = fmax( r2, min_r*min_r );
	double r = sqrt( r2 );
	//
	// very simple short-range repulsive force
	//
	double coef = ( 1 - cutoff / r ) / r2 / mass;
	particles[p].ax += coef * dx;
	particles[p].ay += coef * dy;

	if (particles[p].ax*particles[p].ax + particles[p].ay*particles[p].ay
			> 10000000)
	{
		particles[p].ax = 0;
		particles[p].ay = 0;
	}
}

//
// move a particle (basic version)
//
void DUT::move(int pIndex)
{
    particle_t &p = particles[pIndex];
    assert(p.id != INVALID);
    //
    // slightly simplified Velocity Verlet integration
    // conserves energy better than explicit Euler method
    //
    p.vx += p.ax * dt;
    p.vy += p.ay * dt;
    p.x  += p.vx * dt;
    p.y  += p.vy * dt;
    //
    // bounce from walls
    //
    while( p.x < 0 || p.x > TileSize )	// RD: while or if!?
    {
	p.x  = p.x < 0 ? -p.x : 2*TileSize-p.x;
	p.vx = -p.vx;
	p.color = BLUE;
    }
    while( p.y < 0 || p.y > TileSize )	// RD: while or if!?
    {
	p.y  = p.y < 0 ? -p.y : 2*TileSize-p.y;
	p.vy = -p.vy;
	p.color = BLUE;
    }
}

//
// move a particle (advanced version for multiple cooperating tiles)
//
void DUT::advanced_move(int pIndex)
{
    assert(particles[pIndex].id != INVALID);
    //
    //  slightly simplified Velocity Verlet integration
    //  conserves energy better than explicit Euler method
    //
    particles[pIndex].vx += particles[pIndex].ax * dt;
    particles[pIndex].vy += particles[pIndex].ay * dt;
    particles[pIndex].x  += particles[pIndex].vx * dt;
    particles[pIndex].y  += particles[pIndex].vy * dt;
    //
    //  bounce from walls or switch to neighbor tile
    //
    if (particles[pIndex].x < LeftBound)
    {
	if (Col != 1)		// switch to West neighbor tile
	{
	    particles[pIndex].color = RED;
	    fugitives[WEST].push_back(particles[pIndex]);
	    particles[pIndex] = pINVALID;
	    return;
	}
	else			// bounce on left wall
	{
	    particles[pIndex].x  = -particles[pIndex].x;
	    particles[pIndex].vx = -particles[pIndex].vx;
	    particles[pIndex].color = BLUE;
	}
    }
    if (particles[pIndex].x > RightBound)
    {
	if (Col != NUM_COLS)	// switch to East neighbor tile
	{
	    particles[pIndex].color = RED;
	    fugitives[EAST].push_back(particles[pIndex]);
	    particles[pIndex] = pINVALID;
	    return;
	}
	else			// bounce on right wall
	{
	    particles[pIndex].x = 2*RightBound-particles[pIndex].x;
	    particles[pIndex].vx= -particles[pIndex].vx;
	    particles[pIndex].color = BLUE;
	}
    }
    if (particles[pIndex].y < BottomBound)
    {
	if (Row != NUM_ROWS)	// switch to South neighbor tile
	{
	    particles[pIndex].color = RED;
	    fugitives[SOUTH].push_back(particles[pIndex]);
	    particles[pIndex] = pINVALID;
	    return;
	}
	else			// bounce on bottom wall
	{
	    particles[pIndex].y  = -particles[pIndex].y;
	    particles[pIndex].vy = -particles[pIndex].vy;
	    particles[pIndex].color = BLUE;
	}
    }
    if (particles[pIndex].y > TopBound)
    {
	if (Row != 1)		// switch to North neighbor tile
	{
	    particles[pIndex].color = RED;
	    fugitives[NORTH].push_back(particles[pIndex]);
	    particles[pIndex] = pINVALID;
	    return;
	}
	else			// bounce on top wall
	{
	    particles[pIndex].y = 2*TopBound-particles[pIndex].y;
	    particles[pIndex].vy = -particles[pIndex].vy;
	    particles[pIndex].color = BLUE;
	}
    }
}

void DUT::recycle_memory(void)
{
    for(int i = 0; i < nParticlesInUse; i++)
    {
	if (particles[i].id == INVALID)
	{
	    empty_cells.push_back(i);
	}
    }
}

int DUT::add_immigrants(void)
{
    recycle_memory();	// inserts incoming immigrants into place
			// of escaping particles to reuse the memory
    int incoming = immigrants.size();
    int lost = nParticlesInUse + incoming - MAX_PARTICLES;
    lost = lost>0 ? lost : 0;
    for(size_t i = 0; i < immigrants.size()-lost; i++)
    {
	if (empty_cells.size()>0)
	{
	    particles[empty_cells.back()] = immigrants[i];
	    empty_cells.pop_back();
	}
	else
	{
	    particles[nParticlesInUse] = immigrants[i];
	    nParticlesInUse++;
	    assert(nParticlesInUse <= MAX_PARTICLES);
	}
    }
    empty_cells.erase(empty_cells.begin(), empty_cells.end());
    return lost;
}

void DUT::transfer_escaping_particles(void)
{
	particle_t num_north, num_east, num_south, num_west;
	//particle_t transfer_size, temp_coord;
	int i;
	num_north.x = int(fugitives[NORTH].size());
	num_east.x = int(fugitives[EAST].size());
	num_south.x = int(fugitives[SOUTH].size());
	num_west.x = int(fugitives[WEST].size());
	int imN = 0, imE = 0, imS = 0, imW = 0;
#ifdef DEBUG
	printf("DUT%d[r%d,c%d]: fugitives\tN=%d\tE=%d\tS=%d\tW=%d\n",
		TILE_NO(Row,Col), Row, Col,
		int(fugitives[NORTH].size()), int(fugitives[EAST].size()),
		int(fugitives[SOUTH].size()), int(fugitives[WEST].size()));
#endif

	if (Row % 2 == 0)	//even row
	{
	    std::vector<particle_t> temp_coords;
	    temp_coords = CoordsSouth->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {
		immigrants.push_back(*iter);
		imS++;
	    }
	    temp_coords.clear();
	    for(i = 0; i < int(num_south.x); i++)
	    {
		particle_t temp_particle = fugitives[SOUTH][i];
		temp_coords.push_back(temp_particle);
	    }
	    CoordsSouth->write(temp_coords);
	    fugitives[SOUTH].erase (fugitives[SOUTH].begin(),fugitives[SOUTH].end());
	    temp_coords.clear();
	    for(i = 0; i < int(num_north.x); i++)
	    {
		particle_t temp_particle = fugitives[NORTH][i];
		temp_coords.push_back(temp_particle);
	    }
	    CoordsNorth->write(temp_coords);
	    fugitives[NORTH].erase (fugitives[NORTH].begin(),fugitives[NORTH].end());
	    temp_coords.clear();
	    temp_coords = CoordsNorth->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {
		immigrants.push_back(*iter);
		imN++;
	    }
	}
	else
	{
	    std::vector<particle_t> temp_coords;
	    for(i = 0; i < int(num_north.x); i++)
	    {
		particle_t temp = fugitives[NORTH][i];
		temp_coords.push_back(temp);
	    }
	    CoordsNorth->write(temp_coords);
	    fugitives[NORTH].erase (fugitives[NORTH].begin(),fugitives[NORTH].end());
	    temp_coords.clear();
	    temp_coords = CoordsNorth->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {

		immigrants.push_back(*iter);
		imN++;
	    }
	    temp_coords.clear();
	    temp_coords = CoordsSouth->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {
		immigrants.push_back(*iter);
		imS++;
	    }
	    temp_coords.clear();
	    for(i = 0; i < int(num_south.x); i++) {
		particle_t temp = fugitives[SOUTH][i];
		temp_coords.push_back(temp);
	    }
	    CoordsSouth->write(temp_coords);
	    fugitives[SOUTH].erase (fugitives[SOUTH].begin(),fugitives[SOUTH].end());
	}

	if (Col % 2 == 0)	// even column
	{
	    std::vector<particle_t> temp_coords;
	    temp_coords = CoordsWest->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {

		immigrants.push_back(*iter);
		imW++;
	    }
	    temp_coords.clear();
	    for(i = 0; i < int(num_west.x); i++) {
		particle_t temp = fugitives[WEST][i];
		temp_coords.push_back(temp);
	    }
	    CoordsWest->write(temp_coords);
	    fugitives[WEST].erase (fugitives[WEST].begin(),fugitives[WEST].end());
	    temp_coords.clear();
	    for(i = 0; i < int(num_east.x); i++) {

		particle_t temp = fugitives[EAST][i];
		temp_coords.push_back(temp);
	    }
	    CoordsEast->write(temp_coords);
	    fugitives[EAST].erase (fugitives[EAST].begin(),fugitives[EAST].end());
	    temp_coords.clear();
	    temp_coords = CoordsEast->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {

		immigrants.push_back(*iter);
		imE++;
	    }
	}
	else
	{
	    std::vector<particle_t> temp_coords;
	    for(i = 0; i < int(num_east.x); i++) {
		particle_t temp = fugitives[EAST][i];
		temp_coords.push_back(temp);
	    }
	    CoordsEast->write(temp_coords);
	    fugitives[EAST].erase (fugitives[EAST].begin(),fugitives[EAST].end());
	    temp_coords.clear();
	    temp_coords = CoordsEast->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {
		immigrants.push_back(*iter);
		imE++;
	    }
	    temp_coords.clear();
	    temp_coords = CoordsWest->read();
	    for(std::vector<particle_t>::iterator iter = temp_coords.begin(); iter != temp_coords.end(); iter++) {
		immigrants.push_back(*iter);
		imW++;
	    }
	    temp_coords.clear();
	    for(i = 0; i < int(num_west.x); i++)
	    {
		particle_t temp = fugitives[WEST][i];
		temp_coords.push_back(temp);

	    }
	    CoordsWest->write(temp_coords);
	    fugitives[WEST].erase (fugitives[WEST].begin(),fugitives[WEST].end());
	}

#ifdef DEBUG
	printf("DUT%d[r%d,c%d]: immigrants\tN=%d\tE=%d\tS=%d\tW=%d\n",
		TILE_NO(Row,Col), Row, Col, imN, imE, imS, imW);
#endif
}

//////////////////////////////////////////////////////////////////////////
// Top Module
//////////////////////////////////////////////////////////////////////////

SC_MODULE(TOP)
{
#if NUM_DUT == 1
    c_particle_double_handshake_channel   c_north;
    c_particle_double_handshake_channel   c_east;
    c_particle_double_handshake_channel   c_south;
    c_particle_double_handshake_channel   c_west;
    c_particle_double_handshake_channel   c_1;
#endif
#if NUM_DUT > 1 //we are assuming it goes to next square number
    //controling links from testbench
    c_particle_double_handshake_channel   c_1;
    c_particle_double_handshake_channel   c_2;
    c_particle_double_handshake_channel   c_3;
    c_particle_double_handshake_channel   c_4;
    //straight links
    c_particle_double_handshake_channel   c23;
    c_particle_double_handshake_channel   c32;
    c_particle_double_handshake_channel   c43;
    c_particle_double_handshake_channel   c34;
    //torus links
    c_particle_double_handshake_channel   c_h1;
    c_particle_double_handshake_channel   c_h2;
    c_particle_double_handshake_channel   c_v1;
    c_particle_double_handshake_channel   c_v2;
#endif
#if NUM_DUT > 4
    //controling links from testbench
    c_particle_double_handshake_channel   c_5;
    c_particle_double_handshake_channel   c_6;
    c_particle_double_handshake_channel   c_7;
    c_particle_double_handshake_channel   c_8;
    c_particle_double_handshake_channel   c_9;
    //straight links
    c_particle_double_handshake_channel   c25;
    c_particle_double_handshake_channel   c45;
    c_particle_double_handshake_channel   c63;
    c_particle_double_handshake_channel   c65;
    c_particle_double_handshake_channel   c36;
    c_particle_double_handshake_channel   c52;
    c_particle_double_handshake_channel   c54;
    c_particle_double_handshake_channel   c56;
    //torus links
    c_particle_double_handshake_channel   c_h3;
    c_particle_double_handshake_channel   c_v3;
#endif

#if NUM_DUT == 1
    DUT		dut;
#endif
#if NUM_DUT > 1 //4 DUTs
    DUT		dut1;
    DUT		dut2;
    DUT		dut3;
    DUT		dut4;
#endif
#if NUM_DUT > 4 //9 DUTs
    DUT		dut5;
    DUT		dut6;
    DUT		dut7;
    DUT		dut8;
    DUT		dut9;
#endif
    Testbench	testbench;

    SC_CTOR(TOP)
#if NUM_DUT == 1
    :c_north("c_north"), c_east("c_east"), c_south("c_south"), c_west("c_west"),
	c_1("c_1"), dut("dut"), testbench("testbench")
#endif
#if NUM_DUT == 4
    :c_1("c_1"), c_2("c_2"), c_3("c_3"), c_4("c_4"),
	c23("c23"), c32("c32"), c43("c43"), c34("c34"),
	c_h1("c_h1"), c_h2("c_h2"), c_v1("c_v1"), c_v2("c_v2"),
	dut1("dut1",1, 1), dut2("dut2",1, 2), dut3("dut3", 2, 1), dut4("dut4",2, 2),
	testbench("testbench")
#endif
#if NUM_DUT == 9
    :c_1("c_1"), c_2("c_2"), c_3("c_3"), c_4("c_4"),
	c23("c23"), c32("c32"), c43("c43"), c34("c34"),
	c_h1("c_h1"), c_h2("c_h2"), c_v1("c_v1"), c_v2("c_v2"),
	c_5("c_5"), c_6("c_6"), c_7("c_7"), c_8("c_8"), c_9("c_9"),
	c25("c25"), c45("c45"), c63("c63"), c65("c65"), c36("c36"),
	c52("c52"), c54("c54"), c56("c56"), c_h3("c_h3"), c_v3("c_v3"),
	dut1("dut1",1, 1), dut2("dut2",1, 2), dut3("dut3", 1, 3), dut4("dut4",2, 1),
	dut5("dut5",2, 2), dut6("dut6",2, 3), dut7("dut7", 3, 1), dut8("dut8", 3, 2),
	dut9("dut9",3, 3), testbench("testbench")
#endif
    {
    }

    void before_end_of_elaboration()
    {
	// port binding
#if NUM_DUT == 1
	//dut(c_1, c_north, c_east, c_south, c_west);
    dut.CoordsControl.bind(c_1);
    dut.CoordsNorth.bind(c_north);
    dut.CoordsEast.bind(c_east);
    dut.CoordsSouth.bind(c_south);
    dut.CoordsWest.bind(c_west);

	//testbench(c_north, c_east, c_south, c_west, c_1);
    testbench.CoordsNorth.bind(c_north);
    testbench.CoordsEast.bind(c_east);
    testbench.CoordsSouth.bind(c_south);
    testbench.CoordsWest.bind(c_west);
    testbench.Coords_1.bind(c_1);
#elif NUM_DUT == 4
	dut1(c_1, c_v1, c23, c32, c_h1);
	dut2(c_2, c_v2, c_h1, c34, c23);
	dut3(c_3, c32, c43, c_v1, c_h2);
	dut4(c_4, c34, c_h2, c_v2, c43);
	testbench.Coords_1.bind(c_1);
	testbench.Coords_2.bind(c_2);
	testbench.Coords_3.bind(c_3);
	testbench.Coords_4.bind(c_4);
	//testbench(c_1, c_2, c_3, c_4);
#elif NUM_DUT == 9
	dut1(c_1, c_v1, c23, c32, c_h1);
	dut2(c_2, c_v2, c25, c34, c23);
	dut3(c_3, c_v3, c_h1, c36, c25);
	dut4(c_4, c32, c43, c52, c_h2);
	dut5(c_5, c34, c45, c54, c43);
	dut6(c_6, c36, c_h2, c56, c45);
	dut7(c_7, c52, c63, c_v1, c_h3);
	dut8(c_8, c54, c65, c_v2, c63);
	dut9(c_9, c56, c_h3, c_v3, c65);
	testbench(c_1, c_2, c_3, c_4, c_5, c_6, c_7, c_8, c_9);
#endif
    }
};

//////////////////////////////////////////////////////////////////////////
// Top Instance
//////////////////////////////////////////////////////////////////////////

TOP top("top");

//////////////////////////////////////////////////////////////////////////
// main function
//////////////////////////////////////////////////////////////////////////

int sc_main(int argc, char* argv[])
{
#if VERBOSITY >= 1
    printf("NoC(%dx%d) particle simulation starts,\n"
		"\tsimulating %d particles over %d steps,\n"
		"\tusing %d tiles managing up to %d particles each.\n",
		NUM_COLS, NUM_ROWS, NUM_PARTICLES,
		NSTEPS, NUM_DUT, MAX_PARTICLES);
#endif
    assert(NUM_DUT == NUM_COLS * NUM_ROWS);
    sc_start();
#if VERBOSITY >= 2
    cout << "  Total simulated time = " << sc_time_stamp() << endl;
#endif

#if VERBOSITY >= 1
    printf("NoC(%dx%d) particle simulation done.\n",
		NUM_COLS, NUM_ROWS);
#endif

    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 noexpandtab: */
