#include "ooo_instrumentation.h"

#include <map>

#include "../internal_representation/design.h"
#include "../internal_representation/thread.h"
#include "../internal_representation/module.h"

#include "../tools/class_is_derived_from.h"
#include "../tools/command_line_parsing.h"
#include "../tools/function_helper.h"
#include "../tools/rose_nodes.h"
#include "../tools/systemc_definitions.h"

#include "../segment_graph/combined_data_conflict_table.h"
#include "../segment_graph/data_conflict_table.h"
#include "../segment_graph/event_conflict_table.h"
#include "../segment_graph/port_call_attribute.h"
#include "../segment_graph/prediction_event_notification_table.h"
#include "../segment_graph/prediction_time_advance_table.h"
#include "../segment_graph/segment.h"
#include "../segment_graph/segment_graph.h"
#include "../segment_graph/time_advance_table.h"

#include "../macros.h"

namespace risc {

namespace inst {

int counter = 0;

InstrumentationTraversal::InstrumentationTraversal(
    SgType* variableType,
    SgExpression* PrimChnlVarRefExp,
    SgExpression* SCChnlVarRefExp,
    SgClassDefinition* SCModuleClassDef,
    std::map<SgNode*, int>& FuncCallExptoID,
    std::string design_file_name)
  : variableType(variableType)
  , PrimChnlVarRefExp(PrimChnlVarRefExp)
  , SCChnlVarRefExp(SCChnlVarRefExp)
  , SCModuleClassDef(SCModuleClassDef)
  , FuncCallExptoID(FuncCallExptoID)
  , _psg_off_instrumented(false)
  , design_file_name(design_file_name)
  , first_func_decl_that_contains_wait(NULL)
{}

void
InstrumentationTraversal::visit(SgNode* astNode)
{
  SgMemberFunctionDeclaration* memfuncdecl
    = isSgMemberFunctionDeclaration(astNode);

  if (memfuncdecl != NULL)
  {
    // Instrument channel classes derived from ::sc_core::sc_prim_channel

    if (risc::tools::is_derived_from_sc_prim_channel(memfuncdecl->
        get_class_scope()))
    {

      // Due to ROSE limitation in template handling, our code instrumentation
      // cannont process template member functions correctly. Refer to Section
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        // TIM, too strict. 
        return;
        fprintf(stderr, "Function %s is a template member function in Class %s"
                " that is derived from ::sc_core::sc_prim_channel.\n",
                templmemfuncdecl->get_name().str(),
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template"
                " declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier
          = memfuncdecl->get_specialFunctionModifier();

        if ( !memfuncmodifier.isConstructor() && // except constructors
             !memfuncmodifier.isDestructor() ) // except destructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgVariableDeclaration* variableDeclaration
                = SageBuilder::buildVariableDeclaration("_sc_chnl_scoped_lock",
                  variableType, SageBuilder::buildCompoundInitializer(
                  SageBuilder::buildExprListExp(PrimChnlVarRefExp),
                  PrimChnlVarRefExp->get_type()));

              SageInterface::prependStatement(variableDeclaration, block);
            }
          }
        }
      }
    }

    // Instrument channel classes derived from ::sc_core::sc_channel

    if (risc::tools::is_derived_from_sc_channel(memfuncdecl->
        get_class_scope()))
    {

      // Due to ROSE limitation in template handling, our code instrumentation
      // cannont process template member functions correctly. Refer to Section
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s"
                " that is derived from ::sc_core::sc_channel.\n",
                templmemfuncdecl->get_name().str(),
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template"
                " declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier
          = memfuncdecl->get_specialFunctionModifier();

        if ( !memfuncmodifier.isConstructor() && // except constructors
             !memfuncmodifier.isDestructor() ) // except destructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgVariableDeclaration* variableDeclaration
                = SageBuilder::buildVariableDeclaration("_sc_chnl_scoped_lock",
                  variableType, SageBuilder::buildCompoundInitializer(
                  SageBuilder::buildExprListExp(SCChnlVarRefExp),
                  SCChnlVarRefExp->get_type()));

              SageInterface::prependStatement(variableDeclaration, block);
            }
          }
        }
      }
    }

    // Instrument module classes derived from ::sc_core::sc_module

    if (risc::tools::is_derived_from_sc_module(memfuncdecl->get_class_scope()) ||
        risc::tools::is_derived_from_sc_channel(memfuncdecl->get_class_scope()))
    {

      // Due to ROSE limitation in template handling, our code instrumentation
      // cannont process template member functions correctly. Refer to Section
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s"
                " that is derived from ::sc_core::sc_module.\n",
                templmemfuncdecl->get_name().str(),
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template"
                " declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier
          = memfuncdecl->get_specialFunctionModifier();

        if (memfuncmodifier.isConstructor()) // constructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 0) // bug fix RD+ZC 03/05/18
            {
              SgVariableDeclaration* InstCntVarDecl
                = SageBuilder::buildVariableDeclaration("_OoO_Instance_Counter",
                    SageBuilder::buildUnsignedIntType(),
                    SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(0))
                );
              InstCntVarDecl->get_declarationModifier().get_storageModifier().
                  setStatic();
                  
              SgExprStatement* assignStmt = SageBuilder::buildAssignStatement(
                SageBuilder::buildVarRefExp("m_instance_id", SCModuleClassDef),
                SageBuilder::buildPlusPlusOp(SageBuilder::buildVarRefExp(
                                               "_OoO_Instance_Counter",
                                               memfuncdecl->get_class_scope()),
                                               SgUnaryOp::postfix));

              SageInterface::prependStatement(assignStmt, block);
              SageInterface::prependStatement(InstCntVarDecl, block);
            }
          }
        }
      }
    }
  }

  // Instrument Instance Counters
  SgClassDefinition* classdf = isSgClassDefinition(astNode);

  if (classdf != NULL)
  {
    if (risc::tools::is_derived_from_sc_module(classdf) ||
        risc::tools::is_derived_from_sc_channel(classdf) )
    {
      /*
      // static unsigned int InstanceCounter
      SgVariableDeclaration* InstCntVarDecl
        = SageBuilder::buildVariableDeclaration("_OoO_Instance_Counter",
            SageBuilder::buildUnsignedIntType());
      InstCntVarDecl->get_declarationModifier().get_storageModifier().
          setStatic();

      SageInterface::prependStatement(InstCntVarDecl, classdf);

      // initialize InstanceCounter
      SgVariableDeclaration* InstCntInitDecl
        = SageBuilder::buildVariableDeclaration("_OoO_Instance_Counter",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(0)),
            classdf);

      SageInterface::insertStatementAfter(classdf->get_declaration(),
                                          InstCntInitDecl);
                                          */
    }
  }

  // Instrument segment boundaries
  SgFunctionCallExp* funccallexp = isSgFunctionCallExp(astNode);
  if (funccallexp != NULL)
  {
    if (FuncCallExptoID.find(funccallexp) != FuncCallExptoID.end()){
      if(PSG_MODE){
        if(_psg_off_instrumented == false){
          _psg_off_instrumented = true;
          first_func_decl_that_contains_wait = SageInterface::getEnclosingFunctionDeclaration(funccallexp);
          

        }
      
        SgExpression *offest_plus_id
          = SageBuilder::buildPntrArrRefExp(
            SageBuilder::buildVarRefExp("psg_seg_id_"+design_file_name,new SgGlobal()),
            SageBuilder::buildIntVal(FuncCallExptoID[funccallexp])
          );

        funccallexp->append_arg(offest_plus_id);
      }
      else{
        //at this point, this is a bug fix for only sc_fifo read.
        if(funccallexp->getAssociatedFunctionDeclaration() != NULL ){
          if(funccallexp->getAssociatedFunctionDeclaration()->get_name().getString() != "read")
            funccallexp->append_arg(SageBuilder::buildIntVal(
              FuncCallExptoID[funccallexp]));
          else {
            SgFunctionCallExp *sc_segid
              = SageBuilder::buildFunctionCallExp(
                  isSgFunctionSymbol(
                    risc::tools::SystemCDefinitions::get_sc_segid_ctor()
                      ->get_symbol_from_symbol_table()));
            sc_segid->append_arg(SageBuilder::buildIntVal(FuncCallExptoID[funccallexp]));
            funccallexp->append_arg(sc_segid);
          }
        }else{ //handles sc_port<sc_fifo_in/out<T>>->read() 
          if(funccallexp->get_function()!=NULL){
            SgArrowExp* arrow_exp = isSgArrowExp(funccallexp->get_function());
            SgDotExp* dot_exp = isSgDotExp(funccallexp->get_function());
            if(arrow_exp){
              SgMemberFunctionRefExp* f = isSgMemberFunctionRefExp(arrow_exp->get_rhs_operand());
              if(SageInterface::get_name(f->getAssociatedMemberFunctionDeclaration()) != "read")
                funccallexp->append_arg(SageBuilder::buildIntVal(
                  FuncCallExptoID[funccallexp]));
              else {
                SgFunctionCallExp *sc_segid
                  = SageBuilder::buildFunctionCallExp(
                      isSgFunctionSymbol(
                        risc::tools::SystemCDefinitions::get_sc_segid_ctor()
                          ->get_symbol_from_symbol_table()));
                sc_segid->append_arg(SageBuilder::buildIntVal(FuncCallExptoID[funccallexp]));
                funccallexp->append_arg(sc_segid);
              }
            }
            else if(dot_exp){
              SgMemberFunctionRefExp* f = isSgMemberFunctionRefExp(dot_exp->get_rhs_operand());
              if(SageInterface::get_name(f->getAssociatedMemberFunctionDeclaration()) != "read")
                funccallexp->append_arg(SageBuilder::buildIntVal(
                  FuncCallExptoID[funccallexp]));
              else {
                SgFunctionCallExp *sc_segid
                  = SageBuilder::buildFunctionCallExp(
                      isSgFunctionSymbol(
                        risc::tools::SystemCDefinitions::get_sc_segid_ctor()
                          ->get_symbol_from_symbol_table()));
                sc_segid->append_arg(SageBuilder::buildIntVal(FuncCallExptoID[funccallexp]));
                funccallexp->append_arg(sc_segid);
              }
            }
          }
        }
        
        
      }
    } 

    SgFunctionDeclaration* funcdecl
      = funccallexp->getAssociatedFunctionDeclaration();

    if (funcdecl != NULL)
    {
      if (funcdecl->get_name().getString() == "create_thread_process" ||
          funcdecl->get_name().getString() == "create_method_process" ||
          funcdecl->get_name().getString() == "create_cthread_process")
        funccallexp->append_arg(SageBuilder::buildVarRefExp(
            "m_instance_id", SCModuleClassDef));
    }
  }
}

int
risc_link(SgProject* project, risc::tools::ParseStatus* parseStatus, const std::string tool_name)
{
  //std::cout << "into link" << std::endl;
  if (project->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  if (project->numberOfFiles() == 0 && project->numberOfDirectories() == 0)
  {
    if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
      std::cout << " risc_link may be encountering an object file" << std::endl;
  } else {
    // normal cases that rose translators will actually do something about
    // the input files and we have SgFile for each of the files.
    if (project->get_compileOnly() ||
        project->get_file(0).get_skipfinalCompileStep() ||
        project->get_file(0).get_skip_unparse())
    {
      if (parseStatus->verbosityLevel >= RISC_VERBOSE)
        std::cout << tool_name << ": skipping risc_link..." << std::endl;
      return 0;
    }
  }

  // Compile the output file from the unparsing
  std::vector<std::string> argcArgvList
    = project->get_originalCommandLineArgumentList();

  // error checking
  if (project->numberOfFiles() != 0)
    ROSE_ASSERT(argcArgvList.size() > 1);

  // strip out any rose options before passing the command line.
  SgFile::stripRoseCommandLineOptions(argcArgvList);

  // strip out edg specific options that would cause an error
  // in the backend linker (compiler).
  SgFile::stripEdgCommandLineOptions(argcArgvList);

  // remove the original compiler/linker name
  argcArgvList.erase(argcArgvList.begin());

  // remove all original file names
  Rose_STL_Container<std::string> sourceFilenames
    = project->get_sourceFileNameList();
  for (Rose_STL_Container<std::string>::iterator i = sourceFilenames.begin();
       i != sourceFilenames.end(); i++)
  {
#if USE_ABSOLUTE_PATHS_IN_SOURCE_FILE_LIST
#error "USE_ABSOLUTE_PATHS_IN_SOURCE_FILE_LIST is not supported yet"
    // DQ (9/1/2006): Check for use of absolute path and convert filename
    // to absolute path if required
    bool usesAbsolutePath = ((*i)[0] == '/');
    if (usesAbsolutePath == false)
    {
      std::string targetSourceFileToRemove
        = StringUtility::getAbsolutePathFromRelativePath(*i);
      if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
        std::cout << "  Converting source file to absolute path to search "
                  << "for it and remove it! targetSourceFileToRemove = "
                  << targetSourceFileToRemove << std::endl;
      argcArgvList.remove(targetSourceFileToRemove);
    } else {
      argcArgvList.remove(*i);
    }
#else
    if (find(argcArgvList.begin(),argcArgvList.end(),*i) != argcArgvList.end())
    {
      argcArgvList.erase(find(argcArgvList.begin(),argcArgvList.end(),*i));
    }
#endif
  }

  for (Rose_STL_Container<std::string>::iterator i = argcArgvList.begin();
       i != argcArgvList.end(); i++)
  {
    std::string::size_type startingQuote = i->find("\"");
    if (startingQuote != std::string::npos)
    {
      std::string::size_type endingQuote = i->rfind("\"");

      // There should be a double quote on both ends of the string
      ROSE_ASSERT (endingQuote != std::string::npos);

      std::string quotedSubstring = i->substr(startingQuote, endingQuote);

      std::string fixedQuotedSubstring
        = std::string("\\\"") + quotedSubstring + std::string("\\\"");

      // Now replace the quotedSubstring with the fixedQuotedSubstring
      i->replace(startingQuote,endingQuote,fixedQuotedSubstring);

      if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
        std::cout << "  Modified argument = " << *i << std::endl;
    }
  }

     const std::string whiteSpace = " ";

  // This is a better implementation since it will include any additional
  // command line options that target the linker
  Rose_STL_Container<std::string> linkingCommand ;

  linkingCommand.push_back(parseStatus->compilerName);
  // Find all object files generated at file level compilation
  // The assumption is that -o objectFileName is made explicit and
  // is generated by SgFile::generateOutputFileName()
  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    // DQ (2/25/2014): If this file was supressed in the compilation to build
    // an object file then it should be supressed in being used
    // in the linking stage.
    if (project->get_file(i).get_skipfinalCompileStep() == false)
    {
      linkingCommand.push_back(project->get_file(i).generateOutputFileName());
    }
  }

  // Add any options specified in the original command line
  // (after preprocessing)
  linkingCommand.insert(linkingCommand.end(),
                        argcArgvList.begin(),
                        argcArgvList.end());

  // Check if -o option exists, otherwise append -o a.out to the command line

  // Additional libraries to be linked with
  if(SageInterface::getProject()->get_openmp_linking())
  {
    // Sara Royuela 12/10/2012:  Add GCC version check
#ifdef USE_ROSE_GOMP_OPENMP_LIBRARY
#if (__GNUC__ < 4 || \
    (__GNUC__ == 4 && (__GNUC_MINOR__ < 4)))
#warning "GNU version lower than expected"
     std::cout << tool_name << ": GCC version must be 4.4.0 or later when linking "
              << "with GOMP OpenMP Runtime Library " << std::endl
              << "(OpenMP tasking calls are not implemented "
              << "in previous versions)" << std::endl;
      ROSE_ASSERT(false);
#endif

      // add libxomp.a , Liao 6/12/2010
      std::string xomp_lib_path(ROSE_INSTALLATION_PATH);
      ROSE_ASSERT(xomp_lib_path.size() != 0);
      // static linking for simplicity
      linkingCommand.push_back(xomp_lib_path + "/lib/libxomp.a");

      // lib path is available if --with-gomp_omp_runtime_library=XXXX is used
      std::string gomp_lib_path(GCC_GOMP_OPENMP_LIB_PATH);
      ROSE_ASSERT(gomp_lib_path.size() != 0);
      linkingCommand.push_back(gomp_lib_path + "/libgomp.a");
      linkingCommand.push_back("-lpthread");
#else
      // GOMP has higher priority when both GOMP and OMNI are specified
      // (wrongfully)
#ifdef OMNI_OPENMP_LIB_PATH
      // a little redundant code to defer supporting 'ROSE_INSTALLATION_PATH'
      // in cmake
      std::string xomp_lib_path(ROSE_INSTALLATION_PATH);
      ROSE_ASSERT(xomp_lib_path.size() != 0);
      linkingCommand.push_back(xomp_lib_path + "/lib/libxomp.a");

      std::string omni_lib_path(OMNI_OPENMP_LIB_PATH);
      ROSE_ASSERT(omni_lib_path.size() != 0);
      linkingCommand.push_back(omni_lib_path + "/libgompc.a");
      linkingCommand.push_back("-lpthread");
#else
     if (parseStatus->warningLevel >= RISC_WARNING)
     std::cout << "Warning: OpenMP lowering is requested but no target runtime "
               << "library is specified!" << std::endl;
#endif
#endif
  }

  if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
    std::cout << "  In risc_link command line = " << std::endl
              << CommandlineProcessing::generateStringFromArgList(
                   linkingCommand,false,false)
              << std::endl;

  int errorcode = system(CommandlineProcessing::generateStringFromArgList(
                        linkingCommand,false,false).c_str());

  if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
    std::cout << "  linker error status = " << errorcode << std::endl;

  //std::cout << "out link" << std::endl;
  // return the error code from the linking
  return errorcode;
}

void
risc_buildCompilerCommandLineOptions(SgFile* file,
                                     std::vector<std::string>& argv,
                                     std::vector<std::string>& compilerCmdLine,
                                     risc::tools::ParseStatus* parseStatus,
                                     const std::string tool_name)
{
  if (file->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
  {
    std::cout << "  risc_buildCompilerCommandLineOptions(): compilerName = "
              << parseStatus->compilerName << std::endl;
  }

  compilerCmdLine.push_back(parseStatus->compilerName);

  // TOO1 (2014-10-09): Use the correct Boost version that ROSE was configured
  // --with-boost
#ifdef ROSE_BOOST_PATH
  // Search dir for header files, after all directories specified by -I but
  // before the standard system directories.
  compilerCmdLine.push_back("-isystem");
  compilerCmdLine.push_back(std::string(ROSE_BOOST_PATH) + "/include");
#endif

  // DQ (7/26/2014): Adding support for C11 (option to backend compiler).
  if (file->get_Cxx11_only() == true)
  {
    if (file->get_Cxx11_gnu_only() == true)
    {
      compilerCmdLine.push_back("-std=gnu++11");
    } else {
      compilerCmdLine.push_back("-std=c++11");
    }
  } else {
    // DQ (7/26/2014): Adding support for C14 (option to backend compiler).
    if (file->get_Cxx14_only() == true)
    {
      // DQ (7/27/2014): These options are not available in GNU g++ yet.
      compilerCmdLine.push_back("-std=c++14");
    }
  }
  // tps (28 Aug 2008) : changed this so it does not pick up mpicc for icc
  std::string name = Rose::StringUtility::stripPathFromFileName(compilerCmdLine[0]);
  if (name == "icpc")
  {
    // This is the Intel C++ compiler: icc, we need to add the -restrict option
    compilerCmdLine.push_back("-restrict");
  }

  if (file->get_strict_language_handling() == true)
  {
    // Check if it is appears as "-ansi" on the original commandline
    if (CommandlineProcessing::isOption(argv,"-","ansi",false) == true)
    {
      if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
        std::cout << "  Option -ansi detected on the original commandline"
                  << std::endl;
    } else {
      // This is might be specific to GNU
      compilerCmdLine.push_back("-ansi");
    }
  }

#ifdef ROSE_USE_NEW_EDG_INTERFACE
  // DQ (1/12/2009): Allow in internal indicator that EDG version 3.10 or 4.0
  // (or greater) is in use to be properly passed on the compilation of
  // the generated code.
  compilerCmdLine.push_back("-DROSE_USE_NEW_EDG_INTERFACE");
#endif

  // Since we need to do this often, support is provided in the utility_functions.C
  // and we can simplify this code.
  std::string currentDirectory = file->getWorkingDirectory();

  // specify compilation only option (new style command line processing)
  if (CommandlineProcessing::isOption(argv,"-","c",false) == true)
  {
    file->set_compileOnly(true);
  }

  compilerCmdLine.push_back("-DUSE_ROSE");

  compilerCmdLine.push_back("-DUSE_ROSE_BACKEND");

  if (file->get_openmp_lowering())
  {
    compilerCmdLine.push_back("-D_OPENMP");
  }

  // DQ (3/31/2004): Now cleaned up source file handling
  Rose_STL_Container<std::string> argcArgvList = argv;

  // DQ (9/25/2007): Moved to std::vector from std::list uniformly within ROSE.
  // Remove the first argument (argv[0])
  argcArgvList.erase(argcArgvList.begin());

  // DQ (4/20/2006): Modified to only do this when generating code and compiling it
  // Remove the source names from the argcArgvList
  // (translated versions of these will be inserted later)
  if (file->get_skip_unparse() == false)
  {
    // DQ (1/24/2010): Now that we have directory support, the parent of a SgFile
    // does not have to be a SgProject.
    SgProject* project = TransformationSupport::getProject(file);
    ROSE_ASSERT(project != NULL);
    Rose_STL_Container<std::string> sourceFilenames
      = project->get_sourceFileNameList();
    for (Rose_STL_Container<std::string>::iterator i = sourceFilenames.begin();
         i != sourceFilenames.end(); i++)
    {
#if USE_ABSOLUTE_PATHS_IN_SOURCE_FILE_LIST
#error "USE_ABSOLUTE_PATHS_IN_SOURCE_FILE_LIST is not supported yet"
      // DQ (9/1/2006): Check for use of absolute path and
      // convert filename to absolute path if required
      bool usesAbsolutePath = ((*i)[0] == '/');
      if (usesAbsolutePath == false)
      {
        std::string targetSourceFileToRemove
          = StringUtility::getAbsolutePathFromRelativePath(*i);
        if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
          std::cout << "  Converting source file to absolute path to search "
                    << "for it and remove it! targetSourceFileToRemove = "
                    << targetSourceFileToRemove << std::endl;
        argcArgvList.remove(targetSourceFileToRemove);
      } else {
        argcArgvList.remove(*i);
      }
#else
      if (find(argcArgvList.begin(), argcArgvList.end(), *i) !=
          argcArgvList.end())
      {
        argcArgvList.erase(find(argcArgvList.begin(), argcArgvList.end(), *i));
      }
#endif
    }
  }

  // AS(080704) Fix so that if user specifies name of -o file
  // rose do not specify another in addition
  bool objectNameSpecified = false;
  for (Rose_STL_Container<std::string>::iterator i = argcArgvList.begin();
       i != argcArgvList.end(); i++)
  {
    if (i->substr(0,2) == "-o")
    {
      // DQ (6/12/2005): Added error checking!
      if (objectNameSpecified == true)
      {
        // Error: "-o" has been specified twice
        std::cout << "Error: \"-o \" has been specified twice" << std::endl;
      }
      ROSE_ASSERT(objectNameSpecified == false);
      objectNameSpecified = true;
    }
  }

  Rose_STL_Container<std::string> tempArgcArgv;
  for (Rose_STL_Container<std::string>::iterator i = argcArgvList.begin();
       i != argcArgvList.end(); i++)
  {
    if(i->substr(0,2) != "-l")
    {
      tempArgcArgv.push_back(*i);
    }
  }
  argcArgvList.swap(tempArgcArgv);
  for (Rose_STL_Container<std::string>::iterator i = argcArgvList.begin();
       i != argcArgvList.end(); i++)
  {
    // DQ (1/26/2006): Fix for 64 bit support.
    std::string::size_type startingQuote = i->find("\"");
    if (startingQuote != std::string::npos)
    {
      // This string at least has a quote
      std::string::size_type endingQuote = i->rfind("\"");
      // There should be a double quote on both ends of the string
      ROSE_ASSERT(endingQuote != std::string::npos);

      std::string::size_type substringWithoutQuotesSize
        = ((endingQuote-1) - (startingQuote+1)) + 1;
      // Generate the string without quotes so that we can rebuild the quoted
      // string. This is more critical if there were escpes before the quotes
      // in the original string.
      std::string quotedSubstringWithoutQuotes
        = i->substr(startingQuote+1,substringWithoutQuotesSize);
      std::string fixedQuotedSubstring
        = std::string("\"") + quotedSubstringWithoutQuotes + std::string("\"");
      // Now replace the quotedSubstring with the fixedQuotedSubstring
      i->replace(startingQuote,endingQuote,fixedQuotedSubstring);
    }
  }

  // Add any options specified by the user (and add space at the end)
  compilerCmdLine.insert(compilerCmdLine.end(),
                         argcArgvList.begin(),
                         argcArgvList.end());

  std::string sourceFileName = file->get_sourceFileNameWithPath();

  std::string oldFileNamePathOnly
    = Rose::StringUtility::getPathFromFileName(sourceFileName.c_str());
  std::string oldFileName = Rose::StringUtility::stripPathFromFileName(sourceFileName.c_str());

  // DQ (4/13/2014): Added support to avoid output of a specified include path
  // twice.
  bool oldFileNamePathOnlyAlreadySpecifiedAsIncludePath = false;
  for (std::vector<std::string>::iterator i = argcArgvList.begin();
       i != argcArgvList.end(); i++)
  {
    std::string s = std::string("-I") + oldFileNamePathOnly;
    if (s == *i)
    {
      oldFileNamePathOnlyAlreadySpecifiedAsIncludePath = true;
    }
  }

  // DQ (4/2/2011): Java does not have -I as an accepted option.
  if (oldFileNamePathOnly.empty() == false &&
      oldFileNamePathOnlyAlreadySpecifiedAsIncludePath == false)
  {
    std::vector<std::string>::iterator iter;
    // find the very first -Ixxx option's position
    for (iter = compilerCmdLine.begin(); iter != compilerCmdLine.end(); iter++)
    {
      std::string cur_string = *iter;
      std::string::size_type pos = cur_string.find("-I",0);
      if (pos==0)
        break;
    }
    if (file->get_unparseHeaderFiles() == false)
    {
      // DQ (9/15/2013): Added support for generated file to be placed into
      // the same directory as the source file. When
      // (get_unparse_in_same_directory_as_input_file() == true) we don't want
      // to add the include path to the source directory.
      SgProject* project = TransformationSupport::getProject(file);
      if (project != NULL)
      {
        if (project->get_unparse_in_same_directory_as_input_file() == false)
        {
          compilerCmdLine.insert(iter, std::string("-I") + oldFileNamePathOnly);
        }
      } else {
          std::cout << "ERROR: In risc_buildCompilerCommandLineOptions(): "
                    << "file = " << file << " has no associated project"
                    << std::endl;
          ROSE_ASSERT(false);
      }
    }
  }

  if (file->get_openmp_lowering())
  {
    std::vector<std::string>::iterator iter;
    std::vector<std::string>::iterator iter_last_inc = compilerCmdLine.begin();
    // find the very last -Ixxx option's position
    // This for loop cannot be merged with the previous one due to
    // iterator invalidation rules.
    for (iter = compilerCmdLine.begin(); iter != compilerCmdLine.end(); iter++)
    {
      std::string cur_string = *iter;
      std::string::size_type pos = cur_string.find("-I",0);
      if (pos==0)
      {
        iter_last_inc = iter;
      }
    }
    // accommodate the insert-before-an-iterator semantics used
    // in vector::insert()
    if (iter_last_inc != compilerCmdLine.end())
      iter_last_inc ++;

#ifdef ROSE_INSTALLATION_PATH
    std::string include_path(ROSE_INSTALLATION_PATH);
    include_path += "/include/rose";
    compilerCmdLine.insert(iter_last_inc, "-I"+include_path);
#endif
  }

  // DQ (4/20/2006): This allows the ROSE translator to be just a wrapper
  // for the backend (vendor) compiler.
  if (file->get_skip_unparse() == false)
  {
    // Generate the name of the ROSE generated source file
    // (instead of the original source file) and this file will be
    // compiled by the backend (vendor) compiler.
    ROSE_ASSERT(file->get_unparse_output_filename().empty() == false);
    compilerCmdLine.push_back(file->get_unparse_output_filename());
  }

  if (file->get_compileOnly() == true)
  {
    std::string objectFileName = file->generateOutputFileName();
    // DQ (7/14/2004): Suggested fix from Andreas, make the object file name
    // explicit
    if (objectNameSpecified == false)
    {
      compilerCmdLine.push_back("-o");
      compilerCmdLine.push_back(currentDirectory + "/" + objectFileName);
    }
  } else {
    compilerCmdLine.push_back("-c");
    // For compile+link mode, -o is used for the final executable, if it exists
    // We make -o objectfile explicit
    std::string objectFileName = file->generateOutputFileName();
    compilerCmdLine.push_back("-o");
    compilerCmdLine.push_back(currentDirectory + "/" + objectFileName);
  }
}

int
risc_compileOutputFile(SgFile* file, risc::tools::ParseStatus* parseStatus, const std::string tool_name)
{

//  std::cout << "into risc_compileOutputFile" << std::endl;
  // RISC only supports C++/SystemC codes
  if (file->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  std::vector<std::string> argv
    = file->get_originalCommandLineArgumentList();
  ROSE_ASSERT(!argv.empty());


  //ZC 2018/09/25 remove .o files
  //however, the first .o file after -c is not removed

  int coFound = 0; //if -c -o are both found
  for(std::vector<std::string>::iterator iter = argv.begin(); 
      iter!=argv.end();
      )
  {
    std::string s = *iter;
    if(s=="-c" || s=="-o") {
      ++coFound;
      iter ++;
    }
    else if(s.find(".o")!=std::string::npos){
      if(coFound==2){
        coFound = 0;
        iter++;
        //std::cout << s <<" " << coFound << std::endl;
      }
      else
        argv.erase(iter);
    }
    else
      iter++;
  }
  //std::cout << std::endl;
  /*for(int i=0;i<argv.size();i++){
    std::cout << argv[i] << std::endl;
  }
  */
  // error checking
  ROSE_ASSERT(argv.size() > 1);

  // strip out any rose options before passing the command line.
  SgFile::stripRoseCommandLineOptions(argv);

  // strip out edg specific options that would cause an error
  // in the backend linker (compiler).
  SgFile::stripEdgCommandLineOptions(argv);

  int returnValueForRose = 0;

  bool use_original_input_file =
    // (1) An ROSE unparsed file was not generated
    file->get_unparse_output_filename().empty() == true ||
    // (2) File has an error and user specified to "keep going"
    (
      (
        file->get_frontendErrorCode() != 0 ||
        file->get_project()->get_midendErrorCode() != 0 ||
        file->get_unparserErrorCode() != 0 ||
        file->get_backendCompilerErrorCode() != 0) &&
      (
        file->get_project()->get_keep_going()
      )
    );

  // TOO1 (05/14/2013): Handling for -rose:keep_going
  // Replace the unparsed file with the original input file.
  if (use_original_input_file)
  {
    std::string outputFilename = file->get_sourceFileNameWithPath();

    SgProject* project = TransformationSupport::getProject(file);
    if (project != NULL)
    {
      if (project->get_unparse_in_same_directory_as_input_file() == true)
      {
        outputFilename
          = Rose::StringUtility::getPathFromFileName(file->get_sourceFileNameWithPath())
            + "/" + RISC_UNPARSE_FILE_PREFIX
            + file->get_sourceFileNameWithoutPath();

        if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
        {
          std::cout << "  In risc_compileOutputFile(): Using filename for "
                    << "unparsed file into same directory as input file: "
                    << "outputFilename = " << outputFilename
                    << std::endl;
        }

        file->set_unparse_output_filename(outputFilename);
      }
    } else {
      if (parseStatus->warningLevel >= 1) /* RISC_WARNING */
      {
        std::cout << "WARNING: In risc_compileOutputFile(): file = " << file
                  <<  " has no associated project" << std::endl;
      }
    }

    if (!file->get_unparse_output_filename().empty())
    {
      boost::filesystem::path original_file = outputFilename;
      boost::filesystem::path unparsed_file
        = file->get_unparse_output_filename();

      if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
      {
        std::cout << "  [DEBUG] " << "unparsed_file "
                  << "'" << unparsed_file << "' "
                  << "exists = "
                  << std::boolalpha
                  << boost::filesystem::exists(unparsed_file)
                  << std::endl;
      }
      // Don't replace the original input file with itself
      if (original_file.string() != unparsed_file.string())
      {
        if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
        {
          std::cout << "  [INFO] " << "Replacing "
                    << "'" << unparsed_file << "' "
                    << "with "
                    << "'" << original_file << "'"
                    << std::endl;
        }

        if (boost::filesystem::exists(unparsed_file))
        {
          boost::filesystem::remove(unparsed_file);
        }

        boost::filesystem::copy_file(original_file,
                                     unparsed_file,
          boost::filesystem::copy_option::overwrite_if_exists);
      }
    }

    if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
    {
      std::cout << "  In risc_compileOutputFile(): outputFilename = "
                << outputFilename << std::endl;
    }

    file->set_unparse_output_filename(outputFilename);
  }

  ROSE_ASSERT(file->get_unparse_output_filename().empty() == false);

  // Now call the compiler that rose is replacing
  if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
  {
    std::cout << "  Calling the backend (vendor's) compiler compilerName = "
              << parseStatus->compilerName << " for file = "
              << file->get_unparse_output_filename() << std::endl;
  }

  // Build the commandline to hand off to the C++ compiler
  std::vector<std::string> compilerCmdLine;
  risc_buildCompilerCommandLineOptions (file, argv,
                                        compilerCmdLine,
                                        parseStatus,
                                        tool_name);

  int returnValueForCompiler = 0;

  // Allow conditional skipping of the final compile step for testing ROSE.
  if (file->get_skipfinalCompileStep() == false)
  {
    // Debugging code
    if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
    {
      std::cout << "  Calling system(compilerCmdLine)" << std::endl;
      std::cout << "  Number of command line arguments: "
                << compilerCmdLine.size() << std::endl;
      for (size_t i = 0; i < compilerCmdLine.size(); ++i)
      {
        std::cout << "  Backend compiler arg[" << i
                  <<  "]: = " << compilerCmdLine[i] << std::endl;
      }
      std::cout << "  End of command line for backend compiler" << std::endl;

      std::cout << "  In risc_compileOutputFile compiler command line  = "
                << std::endl
                << CommandlineProcessing::generateStringFromArgList(
                     compilerCmdLine, false, false)
                << std::endl;
    }

    returnValueForCompiler
      = system(CommandlineProcessing::generateStringFromArgList(
                 compilerCmdLine, false, false).c_str());

    // TOO1 (05/14/2013): Handling for -rose:keep_going
    //
    // Compilation failed =>
    //
    //   1. Unparsed file is invalid => Try compiling the original input file
    //   2. Original input file is invalid => abort
    if (returnValueForCompiler != 0)
    {
      file->set_backendCompilerErrorCode(-1);
      if (file->get_project()->get_keep_going() == true)
      {
        // 1. We already failed the compilation of the ROSE unparsed file.
        // 2. Now we tried to compile the original input file --
        //    that was just compiled above -- and failed also.
        if (file->get_unparsedFileFailedCompilation())
        {
          file->set_backendCompilerErrorCode(-1);

          std::cout  << "[FATAL] "
                     << "Original input file is invalid: "
                     << "'" << file->getFileName() << "'"
                     << std::endl;
          exit(1);
        } else {
          // The ROSE unparsed file is invalid...
          file->set_frontendErrorCode(-1);
          file->set_unparsedFileFailedCompilation(true);

          // So try to compile the original input file instead...
          returnValueForCompiler = risc_compileOutputFile(file, parseStatus, tool_name);
        }
      }
    }
  } else {
    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
      std::cout << "  COMPILER NOT CALLED: compilerNameString = <unknown>"
                << std::endl;

      if ( (file->get_binary_only() == false) &&
           (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE) )
      {
        std::cout << "  Skipped call to backend vendor compiler!" << std::endl;
      }
  }

  if (returnValueForCompiler < 0)
  {
    perror("Serious Error returned from vendor compiler");
  }

  // Assemble an exit status that combines the values for ROSE and
  // the C++/C compiler, and return an exit status which is the boolean OR of
  // the bits from the EDG/SAGE/ROSE and the compile step
  int finalCompiledExitStatus = returnValueForRose | returnValueForCompiler;

  if (finalCompiledExitStatus != 0)
  {
    // If this it is non-zero then make it 1 to be more clear
    // to external tools (e.g. make)
    finalCompiledExitStatus = 1;
  }

  // DQ (9/19/2006): We need to invert the test result (return code) for
  // negative tests (where failure is expected and success is an error).
  if (file->get_negative_test() == true)
  {
    if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
      std::cout << "  This is a negative tests, so an error in compilation is "
                << "a PASS and successful compilation is a FAIL "
                << "(vendor compiler return value = "
                << returnValueForCompiler << ")" << std::endl;

    finalCompiledExitStatus
      = (finalCompiledExitStatus == 0) ? /* error */ 1 : /* success */ 0;
  }

  if(finalCompiledExitStatus != 0) {
    std::cout << "Error: Output file could not be generated." << std::endl;
    exit(1);
  }

  //std::cout << "out risc_compileOutputFile" << std::endl;
  return finalCompiledExitStatus;
}

int
risc_compileOutput(SgProject* project, risc::tools::ParseStatus* parseStatus, const std::string tool_name)
{
  int errorCode = 0;
  int linkingReturnVal = 0;
  int i = 0;
  //std::cout << "into risc_compileOutput" << std::endl;
  // RISC only supports C++/SystemC codes
  if (project->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  if (project->get_stop_after_compilation_do_not_assemble_file() == true)
  {
    std::cout << tool_name << ": GNU \"-S\" option is not supported for more than "
              << "the C language in ROSE at present! "
              << "However, " << tool_name << " only supports C++."
              << std::endl;
    ROSE_ASSERT(false);
  }

  // case 1: preprocessing only
  if (project->get_C_PreprocessorOnly() == true)
  {
    // DQ (10/16/2005): Handle special case
    // (issue a single compile command for all files)
    std::vector<std::string> argv
      = project->get_originalCommandLineArgumentList();

    // strip out any rose options before passing the command line.
    SgFile::stripRoseCommandLineOptions(argv);

    // strip out edg specific options that would cause an error
    // in the backend linker (compiler).
    SgFile::stripEdgCommandLineOptions(argv);

    std::vector<std::string> originalCommandLine = argv;
    ROSE_ASSERT(!originalCommandLine.empty());

    std::string & compilerNameString = originalCommandLine[0];
    compilerNameString = parseStatus->compilerName;

    // TOO1 (2014-10-09): Use the correct Boost version
    // that ROSE was configured --with-boost
#ifdef ROSE_BOOST_PATH
    // Search dir for header files, after all directories specified by -I but
    // before the standard system directories.
    originalCommandLine.push_back("-isystem");
    originalCommandLine.push_back(std::string(ROSE_BOOST_PATH) + "/include");
#endif

    // Debug: Output commandline arguments before actually executing
    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
    {
      for (unsigned int i=0; i < originalCommandLine.size(); ++i)
      {
        std::cout << "  originalCommandLine[" << i << "] = "
                  << originalCommandLine[i] << std::endl;
      }
    }

    errorCode = system(CommandlineProcessing::generateStringFromArgList(
                         originalCommandLine, false, false).c_str());

    ROSE_ASSERT(SgNode::get_globalFunctionTypeTable() != NULL);
    ROSE_ASSERT(SgNode::get_globalFunctionTypeTable()->get_parent() != NULL);

  } else { // non-preprocessing-only case
    // case 2: compilation  for each file
    // Typical case
    for (i = 0; i < project->numberOfFiles(); i++)
    {
      int localErrorCode = 0;
      SgFile& file = project->get_file(i);

      localErrorCode = risc_compileOutputFile(&file, parseStatus, tool_name);

      if (localErrorCode > errorCode)
      {
        errorCode = localErrorCode;
      }
    }

    // case 3: linking at the project level
#ifndef _MSC_VER
    linkingReturnVal = risc_link(project, parseStatus, tool_name);
#else
    #pragma message ("sageSupport.C : linkingReturnVal = link (compilerName); "\
                     "not implemented yet.")
#endif
  } // end if preprocessing-only is false


  //std::cout << "out risc_compileOutput" << std::endl;
  // return errorCode;
  return errorCode + linkingReturnVal;
}

void
risc_unparse(risc::Design &design,
    risc::tools::ParseStatus* parseStatus,
    const std::string tool_name,
    UnparseFormatHelp *unparseFormatHelp,
    UnparseDelegate* unparseDelegate)
{
  SgProject* project = design.project_;

  // RISC only supports C++/SystemC codes
  if (project->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  // Hack the prefix of unparsed output filenames
//if (parseStatus->verbosityLevel >= RISC_VERY_VERBOSE)
//  std::cout << tool_name << ": hacking the filename prefix..." << std::endl;

  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    SgFile& file = project->get_file(i);
    if (file.get_unparse_output_filename().empty() == true)
    {
      file.set_unparse_output_filename(RISC_UNPARSE_FILE_PREFIX +
        file.get_sourceFileNameWithoutPath());
    }
  }

  // generate C++ source code
  project->unparse(unparseFormatHelp,unparseDelegate);
}

/*** THIS PATCHING AROUND ROSE LIMITATIONS IS NOT NEEDED ANY MORE
     WITH THE NEW ROSE LIBRARY VERSION 0.9.10.X (RD, 05/30/18)
void
risc_patch(risc::Design &design,
    risc::tools::ParseStatus* parseStatus,
    const char *methodPattern,
    const char *threadPattern,
    const char *cthreadPattern,
    const std::string tool_name)
{
  SgProject* project = design.project_;

  // RISC only supports C++/SystemC codes
  if (project->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  // Patch the static_cast errors
  char SysCmd[300];

  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    SgFile& file = project->get_file(i);
    sprintf(SysCmd, methodPattern,
            file.get_sourceFileNameWithoutPath().c_str());

    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
      std::cout << "  " << SysCmd << std::endl;

    if (system(SysCmd) != 0)
    {
      std::cout << "\nERROR: Patch File ";
      std::cout << file.get_sourceFileNameWithoutPath();
      std::cout << "failed" << std::endl;
      exit(1);
    }

    sprintf(SysCmd, threadPattern,
            file.get_sourceFileNameWithoutPath().c_str());

    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
      std::cout << "  " << SysCmd << std::endl;

    if (system(SysCmd) != 0)
    {
      std::cout << "\nERROR: Patch File ";
      std::cout << file.get_sourceFileNameWithoutPath();
      std::cout << "failed" << std::endl;
      exit(1);
    }

    sprintf(SysCmd, cthreadPattern,
            file.get_sourceFileNameWithoutPath().c_str());

    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
      std::cout << "  " << SysCmd << std::endl;

    if (system(SysCmd) != 0)
    {
      std::cout << "\nERROR: Patch File ";
      std::cout << file.get_sourceFileNameWithoutPath();
      std::cout << "failed" << std::endl;
      exit(1);
    }
  }

  if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
    std::cout << "  source file(s) generated" << std::endl;
}
***/

int
risc_compile(risc::Design &design, risc::tools::ParseStatus* parseStatus,
    const std::string tool_name)
{
  //std::cout << "into risc_compile" <<std::endl;
  int finalCombinedExitStatus = 0;
  SgProject* project = design.project_;

  // RISC only supports C++/SystemC codes
  if (project->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  if (project->numberOfFiles() > 0 || project->numberOfDirectories() > 0)
  {
    // Compile generated C++ source code with vendor compiler.
    // Generate object file (required for further template processing
    // if templates exist).
//    if (parseStatus->verbosityLevel >= RISC_VERY_VERBOSE)
//      std::cout << tool_name << ": compiling unparsed source file(s) with "
//                << "vendor compiler..." << std::endl;

      // Use risc_compileOutput
      finalCombinedExitStatus = risc_compileOutput(project, parseStatus, tool_name);
  }
  else
  {
    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
      std::cout << "   project->get_compileOnly() = "
                << (project->get_compileOnly() ? "true" : "false")
                << std::endl;

    bool callTemplateInstantation
         = (project->get_template_instantiation_mode() == SgProject::e_none);

    if (callTemplateInstantation == true)
    {
      std::cout << "ROSE: I don't think we need to call instantiateTemplates() "
                << "any more!" << std::endl;
      ROSE_ASSERT(false);
    }

//    if (parseStatus->verbosityLevel >= RISC_VERY_VERBOSE)
//      std::cout << tool_name << ": linking object file(s) with vendor compiler..."
//                << std::endl;

    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE)
      std::cout << "  Link using the linker = " << parseStatus->compilerName
                << std::endl;
    finalCombinedExitStatus = project->link(parseStatus->compilerName);
  }

  // Message from backend to user.
  if (parseStatus->verbosityLevel >= RISC_PEDANTIC)
    std::cout << "  Source file(s) compiled with vendor compiler. "
              << "(exit status = " << finalCombinedExitStatus << ")."
              << std::endl << std::endl;

  // Set the final error code to be returned to the user.
  project->set_backendErrorCode(finalCombinedExitStatus);

  //std::cout << "into risc_compile" <<std::endl;
  return project->get_backendErrorCode();
}

int
risc_backend(risc::Design &design,
    risc::tools::ParseStatus* parseStatus,
    const std::string tool_name,
    UnparseFormatHelp *unparseFormatHelp,
    UnparseDelegate* unparseDelegate)
{
  SgProject* project = design.project_;

  // RISC only supports C++/SystemC codes
  if (project->get_Cxx_only() == false)
  {
    std::cout << tool_name << ": only supports C++ at present!" << std::endl;
    ROSE_ASSERT(false);
  }

  if (project->get_binary_only() == true)
  {
    ROSE_ASSERT(project != NULL);

    if (parseStatus->verbosityLevel >= RISC_VERBOSE)
      std::cout << tool_name << ": Binary executables are unparsed, but not passed to "
                << "C++ compiler as assembly source code" << std::endl;

    project->skipfinalCompileStep(true);
  }

  if (project->get_useBackendOnly() == false)
  {
    if (parseStatus->verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << tool_name << ": unparsing source files..." << std::endl;
    }

    risc_unparse(design, parseStatus, tool_name, unparseFormatHelp, unparseDelegate);

/*** THIS PATCHING AROUND ROSE LIMITATIONS IS NOT NEEDED ANY MORE
     WITH THE NEW ROSE LIBRARY VERSION 0.9.10.X (RD, 05/30/18)
    if (parseStatus->verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << tool_name << ": patching ROSE limitations..."
                << std::endl;
    }
    risc_patch(design, parseStatus,
               "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,"
               "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_method_process "
               "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
               "\\5, \\6, \\7)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
               "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,"
               "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_thread_process "
               "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
               "\\5, \\6, \\7)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
               "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,"
               "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_cthread_process "
               "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
               "\\5, \\6, \\7)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
               tool_name);
***/
  }

  if (parseStatus->verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << tool_name << ": compiling executable..."
              << std::endl;
  }

  return risc_compile(design, parseStatus, tool_name);
}

void
instrumentor(risc::Design &design,
    risc::sg::SegmentGraph &sg,
    risc::sg::DataConflictTable* data_conflict_table,
    risc::sg::EventConflictTable* event_conflict_table,
    risc::sg::TimeAdvanceTable* time_advance_table,
    risc::sg::CombinedDataConflictTable* combined_data_conflict_table,
    risc::sg::PredictionTimeAdvanceTable* prediction_time_advance_table,
    risc::sg::PredictionEventNotificationTable* prediction_event_notification_table,
    risc::tools::ParseStatus* parseStatus,
    std::string design_file_name,
    std::vector< std::vector<int> > &psg_seg_id_array,
    std::vector<std::string> &file_names,

    std::vector< std::vector<int> > &channel_seg_id_array,
    std::vector<std::string> &channel_seg_id_array_names
    )
{
  SgProject* project = design.project_;
  ROSE_ASSERT (project != NULL);
  SgClassDeclaration* ChnlScpLockClassDecl = NULL;
  SgFunctionDeclaration* SCMainFuncDecl = NULL;
  SgType* variableType = NULL;
  SgExpression* PrimChnlVarRefExp = NULL;
  SgExpression* SCChnlVarRefExp = NULL;
  SgClassDefinition* PrimChnlClassDef = NULL;
  SgClassDefinition* SCChnlClassDef = NULL;
  SgClassDefinition* SCModuleClassDef = NULL;

  std::map<SgNode*, int> FuncCallExptoID;

  for(risc::ModuleVectorConstIter
      iter_module  = design.module_definitions_.begin();
      iter_module != design.module_definitions_.end();
      iter_module++) {

    for(risc::ThreadVectorConstIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      FuncCallExptoID[(*iter_thread)->starting_node_]
        = (*iter_thread)->starting_id_;
    }
  }

  std::pair<risc::sg::VertexIterator, risc::sg::VertexIterator>
    vertices_range = vertices(sg.graph_);

  for(risc::sg::VertexIterator
      iter  = vertices_range.first;
      iter != vertices_range.second;
      iter++) {

    if (parseStatus->verbosityLevel >= RISC_VERY_VERY_VERBOSE) {
      if (sg.graph_[*iter].boundary_node_)
        std::cout << "  Segment boundary node: "
                  << sg.graph_[*iter].boundary_node_->class_name() << " "
                  << sg.graph_[*iter].boundary_node_->unparseToString()
                  << std::endl;
    }

    
    const SgExprStatement* waitexprstmt
      = isSgExprStatement(sg.graph_[*iter].boundary_node_);

    if (waitexprstmt != NULL) {
      SgFunctionCallExp* funccallexp
        = isSgFunctionCallExp(waitexprstmt->get_expression());
      if (funccallexp != NULL) {

        //befor the tag implementation
        FuncCallExptoID[funccallexp] = sg.graph_[*iter].id_;
        //tag: psg cloning ZC step 5 // proved wrong
        //psg node with is_channel_segment_=true should not be here because
        //it will be instrumented in another way
        //befor the tag implementation
        /*if(!PSG_MODE)
          FuncCallExptoID[funccallexp] = sg.graph_[*iter].id_;
        else{ //PSG_MODE
          if(sg.graph_[*iter].is_channel_segment_ == false){
            FuncCallExptoID[funccallexp] = sg.graph_[*iter].id_;
          }
        }*/
        //end
      }
    }

    SgFunctionCallExp* boundaryexpr
      = isSgFunctionCallExp(const_cast<SgNode*>(
          sg.graph_[*iter].boundary_node_));

    if (boundaryexpr != NULL) {
      //befor the tag implementation
      FuncCallExptoID[boundaryexpr] = sg.graph_[*iter].id_;
      //tag: psg cloning ZC step 5 // proved wrong
      //psg node with is_channel_segment_=true should not be here because
      //it will be instrumented in another way
      /*if(!PSG_MODE)
        FuncCallExptoID[boundaryexpr] = sg.graph_[*iter].id_;
      else{ //PSG_MODE
        if(sg.graph_[*iter].is_channel_segment_ == false){
          FuncCallExptoID[boundaryexpr] = sg.graph_[*iter].id_;
        }
      }*/

      //end
    }
  }

  // Find the class declarations of chnl_scoped_lock, sc_prim_channel,
  // sc_channel and sc_module
  const Rose_STL_Container<SgNode*> &classDefinitionList
    = risc::tools::RoseNodes::get_class_defs();

  for (Rose_STL_Container<SgNode*>::const_iterator
      it  = classDefinitionList.begin();
      it != classDefinitionList.end();
      it++)
  {
    SgClassDefinition* classdf = isSgClassDefinition(*it);
    if (classdf != NULL)
    {
      if (classdf->get_qualified_name().getString() ==
          "::sc_core::chnl_scoped_lock")
      {
        // SgClassDeclaration of "::sc_core::chnl_scoped_lock"
        ChnlScpLockClassDecl = classdf->get_declaration();
      }

      if (classdf->get_qualified_name().getString() ==
          "::sc_core::sc_prim_channel")
      {
        // SgClassDefinition of "::sc_core::sc_prim_channel"
        PrimChnlClassDef = classdf;
      }

      if (classdf->get_qualified_name().getString() ==
          "::sc_core::sc_channel")
      {
        // SgClassDefinition of "::sc_core::sc_channel"
        SCChnlClassDef = classdf;
      }

      if (classdf->get_qualified_name().getString() ==
          "::sc_core::sc_module")
      {
        // SgClassDefinition of "::sc_core::sc_module"
        SCModuleClassDef = classdf;
      }

      if (ChnlScpLockClassDecl && PrimChnlClassDef &&
          SCChnlClassDef && SCModuleClassDef) {
        break;
      }
    }
  }
  ROSE_ASSERT(ChnlScpLockClassDecl != NULL);
  ROSE_ASSERT(PrimChnlClassDef != NULL);
  ROSE_ASSERT(SCChnlClassDef != NULL);
  ROSE_ASSERT(SCModuleClassDef != NULL);

  variableType = ChnlScpLockClassDecl->get_type();
  ROSE_ASSERT(variableType != NULL);

  PrimChnlVarRefExp = SageBuilder::buildVarRefExp("m_mutex", PrimChnlClassDef);
  ROSE_ASSERT(PrimChnlVarRefExp != NULL);

  SCChnlVarRefExp = SageBuilder::buildVarRefExp("m_mutex", SCChnlClassDef);
  ROSE_ASSERT(SCChnlVarRefExp != NULL);

  // Find the function declaration of sc_main
  if(!PSG_MODE){
    SCMainFuncDecl = risc::tools::SystemCDefinitions::get_sc_main()->get_declaration();

    ROSE_ASSERT(SCMainFuncDecl != NULL);
  }

  // Build the traversal object
  InstrumentationTraversal traversal(variableType, PrimChnlVarRefExp,
                                     SCChnlVarRefExp, SCModuleClassDef,
                                     FuncCallExptoID, design_file_name);

  // Call the traversal starting at the project node of the AST
  traversal.traverseInputFiles(project, preorder);

  // Instrument all annotated function declarations in interfaces
  std::list<SgClassDefinition*> annotatedIntfList
    = risc::tools::get_all_used_interfaces();
  for (std::list<SgClassDefinition*>::iterator
       iter  = annotatedIntfList.begin();
       iter != annotatedIntfList.end();
       iter++) {
    std::vector<SgNode*> memFuncDeclList
      = NodeQuery::querySubTree(*iter, V_SgMemberFunctionDeclaration);

    for (std::vector<SgNode*>::iterator
         memFuncDeclIter  = memFuncDeclList.begin();
         memFuncDeclIter != memFuncDeclList.end();
         memFuncDeclIter++) {
      SgMemberFunctionDeclaration* memFuncDecl
        = isSgMemberFunctionDeclaration(*memFuncDeclIter);

      if (risc::tools::is_annotated_function(memFuncDecl)) {
        SgFunctionParameterList* newFuncParam
          = SageBuilder::buildFunctionParameterList();
        for (SgInitializedNamePtrList::iterator funcParamIter
               = memFuncDecl->get_parameterList()->get_args().begin();
             funcParamIter
               != memFuncDecl->get_parameterList()->get_args().end();
             funcParamIter++) {
          SgInitializedName* param
            = SageBuilder::buildInitializedName(
                (*funcParamIter)->get_name().getString(),
                (*funcParamIter)->get_type());
          SageInterface::appendArg(newFuncParam, param);
        }

        SgInitializedName* segId
          = SageBuilder::buildInitializedName("_seg_id",
                                              SageBuilder::buildIntType());
        SageInterface::appendArg(newFuncParam, segId);

        SgMemberFunctionDeclaration* newMemFuncDecl
          = SageBuilder::buildNondefiningMemberFunctionDeclaration(
            SageInterface::get_name(memFuncDecl),
            memFuncDecl->get_orig_return_type(),
            newFuncParam,
            memFuncDecl->get_class_scope());

        if (memFuncDecl->get_functionModifier().isVirtual()) {
          newMemFuncDecl->get_functionModifier().setVirtual();
        }

        if (memFuncDecl->get_functionModifier().isPureVirtual()) {
          newMemFuncDecl->get_functionModifier().setPureVirtual();
        }

        if (memFuncDecl->get_declarationModifier().get_accessModifier().
            isPublic()) {
          newMemFuncDecl->get_declarationModifier().get_accessModifier().
            setPublic();
        }

        if (memFuncDecl->get_declarationModifier().get_accessModifier().
            isProtected()) {
          newMemFuncDecl->get_declarationModifier().get_accessModifier().
            setProtected();
        }

        if (memFuncDecl->get_declarationModifier().get_accessModifier().
            isPrivate()) {
          newMemFuncDecl->get_declarationModifier().get_accessModifier().
            setPrivate();
        }

        SageInterface::insertStatement(memFuncDecl, newMemFuncDecl, false);
        SageInterface::removeStatement(memFuncDecl);
      }
    }
  }

  // Instrument all annotated function declarations and definitions in channels
  std::list<SgFunctionDeclaration*> annotatedFuncDecList
    = risc::tools::get_all_implemented_interface_functions(true, true);
  for (std::list<SgFunctionDeclaration*>::iterator
       iter  = annotatedFuncDecList.begin();
       iter != annotatedFuncDecList.end();
       iter++) {
    SgFunctionDefinition* funcdf
      = isSgFunctionDefinition((*iter)->get_definition());

    if (funcdf != NULL) {
      SgInitializedName* segid
        = SageBuilder::buildInitializedName("_seg_id",
                                            SageBuilder::buildIntType());
      segid->set_scope(funcdf);
      funcdf->insert_symbol(segid->get_name(), new SgVariableSymbol(segid));
      (*iter)->append_arg(segid);

      SgBasicBlock* block = funcdf->get_body();
      if (block != NULL) {
        Rose_STL_Container<SgNode*> funcCallExpList
          = NodeQuery::querySubTree(block, V_SgFunctionCallExp);
        for (Rose_STL_Container<SgNode*>::iterator
             it = funcCallExpList.begin();
             it != funcCallExpList.end(); it++)
        {
          SgFunctionCallExp* funccallexp = isSgFunctionCallExp(*it);
          if (funccallexp != NULL) {
            SgFunctionDeclaration* funcdecl
              = funccallexp->getAssociatedFunctionDeclaration();

            if (funcdecl != NULL) {
              if (funcdecl->get_name().getString() == "wait") {
                funccallexp->append_arg(SageBuilder::buildVarRefExp(segid,
                                                                    funcdf));
              }
            }
          }
        }
      }
    }
  }
  
  if(PSG_MODE){
    SgVariableDeclaration* _psg_off 
      = SageBuilder::buildVariableDeclaration(
          "psg_seg_id_"+design_file_name,
          SageBuilder::buildArrayType(SageBuilder::buildUnsignedIntType()));
    SageInterface::setExtern (_psg_off);

    if(traversal.first_func_decl_that_contains_wait!=NULL){
      SageInterface::insertStatementBefore(traversal.first_func_decl_that_contains_wait,
                                         _psg_off); 
    }
    std::vector<SgVariableDeclaration* > port_call_vars;
    for(std::set<risc::sg::VertexDescriptor>::iterator iter = sg.port_call_seg_vertexs_.begin();
        iter != sg.port_call_seg_vertexs_.end();
        iter ++)
    {
      SgVariableDeclaration* var 
      = SageBuilder::buildVariableDeclaration(
          "port_call_seg_ids_"+
            sg.graph_[*iter].unique_port_call_name_
          ,
          SageBuilder::buildArrayType(SageBuilder::buildIntType()));
      SageInterface::setExtern (var);
      if(traversal.first_func_decl_that_contains_wait!=NULL){
        SageInterface::insertStatementBefore(traversal.first_func_decl_that_contains_wait,
                                         var); 
      }


      //set_upcoming_segment_id

      // sc_core::sc_simcontext *temp = sc_get_curr_simcontext();
      SgFunctionCallExp *get_sim_context_func_call
        = SageBuilder::buildFunctionCallExp(
            isSgFunctionSymbol(
              risc::tools::SystemCDefinitions::get_sc_get_curr_simcontext()
                ->get_symbol_from_symbol_table()));
      SgAssignInitializer *init_sim_context
        = SageBuilder::buildAssignInitializer(get_sim_context_func_call);

      SgVariableDeclaration *sc_sim_context
        = SageBuilder::buildVariableDeclaration(
            generate_unique_name("sim_context"),
            SageBuilder::buildPointerType(
              risc::tools::SystemCDefinitions::get_sc_simcontext()
                ->get_declaration()->get_type()),
            init_sim_context);

      SageInterface::insertStatement(
          SageInterface::getEnclosingStatement(sg.graph_[*iter].port_call_func_exp_),
          sc_sim_context);


      // sc_core::sc_process_b *temp2 = temp->get_curr_proc();
      SgArrowExp *arrow_get_curr_proc
        = SageBuilder::buildArrowExp(
            SageBuilder::buildVarRefExp(sc_sim_context),
            SageBuilder::buildMemberFunctionRefExp(
              isSgMemberFunctionSymbol(
                risc::tools::SystemCDefinitions::get_sc_curr_proc()
                  ->get_symbol_from_symbol_table()),
              true,
              false));

      SgFunctionCallExp *get_curr_proc
        = SageBuilder::buildFunctionCallExp(arrow_get_curr_proc);

      SgAssignInitializer *init_process_b
        = SageBuilder::buildAssignInitializer(get_curr_proc);

      SgVariableDeclaration *sc_process_b
        = SageBuilder::buildVariableDeclaration(
            generate_unique_name("process_b"),
            SageBuilder::buildPointerType(
              risc::tools::SystemCDefinitions::get_sc_process_b()
                ->get_declaration()->get_type()),
            init_process_b);

      SageInterface::insertStatement(
          SageInterface::getEnclosingStatement(sg.graph_[*iter].port_call_func_exp_),
          sc_process_b);


      // temp2->set_upcoming_segment_ids(array);
      SgArrowExp *arrow_set_upcoming_ids
        = SageBuilder::buildArrowExp(
            SageBuilder::buildVarRefExp(sc_process_b),
            SageBuilder::buildMemberFunctionRefExp(
              isSgMemberFunctionSymbol(
                risc::tools::SystemCDefinitions::get_set_upcoming_segment_ids()
                  ->get_symbol_from_symbol_table()),
              true,
              false));

      std::vector<SgExpression*> params_set_ids;
      params_set_ids.push_back(SageBuilder::buildVarRefExp(var));

      SgExprListExp *expr_list_seg_ids
        = SageBuilder::buildExprListExp(params_set_ids);

      SgFunctionCallExp *set_upcoming_ids
        = SageBuilder::buildFunctionCallExp(
            arrow_set_upcoming_ids, expr_list_seg_ids);

      SgExprStatement *upcoming_ids_statement
        = SageBuilder::buildExprStatement(set_upcoming_ids);

      SageInterface::insertStatement(
          SageInterface::getEnclosingStatement(sg.graph_[*iter].port_call_func_exp_),
          upcoming_ids_statement);
      //end
    }
    instrument_wait_in_channels(sg);
  }
  else{
    int file_names_id = 0;
    for(std::vector<std::string>::iterator file_iter = file_names.begin();
      file_iter != file_names.end();
      file_iter ++)
    {
      file_names_id++;
      int psg_seg_id1_size = psg_seg_id_array[file_names_id].size();
      std::vector<SgExpression *> psg_seg_id1_value(psg_seg_id1_size);
    
      for (int i = 0; i < psg_seg_id1_size; i++) {        
        psg_seg_id1_value.at(i)
          = SageBuilder::buildIntVal(psg_seg_id_array[file_names_id][i]);      
      }

      SgVariableDeclaration* psg_seg_id1_array
        = SageBuilder::buildVariableDeclaration(
            "psg_seg_id_"+(*file_iter),
            SageBuilder::buildArrayType(SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildIntVal(psg_seg_id1_size)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              psg_seg_id1_value)));
                
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                         psg_seg_id1_array);      
    }

    int names_id = 0;
    for(std::vector<std::string>::iterator name_iter = channel_seg_id_array_names.begin();
      name_iter != channel_seg_id_array_names.end();
      name_iter ++)
    {
      
      int psg_seg_id1_size = channel_seg_id_array[names_id].size();
      std::vector<SgExpression *> psg_seg_id1_value(psg_seg_id1_size);
    
      for (int i = 0; i < psg_seg_id1_size; i++) {        
        psg_seg_id1_value.at(i)
          = SageBuilder::buildIntVal(channel_seg_id_array[names_id][i]);      
      }

      SgVariableDeclaration* psg_seg_id1_array
        = SageBuilder::buildVariableDeclaration(
            "port_call_seg_ids_"+(*name_iter),
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
            SageBuilder::buildIntVal(psg_seg_id1_size)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              psg_seg_id1_value)));
                
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                         psg_seg_id1_array); 

      names_id ++;     
    }


    
    // Output the data conflict table
    if (data_conflict_table) {
      int DataConflTableSize = data_conflict_table->size_of_conflict_table_;
      std::vector<SgExpression *> DataConflTableValue(DataConflTableSize * DataConflTableSize);
      int counter = 0;
      for (int i = 0; i < DataConflTableSize; i++) {
        for (int j = 0; j < DataConflTableSize; j++) {
          DataConflTableValue.at(counter)
            = SageBuilder::buildIntVal(data_conflict_table->conflict_table_[i][j]);
          counter++;
        }
      }

      SgVariableDeclaration* DataConflTableDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Data_Conflict_Table",
            SageBuilder::buildArrayType(SageBuilder::buildBoolType(),
            SageBuilder::buildIntVal(DataConflTableSize * DataConflTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              DataConflTableValue)));

      SageInterface::insertStatementBefore(SCMainFuncDecl, DataConflTableDecl);

      // Output the data conflict table size
      SgVariableDeclaration* DataConflTableSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Data_Conflict_Table_Size",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              DataConflTableSize)));

      SageInterface::insertStatementBefore(SCMainFuncDecl, DataConflTableSizeDecl);
    }

    // Output the event notification table
    if(event_conflict_table) {
      int EventNotifyTableSize = event_conflict_table->size_of_conflict_table_;
      std::vector<SgExpression *> EventNotifyTableValue(EventNotifyTableSize * EventNotifyTableSize);
      int counter = 0;
      for (int i = 0; i < EventNotifyTableSize; i++) {
        for (int j = 0; j < EventNotifyTableSize; j++) {
          EventNotifyTableValue.at(counter)
            = SageBuilder::buildIntVal(event_conflict_table->conflict_table_[i][j]);
          counter++;
        }
      }

      SgVariableDeclaration* EventNotifyTableDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Event_Notify_Table",
            SageBuilder::buildArrayType(SageBuilder::buildBoolType(),
            SageBuilder::buildIntVal(EventNotifyTableSize * EventNotifyTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              EventNotifyTableValue)));

      SageInterface::insertStatementBefore(SCMainFuncDecl, EventNotifyTableDecl);

      // Output the event notification table size
      SgVariableDeclaration* EventNotifyTableSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Event_Notify_Table_Size",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              EventNotifyTableSize)));

      SageInterface::insertStatementBefore(SCMainFuncDecl, EventNotifyTableSizeDecl);
    }

    // Output the segment and instance id to index lookup table for conflict tables
    int MaxNumofInst = data_conflict_table->max_instances_;
    int NumofSeg = data_conflict_table->number_of_segments_;
    std::vector<SgExpression *> ConflictLookupTableValue;
    for (int i = 1; i < MaxNumofInst + 1; i++) { //ZC
      for (int j = 0; j < NumofSeg; j++)
      {
        // TODO: Temporal patch: all unanalyzed segment boundaries are given ID 0
        if (j == 0 &&
            data_conflict_table->instance_id_to_index_lookup_[i][j] == -1) {

          ConflictLookupTableValue.push_back(SageBuilder::buildIntVal(0));
        } else {

          ConflictLookupTableValue.push_back(SageBuilder::buildIntVal(
            data_conflict_table->instance_id_to_index_lookup_[i][j]));
        }
      }
    }

    SgVariableDeclaration* ConflictLookupTableDecl
      = SageBuilder::buildVariableDeclaration(
          "sc_core::_OoO_Conflict_Index_Lookup_Table",
          SageBuilder::buildArrayType(SageBuilder::buildIntType(),
          SageBuilder::buildIntVal((MaxNumofInst + 1) * NumofSeg)),
          SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
            ConflictLookupTableValue)));

    SageInterface::insertStatementBefore(SCMainFuncDecl, ConflictLookupTableDecl);

    // Output the maximum number of module instances and number of segments
    SgVariableDeclaration* MaxNumofInstDecl
      = SageBuilder::buildVariableDeclaration(
          "sc_core::_OoO_Max_Number_of_Instances",
          SageBuilder::buildUnsignedIntType(),
          SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
            MaxNumofInst)));

    SageInterface::insertStatementBefore(SCMainFuncDecl, MaxNumofInstDecl);

    SgVariableDeclaration* NumofSegDecl
      = SageBuilder::buildVariableDeclaration(
          "sc_core::_OoO_Number_of_Segments",
          SageBuilder::buildUnsignedIntType(),
          SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
            NumofSeg)));

    SageInterface::insertStatementBefore(SCMainFuncDecl, NumofSegDecl);

    // Output the current time advance table
    int CurrTimeAdvTableSize = time_advance_table->size_;

    std::vector<SgExpression *> CurrTimeAdvTableTimeValue(CurrTimeAdvTableSize);
    int counter = 0;
    for (int i = 0; i < CurrTimeAdvTableSize; i++)
    {
      if (time_advance_table->current_time_advance_[i].infinite_) {

        // -1 stands for infinity
        CurrTimeAdvTableTimeValue.at(counter)
          = SageBuilder::buildLongLongIntVal(-1);
      } else {
        CurrTimeAdvTableTimeValue.at(counter)
          = SageBuilder::buildLongLongIntVal(
              time_advance_table->current_time_advance_[i].time_units_);
      }
      counter++;
    }

    SgVariableDeclaration* CurrTimeAdvTableTimeDecl
      = SageBuilder::buildVariableDeclaration(
          "sc_core::_OoO_Curr_Time_Advance_Table_Time",
          SageBuilder::buildArrayType(SageBuilder::buildLongLongType(),
                                      SageBuilder::buildIntVal(
                                        CurrTimeAdvTableSize)),
          SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
            CurrTimeAdvTableTimeValue)));

    std::vector<SgExpression *> CurrTimeAdvTableDeltaValue(CurrTimeAdvTableSize);
    counter = 0;
    for (int i = 0; i < CurrTimeAdvTableSize; i++)
    {
      if (time_advance_table->current_time_advance_[i].infinite_) {
        // -1 stands for infinity
        CurrTimeAdvTableDeltaValue.at(counter) = SageBuilder::buildIntVal(-1);
      } else {
        CurrTimeAdvTableDeltaValue.at(counter)
          = SageBuilder::buildIntVal(
              time_advance_table->current_time_advance_[i].delta_count_);
      }
      counter++;
    }

    SgVariableDeclaration* CurrTimeAdvTableDeltaDecl
      = SageBuilder::buildVariableDeclaration(
          "sc_core::_OoO_Curr_Time_Advance_Table_Delta",
          SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                      SageBuilder::buildIntVal(
                                        CurrTimeAdvTableSize)),
          SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
            CurrTimeAdvTableDeltaValue)));

    SageInterface::insertStatementBefore(SCMainFuncDecl,
                                         CurrTimeAdvTableTimeDecl);
    SageInterface::insertStatementBefore(SCMainFuncDecl,
                                         CurrTimeAdvTableDeltaDecl);

    // Output the current time advance table size
    SgVariableDeclaration* CurrTimeAdvTableSizeDecl
      = SageBuilder::buildVariableDeclaration(
          "sc_core::_OoO_Curr_Time_Advance_Table_Size",
          SageBuilder::buildUnsignedIntType(),
          SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
            CurrTimeAdvTableSize)));

    SageInterface::insertStatementBefore(SCMainFuncDecl,
                                         CurrTimeAdvTableSizeDecl);

    // Output the next time advance table
    if(time_advance_table) {
      int NextTimeAdvTableSize = time_advance_table->size_;

      std::vector<SgExpression *> NextTimeAdvTableTimeValue(NextTimeAdvTableSize);
      int counter = 0;
      for (int i = 0; i < NextTimeAdvTableSize; i++)
      {
        if (time_advance_table->next_time_advance_[i].infinite_) {
          // -1 stands for infinity
          NextTimeAdvTableTimeValue.at(counter)
            = SageBuilder::buildLongLongIntVal(-1);
        } else {
          NextTimeAdvTableTimeValue.at(counter)
            = SageBuilder::buildLongLongIntVal(
                time_advance_table->next_time_advance_[i].time_units_);
        }
        counter++;
      }

      SgVariableDeclaration* NextTimeAdvTableTimeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Next_Time_Advance_Table_Time",
            SageBuilder::buildArrayType(SageBuilder::buildLongLongType(),
                                        SageBuilder::buildIntVal(
                                          NextTimeAdvTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              NextTimeAdvTableTimeValue)));

      std::vector<SgExpression *> NextTimeAdvTableDeltaValue(NextTimeAdvTableSize);
      counter = 0;
      for (int i = 0; i < NextTimeAdvTableSize; i++)
      {
        if (time_advance_table->next_time_advance_[i].infinite_) {
          // -1 stands for infinity
          NextTimeAdvTableDeltaValue.at(counter) = SageBuilder::buildIntVal(-1);
        } else {
          NextTimeAdvTableDeltaValue.at(counter)
            = SageBuilder::buildIntVal(
                time_advance_table->next_time_advance_[i].delta_count_);
        }
        counter++;
      }

      SgVariableDeclaration* NextTimeAdvTableDeltaDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Next_Time_Advance_Table_Delta",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(
                                          NextTimeAdvTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              NextTimeAdvTableDeltaValue)));

      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           NextTimeAdvTableTimeDecl);
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           NextTimeAdvTableDeltaDecl);
    }

    // Output the next time advance table size
    if(time_advance_table) {
      int NextTimeAdvTableSize = time_advance_table->size_;

      SgVariableDeclaration* NextTimeAdvTableSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Next_Time_Advance_Table_Size",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              NextTimeAdvTableSize)));

      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           NextTimeAdvTableSizeDecl);

      // Output the segment id to index lookup table for time advance tables
      int TimeAdvLookupTableSize = time_advance_table->size_;
      std::vector<SgExpression *> TimeAdvLookupTableValue(TimeAdvLookupTableSize);
      int counter = 0;
      for (int i = 0; i < TimeAdvLookupTableSize; i++) {
        TimeAdvLookupTableValue.at(counter)
          = SageBuilder::buildIntVal(time_advance_table->get_segment_id(i));
        counter++;
      }

      SgVariableDeclaration* TimeAdvLookupTableDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Time_Advance_Index_Lookup_Table",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(
                                          TimeAdvLookupTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              TimeAdvLookupTableValue)));

      SageInterface::insertStatementBefore(SCMainFuncDecl, TimeAdvLookupTableDecl);

      // Output the size of the segment id to index lookup table
      SgVariableDeclaration* TimeAdvLookupTableSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Time_Advance_Index_Lookup_Table_Size",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              TimeAdvLookupTableSize)));

      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           TimeAdvLookupTableSizeDecl);
    }


    if(combined_data_conflict_table) {

      int CombinedDataConflictTableSize
        = combined_data_conflict_table->size_of_conflict_table_;

      // preparing array
      std::vector<SgExpression *> CombinedDataConflictTableSizeValue(
          CombinedDataConflictTableSize * CombinedDataConflictTableSize);

      int counter = 0;
      for(int i = 0; i < CombinedDataConflictTableSize; i++) {
        for(int j = 0; j < CombinedDataConflictTableSize; j++) {
          CombinedDataConflictTableSizeValue.at(counter)
            = SageBuilder::buildIntVal(
                combined_data_conflict_table->combined_table_[i * CombinedDataConflictTableSize + j]);
          counter++;
        }
      }

      // preparing index loopup array
      std::vector<SgExpression *> CombinedDataConflictTableLookup;
      for(int i = 1; i < combined_data_conflict_table->max_instances_ + 1; i++) { //ZC
        for(int j = 0; j < combined_data_conflict_table->number_of_segments_; j++) {

          CombinedDataConflictTableLookup.push_back(
              SageBuilder::buildIntVal(combined_data_conflict_table->instance_id_to_index_lookup_[i][j]));
        }
      }

      // table
      SgVariableDeclaration* CombinedDataConflictTable
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Combined_Data_Conflict_Table",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(CombinedDataConflictTableSize * CombinedDataConflictTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              CombinedDataConflictTableSizeValue)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, CombinedDataConflictTable);

      // table size
      SgVariableDeclaration* TimeAdvLookupTableSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Combined_Data_Conflict_Table_Size",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              CombinedDataConflictTableSize * CombinedDataConflictTableSize)));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           TimeAdvLookupTableSizeDecl);

      // index lookup
      SgVariableDeclaration* CombinedDataConflictTableLookupDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Combined_Data_Conflict_Lookup_Table",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(combined_data_conflict_table->number_of_segments_ * (combined_data_conflict_table->max_instances_ + 1))),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              CombinedDataConflictTableLookup)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, CombinedDataConflictTableLookupDecl);

      // number segments
      SgVariableDeclaration* CombinedDataConflictTableNumberSegmentsDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Combined_Data_Conflict_Lookup_Table_Number_Segments",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              combined_data_conflict_table->number_of_segments_)));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           CombinedDataConflictTableNumberSegmentsDecl);

      // max number of instances
      SgVariableDeclaration* CombinedDataConflictTablesMaxInstancesDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Combined_Data_Conflict_Lookup_Table_Max_Instances",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              combined_data_conflict_table->max_instances_)));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           CombinedDataConflictTablesMaxInstancesDecl);
    }

    if(prediction_time_advance_table) {
      int PredictionTimeAdvanceTableSize
        = prediction_time_advance_table->table_.size();

      std::vector<SgExpression *> PredictionTimeAdvanceTableTimeUnits(
          PredictionTimeAdvanceTableSize * (prediction_time_advance_table->prediction_steps_ + 1));

      std::vector<SgExpression *> PredictionTimeAdvanceTableTimeDelta(
          PredictionTimeAdvanceTableSize * (prediction_time_advance_table->prediction_steps_ + 1));

      int counter = 0;
      for(int i = 0; i < PredictionTimeAdvanceTableSize; i++) {
        for(int j = 0; j < prediction_time_advance_table->prediction_steps_ + 1; j++) {

          int delta_val;
          long long int unit_val;

          if(prediction_time_advance_table->table_[i][j].infinite_) {
            delta_val = -1;
            unit_val = -1;
          } else {
            delta_val = prediction_time_advance_table->table_[i][j].delta_count_;
            unit_val = prediction_time_advance_table->table_[i][j].time_units_;
          }

          PredictionTimeAdvanceTableTimeUnits.at(counter)
            = SageBuilder::buildLongLongIntVal(unit_val);
          PredictionTimeAdvanceTableTimeDelta.at(counter)
            = SageBuilder::buildIntVal(delta_val);

          counter++;
        }
      }

      // preparing index loopup array
      std::vector<SgExpression *> PredictionTimeAdvanceTableLookup(
          prediction_time_advance_table->index_to_segment_.size());
      counter = 0;
      for(int i = 0; i < prediction_time_advance_table->index_to_segment_.size(); i++) {
        PredictionTimeAdvanceTableLookup.at(counter)
          = SageBuilder::buildIntVal(prediction_time_advance_table->index_to_segment_[i]);
        counter++;
      }

      // time units array
      SgVariableDeclaration* PredictionTimeAdvanceTableTimeUnitsDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Time_Advance_Table_Time_Units",
            SageBuilder::buildArrayType(SageBuilder::buildLongLongType(),
                                        SageBuilder::buildLongLongIntVal(PredictionTimeAdvanceTableSize * (prediction_time_advance_table->prediction_steps_ + 1))),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              PredictionTimeAdvanceTableTimeUnits)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, PredictionTimeAdvanceTableTimeUnitsDecl);

      // delta units array
      SgVariableDeclaration* PredictionTimeAdvanceTableTimeDeltaDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Time_Advance_Table_Delta",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(PredictionTimeAdvanceTableSize * (prediction_time_advance_table->prediction_steps_ + 1))),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              PredictionTimeAdvanceTableTimeDelta)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, PredictionTimeAdvanceTableTimeDeltaDecl);

      // number of prediction steps
      SgVariableDeclaration* PredictionTimeAdvanceTabeTimeStepsDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Time_Advance_Table_Number_Steps",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              prediction_time_advance_table->prediction_steps_)));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           PredictionTimeAdvanceTabeTimeStepsDecl);

      // number of segments
      SgVariableDeclaration* PredictionTimeAdvanceTabeTimeSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Time_Advance_Table_Number_Segments",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              prediction_time_advance_table->index_to_segment_.size())));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           PredictionTimeAdvanceTabeTimeSizeDecl);

      // index lookup
      SgVariableDeclaration* PredictionTimeAdvanceTabeTimeLookupDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Time_Advance_Lookup_Table",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(prediction_time_advance_table->index_to_segment_.size())),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              PredictionTimeAdvanceTableLookup)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, PredictionTimeAdvanceTabeTimeLookupDecl);
    }

    if(prediction_event_notification_table) {

      int PredictionEventNotificationTableSize
        = prediction_event_notification_table->size_of_conflict_table_;

      std::vector<SgExpression *> PredictionEventNotificationTableTimeUnits(
          PredictionEventNotificationTableSize * PredictionEventNotificationTableSize);
      std::vector<SgExpression *> PredictionEventNotificationTableTableTimeDelta(
          PredictionEventNotificationTableSize * PredictionEventNotificationTableSize);
      counter = 0;
      for(int i = 0; i < PredictionEventNotificationTableSize; i++) {
        for(int j = 0; j < PredictionEventNotificationTableSize; j++) {

          int delta_val;
          long long int unit_val;

          if(prediction_event_notification_table->combined_table_[i][j].infinite_) {
            delta_val = -1;
            unit_val = -1;
          } else {
            delta_val = prediction_event_notification_table->combined_table_[i][j].delta_count_;
            unit_val = prediction_event_notification_table->combined_table_[i][j].time_units_;
          }

          PredictionEventNotificationTableTimeUnits.at(counter)
            = SageBuilder::buildLongLongIntVal(unit_val);
          PredictionEventNotificationTableTableTimeDelta.at(counter)
            = (SageBuilder::buildIntVal(delta_val));

          counter++;
        }
      }

      // preparing index loopup array
      std::vector<SgExpression *> PredictionEventNotificationLookupTable;

      for(int i = 1; i < prediction_event_notification_table->max_instances_ + 1; i++) { //ZC
        for(int j = 0; j < prediction_event_notification_table->number_of_segments_; j++) {

          PredictionEventNotificationLookupTable.push_back(
              SageBuilder::buildIntVal(prediction_event_notification_table->instance_id_to_index_lookup_[i][j]));
        }
      }

      // time units array
      SgVariableDeclaration* PredictionEventNotificationTableTimeUnitsDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Event_Notification_Table_Time_Units",
            SageBuilder::buildArrayType(SageBuilder::buildLongLongType(),
                                        SageBuilder::buildLongLongIntVal(PredictionEventNotificationTableSize * PredictionEventNotificationTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              PredictionEventNotificationTableTimeUnits)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, PredictionEventNotificationTableTimeUnitsDecl);

      // delta units array
      SgVariableDeclaration* PredictionEventNotificationTableTimeDeltaDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Event_Notification_Table_Delta",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal(PredictionEventNotificationTableSize * PredictionEventNotificationTableSize)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              PredictionEventNotificationTableTableTimeDelta)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, PredictionEventNotificationTableTimeDeltaDecl);

      // number of segments
      SgVariableDeclaration* PredictionEventNotificationTabeTimeSizeDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Event_Notification_Table_Number_Segments",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              prediction_event_notification_table->number_of_segments_)));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           PredictionEventNotificationTabeTimeSizeDecl);

      // index lookup
      SgVariableDeclaration* PredictionEventNotificationTabeTimeLookupDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Event_Notification_Lookup_Table",
            SageBuilder::buildArrayType(SageBuilder::buildIntType(),
                                        SageBuilder::buildIntVal((prediction_event_notification_table->max_instances_ + 1)
                                          * prediction_event_notification_table->number_of_segments_)),
            SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(
              PredictionEventNotificationLookupTable)));
      SageInterface::insertStatementBefore(SCMainFuncDecl, PredictionEventNotificationTabeTimeLookupDecl);

      // max number of instances
      SgVariableDeclaration* PredictionEventNotificationTableMaxInstancesDecl
        = SageBuilder::buildVariableDeclaration(
            "sc_core::_OoO_Prediction_Event_Notification_Table_Max_Instances",
            SageBuilder::buildUnsignedIntType(),
            SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(
              prediction_event_notification_table->max_instances_)));
      SageInterface::insertStatementBefore(SCMainFuncDecl,
                                           PredictionEventNotificationTableMaxInstancesDecl);
    }

    // Instrumenting segment passing before channel call
    instrument_port_calls(sg);

    // Instrumenting segment reading at wait statement
    instrument_wait_in_channels(sg);
    // Test the new AST
    //AstTests::runAllTests(project);
  }
}



void instrument_port_calls(risc::sg::SegmentGraph &sg)
{
  for(std::set<SgFunctionCallExp*>::iterator
      port_call_iter  = sg.port_calls_.begin();
      port_call_iter != sg.port_calls_.end();
      port_call_iter++) {

    // here we build this
    // int array[3] = {5, 6, 7};
    // sc_core::sc_simcontext *temp = sc_get_curr_simcontext();
    // sc_core::sc_process_b *temp2 = temp->get_curr_proc();
    // temp2->set_upcoming_segment_ids(array);

    risc::sg::PortCallAttribute *port_call_attribute
      = dynamic_cast<risc::sg::PortCallAttribute*>(
          (*port_call_iter)->getAttribute("port_call"));

    // build the array with the variable declarations
    std::vector<SgExpression*> segment_id_exprs;
    if(port_call_attribute) {

      for(std::list<risc::sg::Segment*>::iterator
          iter_id  = port_call_attribute->segment_ids_.begin();
          iter_id != port_call_attribute->segment_ids_.end();
          iter_id++) {

        segment_id_exprs.push_back(SageBuilder::buildIntVal((*iter_id)->id_));
      }
    } else {
      assert(NULL);
    }

    SgExprListExp *expr_list = SageBuilder::buildExprListExp(segment_id_exprs);

    SgAggregateInitializer *initializer
      = SageBuilder::buildAggregateInitializer(expr_list);

    SgArrayType *type
      = SageBuilder::buildArrayType(
          SageBuilder::buildIntType(),
          SageBuilder::buildIntVal(segment_id_exprs.size()));

    SgVariableDeclaration *channel_seg_ids
      = SageBuilder::buildVariableDeclaration(
          generate_unique_name("channel_seg_ids_"),
          type,
          initializer);

    SageInterface::insertStatement(
        SageInterface::getEnclosingStatement(*port_call_iter),
        channel_seg_ids);

    // sc_core::sc_simcontext *temp = sc_get_curr_simcontext();
    SgFunctionCallExp *get_sim_context_func_call
      = SageBuilder::buildFunctionCallExp(
          isSgFunctionSymbol(
            risc::tools::SystemCDefinitions::get_sc_get_curr_simcontext()
              ->get_symbol_from_symbol_table()));
    SgAssignInitializer *init_sim_context
      = SageBuilder::buildAssignInitializer(get_sim_context_func_call);

    SgVariableDeclaration *sc_sim_context
      = SageBuilder::buildVariableDeclaration(
          generate_unique_name("sim_context"),
          SageBuilder::buildPointerType(
            risc::tools::SystemCDefinitions::get_sc_simcontext()
              ->get_declaration()->get_type()),
          init_sim_context);

    SageInterface::insertStatement(
        SageInterface::getEnclosingStatement(*port_call_iter),
        sc_sim_context);


    // sc_core::sc_process_b *temp2 = temp->get_curr_proc();
    SgArrowExp *arrow_get_curr_proc
      = SageBuilder::buildArrowExp(
          SageBuilder::buildVarRefExp(sc_sim_context),
          SageBuilder::buildMemberFunctionRefExp(
            isSgMemberFunctionSymbol(
              risc::tools::SystemCDefinitions::get_sc_curr_proc()
                ->get_symbol_from_symbol_table()),
            true,
            false));

    SgFunctionCallExp *get_curr_proc
      = SageBuilder::buildFunctionCallExp(arrow_get_curr_proc);

    SgAssignInitializer *init_process_b
      = SageBuilder::buildAssignInitializer(get_curr_proc);

    SgVariableDeclaration *sc_process_b
      = SageBuilder::buildVariableDeclaration(
          generate_unique_name("process_b"),
          SageBuilder::buildPointerType(
            risc::tools::SystemCDefinitions::get_sc_process_b()
              ->get_declaration()->get_type()),
          init_process_b);

    SageInterface::insertStatement(
        SageInterface::getEnclosingStatement(*port_call_iter),
        sc_process_b);


    // temp2->set_upcoming_segment_ids(array);
    SgArrowExp *arrow_set_upcoming_ids
      = SageBuilder::buildArrowExp(
          SageBuilder::buildVarRefExp(sc_process_b),
          SageBuilder::buildMemberFunctionRefExp(
            isSgMemberFunctionSymbol(
              risc::tools::SystemCDefinitions::get_set_upcoming_segment_ids()
                ->get_symbol_from_symbol_table()),
            true,
            false));

    std::vector<SgExpression*> params_set_ids;
    params_set_ids.push_back(SageBuilder::buildVarRefExp(channel_seg_ids));

    SgExprListExp *expr_list_seg_ids
      = SageBuilder::buildExprListExp(params_set_ids);

    SgFunctionCallExp *set_upcoming_ids
      = SageBuilder::buildFunctionCallExp(
          arrow_set_upcoming_ids, expr_list_seg_ids);

    SgExprStatement *upcoming_ids_statement
      = SageBuilder::buildExprStatement(set_upcoming_ids);

    SageInterface::insertStatement(
        SageInterface::getEnclosingStatement(*port_call_iter),
        upcoming_ids_statement);
  }
}

void instrument_wait_in_channels(risc::sg::SegmentGraph &sg)
{
  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator vertex_iter;
  std::pair<vertex_iter, vertex_iter> vp;
  for(vp = boost::vertices(sg.graph_); vp.first != vp.second; ++vp.first) {


    if(!sg.graph_[*vp.first].is_channel_segment_) {
      continue;
    }

    
    SgExprStatement const *boundary_expr
      = isSgExprStatement(sg.graph_[*vp.first].boundary_node_);

    if(boundary_expr==NULL) continue;
    
    SgFunctionCallExp *func_call
      = isSgFunctionCallExp(boundary_expr->get_expression());

    SgExprListExp *params = func_call->get_args();

    SgExpressionPtrList &list = params->get_expressions();


    if(isSgIntVal(list.back()) || PSG_MODE) {
      // sc_core::sc_simcontext *temp = sc_get_curr_simcontext();
      SgFunctionCallExp *get_sim_context_func_call
        = SageBuilder::buildFunctionCallExp(
            isSgFunctionSymbol(
              risc::tools::SystemCDefinitions::get_sc_get_curr_simcontext()
                ->get_symbol_from_symbol_table()));

      SgAssignInitializer *init_sim_context
        = SageBuilder::buildAssignInitializer(get_sim_context_func_call);

      SgVariableDeclaration *sc_sim_context
        = SageBuilder::buildVariableDeclaration(
            generate_unique_name("sim_context"),
            SageBuilder::buildPointerType(
              risc::tools::SystemCDefinitions::get_sc_simcontext()
                ->get_declaration()->get_type()),
            init_sim_context);

      SageInterface::insertStatement(
          const_cast<SgExprStatement*>(boundary_expr),
          sc_sim_context);


      // sc_core::sc_process_b *temp2 = temp->get_curr_proc();
      SgArrowExp *arrow_get_curr_proc
        = SageBuilder::buildArrowExp(
            SageBuilder::buildVarRefExp(sc_sim_context),
            SageBuilder::buildMemberFunctionRefExp(
              isSgMemberFunctionSymbol(
                risc::tools::SystemCDefinitions::get_sc_curr_proc()
                  ->get_symbol_from_symbol_table()),
              true,
              false));

      SgFunctionCallExp *get_curr_proc
        = SageBuilder::buildFunctionCallExp(arrow_get_curr_proc);

      SgAssignInitializer *init_process_b
        = SageBuilder::buildAssignInitializer(get_curr_proc);

      SgVariableDeclaration *sc_process_b
        = SageBuilder::buildVariableDeclaration(
            generate_unique_name("process_b"),
            SageBuilder::buildPointerType(
              risc::tools::SystemCDefinitions::get_sc_process_b()
                ->get_declaration()->get_type()),
            init_process_b);

      SageInterface::insertStatement(
          const_cast<SgExprStatement*>(boundary_expr),
          sc_process_b);

      // Getting the upcomig id
      SgArrowExp *arrow_set_upcoming_ids
        = SageBuilder::buildArrowExp(
            SageBuilder::buildVarRefExp(sc_process_b),
            SageBuilder::buildMemberFunctionRefExp(
              isSgMemberFunctionSymbol(
                risc::tools::SystemCDefinitions::get_get_upcoming_segment_ids()
                  ->get_symbol_from_symbol_table()),
              true,
              false));

      SgExprListExp *expr_list_seg_ids = SageBuilder::buildExprListExp();

      SgFunctionCallExp *get_upcoming_ids
        = SageBuilder::buildFunctionCallExp(
            arrow_set_upcoming_ids, expr_list_seg_ids);

      //tag: psg cloning ZC step 5
      //when psg mode, should instrument segment id
      int id_offset;
      //if(PSG_MODE) id_offset = sg.graph_[*vp.first].id_; // I think this line is wrong
      if(PSG_MODE) id_offset = sg.graph_[*vp.first].wait_offset_;
      else id_offset = sg.graph_[*vp.first].wait_offset_;
      //end
      SgExpression *id_add_expr
        = SageBuilder::buildAddOp(
            get_upcoming_ids,
            SageBuilder::buildIntVal(id_offset));

      SgExpression *id_add_deref
        = SageBuilder::buildPointerDerefExp(id_add_expr);

      SgAssignInitializer *init_var
        = SageBuilder::buildAssignInitializer(id_add_deref);

      SgVariableDeclaration *id_var
        = SageBuilder::buildVariableDeclaration(
            generate_unique_name("wait_id"),
            SageBuilder::buildIntType(),
            init_var);

      SageInterface::insertStatement(
          const_cast<SgExprStatement*>(boundary_expr),
          id_var);

      // Exchanging integer segment id (constant) with variable
      list.pop_back();
      list.push_back(SageBuilder::buildVarRefExp(id_var));
    }
  }
}

std::string generate_unique_name(std::string var_name)
{
  std::ostringstream name_stream;
  name_stream << "__risc_temp_" << counter << "_" << var_name;
  std::string name(name_stream.str());
  counter++;

  return name;
}

} // end of namespace inst

} // end of namespace risc

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
