#include "design.h"

#include <vector>
#include <ctime>
#include <string.h>
#include <unistd.h>
#include "variable.h"

#include "cthread.h"
#include "dynamic_internal_representation.h"
#include "dynamic_path_instance_mapper.h"
#include "event.h"
#include "function.h"
#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "inoutport.h"
#include "inport.h"
#include "instance.h"
#include "instance_tree.h"
#include "interface.h"
#include "method.h"
#include "module.h"
#include "module_instance.h"
#include "outport.h"
#include "primitive_channel.h"
#include "primitive_channel_instance.h"
#include "static_path_instance_mapper.h"
#include "thread.h"

#include "../ast_traverser/traverser.h"
#include "../ast_traverser/path_checker.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/command_line_parsing.h"
#include "../tools/file_classifier.h"
#include "../tools/function_helper.h"
#include "../tools/misc.h"
#include "../tools/rose_nodes.h"
#include "../tools/systemc_definitions.h"
#include "../instrumentation/dyn_analysis_instr.h"

void yyerror(const char *s);
int yylex();


risc::Design::Design(int argc, char* argv[],
      std::vector<std::string> psg_file_names, bool is_psg_mode,
      bool PrintFrontendArgs)

{
  risc::GlobalDesign::design_static = this;

  old_argc_ = argc;
  old_argv_ = argv;

  psg_mode = is_psg_mode;
  psg_count = 0;
  // here we are parsing out the filename for the dynamic analysis
  dynamic_analysis_filename_
    = tools::dynamic_analysis_input_filename(argc, argv);

  instrumentor_output_filename_
    = tools::instrumentor_output_filename(argc, argv);

  if(PrintFrontendArgs) { // (RD, 07/06/18)
    std::cout << std::endl
              << "RISC frontend:";
    for(int i=0; i<argc; i++) {
      std::cout << " " << argv[i];
    }
    std::cout << std::endl;
  }
  project_ = frontend(argc, argv);

  if(project_ == NULL) {
    std::cout << std::endl << std::endl;
    std::cout << "Error: The AST for this design could not be created."
              << std::endl
              << "No further analysis possible."
              << std::endl;
    exit(1);
  }

  // the ordering is important
  initialize_interfaces();
  initialize_definitions();
  
 
  
  //initialize_port_mapping();
  /*
  std::cout << "\n\ndesign 74 test if only .h files are able to generate IR for modules and instances" << std::endl;
  
  for(ModuleVector::iterator test_module_iter = module_definitions_.begin(); test_module_iter!= module_definitions_.end();
  test_module_iter++)
  {
    std::cout << "*test_module_iter -> get_name() :    " << (*test_module_iter) -> get_name() << std::endl;
    std::cout << "*test_module_iter -> get_qualified_type_name() :    " << (*test_module_iter) -> get_qualified_type_name() << std::endl;
    for(ModuleInstanceVector::iterator test_sub_module_instance_iter = (*test_module_iter) -> sub_modules_.begin(); test_sub_module_instance_iter != (*test_module_iter) -> sub_modules_.end();test_sub_module_instance_iter++){
      std::cout << "*test_sub_module_instance_iter -> get_name() :    " << (*test_sub_module_instance_iter) -> get_name() << std::endl;
      std::cout << "*test_sub_module_instance_iter -> get_ast_type_name() :    " << (*test_sub_module_instance_iter) -> get_ast_type_name() << std::endl;
      std::cout << "*test_sub_module_instance_iter -> get_instance_qualified_name() :    " << (*test_sub_module_instance_iter) -> get_instance_qualified_name() << std::endl;
      
      
    }
    std::cout<<"\n\n"<<std::endl;
  }
  */
  
  if(!psg_mode){
    for(std::vector<std::string>::iterator f_iter = psg_file_names.begin();
      f_iter != psg_file_names.end();
      f_iter ++)
    {
      load_port_mapping(*f_iter);
      
    }
    
    initialize_top_modules();
    initialize_global_variables();
    initialize_global_functions();
    initialize_path_instance_mapper();
    
    initialize_port_mapping();
    
    port_mapping_cache_.initialize(path_instance_mapper_);
  }
  else{
    initialize_port_mapping();
    const std::vector<SgNode*> &func_decls = risc::tools::RoseNodes::get_func_decls();
    char cwd[1024];
    if (getcwd(cwd, sizeof(cwd)) == NULL)
      
      perror("getcwd() error");
    
    for(std::vector<SgNode*>::const_iterator
      iter  = func_decls.begin();
      iter != func_decls.end();
      iter++) 
    {
      SgFunctionDeclaration *func_decl = isSgFunctionDeclaration(*iter);
      std::string file_name = std::string(func_decl->get_file_info()->get_filename()) ;
      if( file_name.compare(std::string(cwd)+"/"+get_design_file_name()+".cpp") == 0 ){
        bool flag = false;
        for( std::vector<Thread*>::iterator iter = simulation_threads.begin();
          iter != simulation_threads.end();
          iter ++)
        {
          std::string s1 = (*iter)->get_ast_node()->get_declaration () ->get_qualified_name().getString();
          std::string s2 = func_decl->get_qualified_name().getString();
          if(s1.compare(s2) == 0 )
          {
          
            flag = true;
            break;
          }
        }
        if(flag) break;
        //else loaded_threads.push_back(new Thread(func_decl->get_definition(), NULL));
        else {
          loaded_threads.push_back(new Thread(func_decl->get_definition(), NULL));
          //psg cloning ZC step 1
          SgMemberFunctionDeclaration *mem_func_decl
          = isSgMemberFunctionDeclaration(func_decl);
          if(NULL != mem_func_decl)
            SgClassDefinition *class_def = mem_func_decl->get_class_scope();
          
        }
      }
    }
    path_instance_mapper_ = NULL;
    port_mapping_cache_.initialize(NULL);
  }

  // check if the design is analyzable for segment graph generation
  if(!is_analyzable()) {
    std::cout << std::endl << std::endl;
    std::cout << "Error: This design is not analyzable."
              << std::endl
              << "At least one module/channel has a port which is mapped to "
              << "two or more channels with different interfaces."
              << std::endl;

    exit(1);
  }
}

risc::Design::~Design()
{
  // clearing all vectors
  for(ModuleVectorIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end(); iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(PrimitiveChannelVectorIter
      iter  = primitive_channel_.begin();
      iter != primitive_channel_.end(); iter++) {
    delete *iter;
    *iter = NULL;
  }

  for(HierarchicalChannelVectorIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end(); iter++) {
    delete *iter;
    *iter = NULL;
  }
  
  for(std::vector<risc::Thread*>::iterator iter = loaded_threads.begin();
      iter != loaded_threads.end(); iter ++ ) {
    delete *iter;
    *iter = NULL;
  }

  if(path_instance_mapper_)
    delete path_instance_mapper_;
}

void risc::Design::initialize_definitions()
{
  std::vector<SgClassDefinition*> channel
    = risc::traverser::find_all_channels();

  for(std::vector<SgClassDefinition*>::iterator
      iter  = channel.begin();
      iter != channel.end();
      iter++) {

    SgClassDefinition* channel = isSgClassDefinition(*iter);
    bool is_already_inserted = false;

    for(HierarchicalChannelVectorConstIter
        channel_iter  = hierarchical_channel_.begin();
        channel_iter != hierarchical_channel_.end();
        channel_iter++) {

      if((*channel_iter)->get_ast_node() == channel) {
        is_already_inserted = true;
      }
    }

    if(is_already_inserted) {
      continue;
    }

    if(tools::is_derived_from_sc_channel(channel)) {

      hierarchical_channel_.push_back(new HierarchicalChannel(channel));
    }

    for(PrimitiveChannelVectorConstIter
        channel_iter  = primitive_channel_.begin();
        channel_iter != primitive_channel_.end();
        channel_iter++) {

      if((*channel_iter)->get_ast_node() == channel) {
        is_already_inserted = true;
      }
    }

    if(is_already_inserted) {
      continue;
    }

    if(tools::is_derived_from_sc_prim_channel(channel)) {

      primitive_channel_.push_back(new PrimitiveChannel(channel));
    }
  }

  // Search for all class definitions that are derived from sc_module and
  // are not defined in a file in the include path
  std::vector<SgClassDefinition*> modules = risc::traverser::find_all_modules();

  for(std::vector<SgClassDefinition*>::iterator
      iter  = modules.begin();
      iter != modules.end();
      iter++) {

    SgClassDefinition* temp_cd = isSgClassDefinition(*iter);
    bool is_already_inserted = false;

    for(ModuleVectorConstIter
        module_iter  = module_definitions_.begin();
        module_iter != module_definitions_.end();
        module_iter++) {

      if((*module_iter)->get_ast_node() == temp_cd) {
        is_already_inserted = true;
      }
    }

    if(is_already_inserted) {
      continue;
    }

    if(temp_cd) {
     
      Module *m = new Module(temp_cd);
      module_definitions_.push_back(m);
      for( std::vector<Thread*>::iterator iter = m->threads_.begin();
        iter != m->threads_.end();
        iter ++)
      {
        simulation_threads.push_back(*iter);
      }
    }
  }
}

risc::Module* risc::Design::find_module_definition(SgClassDefinition* cd)
{
  for(ModuleVectorConstIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {
    if((*iter)->get_ast_node() == cd) {
      return *iter;
    }
  }

  return NULL;
}

risc::HierarchicalChannel*
risc::Design::find_channel_definition(SgClassDefinition* cd)
{
  for(HierarchicalChannelVectorConstIter
      iter  = hierarchical_channel_.begin() ;
      iter != hierarchical_channel_.end();
      iter++) {

    if((*iter)->get_ast_node() == cd) {
      return *iter;
    }
  }

  return NULL;
}

risc::PrimitiveChannel*
risc::Design::find_prim_channel_definition(SgClassDefinition* cd)
{
  for(PrimitiveChannelVectorConstIter
      iter  = primitive_channel_.begin() ;
      iter != primitive_channel_.end();
      iter++) {
    if((*iter)->get_ast_node() == cd) {
      return *iter;
    }
  }

  return NULL;
}

risc::Interface*
risc::Design::find_interface(SgClassDefinition* cd)
{
  for(InterfaceVectorConstIter
      iter  = interfaces_.begin() ;
      iter != interfaces_.end();
      iter++) {
    if((*iter)->get_ast_node() == cd) {
      return *iter;
    }
  }

  return NULL;
}

void risc::Design::initialize_top_modules()
{
  //FIXME get sc_main function and search for members that are variable
  //      instances and derived from sc_module

  SgFunctionDefinition* func_def
    = tools::SystemCDefinitions::get_sc_main();
  SgBasicBlock* bb = func_def->get_body();

  for(std::vector<SgStatement*>::iterator
      iter  = bb->get_statements().begin();
      iter != bb->get_statements().end();
      iter++) {

    SgVariableDeclaration* var_decl =  isSgVariableDeclaration(*iter);

    if(var_decl) {

#if 0 /*** Tim's selective filtering *************************************/
      SgType* type = var_decl->get_decl_item(0)->get_type();

      // in this case we are addressing an int, float, ...
      if(tools::is_builtin_type(type))
        continue;

      SgArrayType* array_type = isSgArrayType(type);

      // in this case we are addressing a array of int, float, ...
      if(array_type)
        if(tools::is_builtin_type(SageInterface::getArrayElementType(type)))
          continue;

      // At this point we can know that it is a class or a corresponding array
      SgClassDefinition* class_def = NULL;

      // If it is an array we have to get the definition.
      // So, we are asking for the definition behind the array
      if(array_type) {

        SgClassType* class_type =
          isSgClassType(SageInterface::getArrayElementType(type));


        SgClassDeclaration* class_decl =
          isSgClassDeclaration(class_type->get_declaration());

        class_decl
          = isSgClassDeclaration(class_decl->get_definingDeclaration());

        class_def = class_decl->get_definition();

      } else {
        class_def = tools::class_definition_of_variable(var_decl);
      }
#endif /******************************************************************/

      /*** Use new class_definition_of_variable (bug fix 06/26/18, RD) ***/
      SgClassDefinition *class_def = tools::class_definition_of_variable(var_decl);
      if (! class_def) {
        continue; // not a class type, ignore!
      }
      /*** (bug fix 06/26/18, RD) ***/

      if(tools::is_derived_from_sc_channel(class_def)) {

        HierarchicalChannel* hc = find_channel_definition(class_def);

        top_modules_.push_back(
                         new HierarchicalChannelInstance(
                                 var_decl->get_definition(), hc));
        continue;
      }

      if(tools::is_derived_from_sc_module(class_def)) {

        Module* md = find_module_definition(class_def);

        top_modules_.push_back(
                         new ModuleInstance(var_decl->get_definition(), md));
      }
    }
  }
}

void risc::Design::initialize_global_variables()
{
  SgDeclarationStatementPtrList &global_decls
    = SageInterface::getGlobalScope(
        tools::SystemCDefinitions::get_sc_main())->get_declarations();

  int counter = 0;

  for(SgDeclarationStatementPtrList::iterator
      iter  = global_decls.begin();
      iter != global_decls.end();
      iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);

    if(var_decl) {

      // check if we added the variable already
      bool exists_already = false;
      for(VariableVectorConstIter
          iter  = global_variables_.begin();
          iter != global_variables_.end();
          iter++) {

        if(var_decl->get_decl_item(0)->search_for_symbol_from_symbol_table()
            == (*iter)->get_ast_node()->get_declaration()->
                   get_decl_item(0)->search_for_symbol_from_symbol_table()) {
          exists_already = true;
        }
      }

      if(exists_already) {
        continue;
      }

      // here we are checking the variable is module
      SgClassType *class_type
        = isSgClassType(var_decl->get_decl_item(0)->get_type());

      // in the following we are checking if the global variable is a potential
      // top module
      if(class_type) {
        SgClassDeclaration *class_decl
          = isSgClassDeclaration(class_type->get_declaration());
        class_decl
          = isSgClassDeclaration(class_decl->get_definingDeclaration());

        if(class_decl) {
          SgClassDefinition *class_def = class_decl->get_definition();

          if(class_def) {

            if(tools::is_derived_from_sc_channel(class_def)) {

              Module* md = find_module_definition(class_def);

              top_modules_.
                push_back(
                    new ModuleInstance(var_decl->get_definition(), md));
              continue;
            }

            if(tools::is_derived_from_sc_module(class_def)) {

              Module* md = find_module_definition(class_def);

              top_modules_.
                push_back(
                    new ModuleInstance(var_decl->get_definition(), md));
            }
          }
        }
      }

      global_variables_.push_back(new Variable(var_decl->get_definition()));
    }
  }
}

void risc::Design::initialize_global_functions()
{
  std::vector<SgNode*> function_definition
    = tools::RoseNodes::get_func_defs();

  int filter = 0;
  for(std::vector<SgNode*>::const_iterator
      i  = function_definition.begin();
      i != function_definition.end(); ++i) {

    SgFunctionDefinition *func = isSgFunctionDefinition(*i);

    // Theses functions are not in the global scope
    if(SageInterface::getScope(func->get_declaration()) !=
        SageInterface::getGlobalScope(func->get_declaration())) {
      continue;
    }

    // Skip all template functions
    if(isSgTemplateFunctionDefinition(func) ||
        isSgTemplateInstantiationFunctionDecl(func->get_declaration()) ||
        isSgTemplateInstantiationMemberFunctionDecl(func->get_declaration())) {
      continue;
    }

    if(func->get_file_info()->isCompilerGenerated()) {
      continue;
    }

    // Skip all functions which are part of a class
    if(func->get_declaration()->get_specialFunctionModifier().isConstructor() ||
        func->get_declaration()->get_specialFunctionModifier().isDestructor() ||
        isSgMemberFunctionDeclaration(func->get_declaration())) {
      continue;
    }

    // These functions are built in functions. So, we can skip them
    if(risc::tools::is_builtin_function(func->get_declaration())) {
      continue;
    }

    // Global functions are not allowed to start with a '_'
    // https://gcc.gnu.org/onlinedocs/libstdc++/manual/source_code_style.html
    if(SageInterface::get_name(func).at(0) == '_' ) {
        //&& SageInterface::get_name(func).at(1) == '_'){
      continue;
    }

    if(SageInterface::get_name(func) == "sc_main")
      continue;

    global_functions_.push_back(new Function(func));
  }
}

void risc::Design::print_design()
{
  std::cout << "Top Modules" << std::endl;
  for(risc::InstanceVectorConstIter
      iter  = top_modules_.begin();
      iter != top_modules_.end();
      iter++) {

    std::cout << " " << (*iter)->get_ast_type_name()
              << " "
              << (*iter)->get_name()
              << std::endl;

    HierarchicalChannelInstance *channel
      = dynamic_cast<HierarchicalChannelInstance*>(*iter);
    if(channel) {

      channel->get_definition()->print_hierarchy(2);
    } else {

      ModuleInstance *m = dynamic_cast<ModuleInstance*>(*iter);
      m->get_definition()->print_hierarchy(2);
    }
  }
  std::cout << std::endl;

  std::cout << "Modules" << std::endl;
  for(risc::ModuleVectorConstIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
    (*iter)->print_module_definition();
  }
  std::cout << std::endl;

  std::cout << "Hierarchical channels" << std::endl;
  for(risc::HierarchicalChannelVectorConstIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
    (*iter)->print_hierarchical_channel_definition();
  }
  std::cout << std::endl;

  std::cout << "Primitive channels" << std::endl;
  for(risc::PrimitiveChannelVectorConstIter
      iter  = primitive_channel_.begin();
      iter != primitive_channel_.end();
      iter++) {
    std::cout << " " << (*iter)->get_name() << std::endl;
  }
}

void risc::Design::print_all_class_definitions()
{
  for(ModuleVectorConstIter iter = module_definitions_.begin();
      iter != module_definitions_.end(); iter++) {

    std::cout << (*iter)->get_ast_type_name() << std::endl;
  }
}

void risc::Design::print_all_global_functions()
{
  std::cout << std::endl << "Global functions" << std::endl;
  for(risc::FunctionVectorConstIter
      iter  = global_functions_.begin();
      iter != global_functions_.end(); iter++) {

    std::cout << " " << (*iter)->get_name() << std::endl;
  }
}

risc::Design* risc::GlobalDesign::design_static = NULL;

risc::Design* risc::get_design()
{
  return risc::GlobalDesign::design_static;
}

SgProject* risc::get_project()
{
  return risc::GlobalDesign::design_static->project_;
}

risc::Variable*
risc::Design::find_global_variable(SgVariableDefinition* var_def)
{
  for(VariableVectorConstIter
      iter  = global_variables_.begin();
      iter != global_variables_.end();
      iter++) {

    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
  return NULL;
}

risc::InstanceTree
risc::Design::get_root()
{
  InstanceTree result;

  if(top_modules_.size() != 1) {
    std::cout << "Number of modules: " << top_modules_.size() << std::endl;
    std::cout << "Only exactly one top module is allowed" << std::endl;
    exit(1);
  }

  InstanceTree::Element top(NULL, top_modules_.at(0));
  result.path_.push_back(top);

  return result;
}

void
risc::Design::reset_instance_counter()
{
  for(ModuleVectorConstIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {
    (*iter)->instance_counter_ = 0;
  }

  for(HierarchicalChannelVectorConstIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {
    (*iter)->instance_counter_ = 0;
  }
}

bool
risc::Design::perform_static_analysis()
{
  return dynamic_analysis_filename_.empty();
}

void risc::Design::add_module_to_definition_list(SgClassDefinition *class_def)
{
  Module *m = new Module(class_def);
  module_definitions_.push_back(m);
}

void
risc::Design::add_hierarchical_channel_to_defintion_list(
    SgClassDefinition *class_def)
{
  HierarchicalChannel *hc = new HierarchicalChannel(class_def);
  hierarchical_channel_.push_back(hc);
}

std::string risc::Design::get_design_file_name()
{
  SgFile& file = project_->get_file(0);

  std::string temp = file.get_sourceFileNameWithoutPath();
  std::size_t found = temp.find_last_of("/\\");

  temp = temp.substr(found+1);

  int size = temp.length();

  if(size >= 2) {
    if(temp.at(size - 2) == '.' &&
        temp.at(size - 1) == 'c') {
      temp = temp.substr(0, size - 2);
    }
  }

  if(size >= 3) {
    if(temp.at(size - 3) == '.' &&
        temp.at(size - 2) == 'c' &&
        temp.at(size - 1) == 'c') {
      temp = temp.substr(0, size - 3);
    }
  }

  if(size >= 4) {
    if(temp.at(size - 4) == '.' &&
        temp.at(size - 3) == 'c' &&
        temp.at(size - 2) == 'p' &&
        temp.at(size - 1) == 'p') {
      temp = temp.substr(0, size - 4);
    }
  }

  return temp;
}

bool risc::Design::is_analyzable()
{
  for(ModuleVectorConstIter
      module_iter  = module_definitions_.begin();
      module_iter != module_definitions_.end();
      module_iter++) {

    if((*module_iter)->instances_.empty()) {
      continue;
    }

    for(InPortVectorConstIter
        port_iter  = (*module_iter)->inports_.begin();
        port_iter != (*module_iter)->inports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(InOutPortVectorConstIter
        port_iter  = (*module_iter)->inoutports_.begin();
        port_iter != (*module_iter)->inoutports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(OutPortVectorConstIter
        port_iter  = (*module_iter)->outports_.begin();
        port_iter != (*module_iter)->outports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }
  }

  for(HierarchicalChannelVectorConstIter
      channel_iter  = hierarchical_channel_.begin();
      channel_iter != hierarchical_channel_.end();
      channel_iter++) {

    for(InPortVectorConstIter
        port_iter  = (*channel_iter)->inports_.begin();
        port_iter != (*channel_iter)->inports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(InOutPortVectorConstIter
        port_iter  = (*channel_iter)->inoutports_.begin();
        port_iter != (*channel_iter)->inoutports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }

    for(OutPortVectorConstIter
        port_iter  = (*channel_iter)->outports_.begin();
        port_iter != (*channel_iter)->outports_.end();
        port_iter++) {

      std::set<risc::Class*> channels;
      (*port_iter)->get_mapped_channel_types(channels);
      if(channels.size() != 1) {
        return false;
      }
    }
  }

  return true;
}

void risc::Design::initialize_interfaces()
{
  SgClassDeclaration *sc_interface
    = tools::SystemCDefinitions::get_sc_interface()->get_declaration();
  sc_interface
    = isSgClassDeclaration(sc_interface->get_firstNondefiningDeclaration());


  const std::vector<SgNode*> &class_defs = tools::RoseNodes::get_class_defs();

  // getting all interfaces
  for(std::vector<SgNode*>::const_iterator
      iter  = class_defs.begin();
      iter != class_defs.end();
      iter++) {

    SgClassDefinition *class_def = isSgClassDefinition(*iter);

    if(class_def->get_inheritances().size() != 1) {
      continue;
    }

    SgClassDeclaration *base_class_decl
      = class_def->get_inheritances().at(0)->get_base_class();
    base_class_decl
      = isSgClassDeclaration(base_class_decl->get_firstNondefiningDeclaration());

    if(base_class_decl == sc_interface) {

      interfaces_.push_back(new Interface(class_def));
    }
  }
}

void risc::Design::initialize_port_mapping()
{
  // Here we setup the port binding for each module.
  for(ModuleVectorConstIter
      iter  = module_definitions_.begin();
      iter != module_definitions_.end();
      iter++) {
    (*iter)->initialize_mapping();
  }
  for(HierarchicalChannelVectorConstIter
      iter  = hierarchical_channel_.begin();
      iter != hierarchical_channel_.end();
      iter++) {
    (*iter)->initialize_mapping();
  }
}

void risc::Design::initialize_path_instance_mapper()
{
  if(perform_static_analysis()) {

    // TODO: We activate this when the design takes ParseStatus as argument
    //std::cout << "static analysis" << std::endl;
    path_instance_mapper_ = new StaticPathInstanceMapper(get_root());
  } else {

    // TODO: We activate this when the design takes ParseStatus as argument
    //std::cout << "dynamic analysis" << std::endl;
    path_instance_mapper_
      = new DynamicPathInstanceMapper(dynamic_analysis_filename_);
  }
}

void risc::Design::store_port_mapping(std::string filename)
{
  if(!PSG_MODE) {
    std::cout << "WARNING: risc::Design::store_port_mapping() is only used under PSG mode." << std::endl;
    return;
  }
  std::ofstream file;
  
  static bool flag = false;
  file.open(filename.c_str(), std::ofstream::out | std::ofstream::app);
  std::map<std::string, VertexDescriptor>  module_definition_vertex_map;
  std::map<std::string, VertexDescriptor>  module_instance_vertex_map;
  std::map<std::string, VertexDescriptor>  port_vertex_map;
  std::map<std::string, VertexDescriptor>  object_vertex_map;
  std::map<std::string, VertexDescriptor>::iterator find_iter;
  
  for(ModuleVector::iterator module_iter = module_definitions_.begin(); 
    module_iter!= module_definitions_.end();
    module_iter++)
  {
    
    for(ModuleInstanceVector::iterator 
      sub_module_instance_iter = (*module_iter) -> sub_modules_.begin(); 
      sub_module_instance_iter != (*module_iter) -> sub_modules_.end();
      sub_module_instance_iter++)
    {
      
      
      for(std::map<Port*, Object*>::iterator mapping_iter = (*sub_module_instance_iter)->mapping_.begin();
        mapping_iter!=(*sub_module_instance_iter)->mapping_.end();
        mapping_iter++)
      {
        if(flag == false) {
          file << "/*\nPort Mapping Information:\n\n";
          flag = true;
        }
        std::string mdname = (*module_iter)->get_qualified_type_name();
        std::string miname = (*sub_module_instance_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString();
        std::string pname = ((*mapping_iter).first)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString() ;
        std::string oname = ((*mapping_iter).second)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString() ;
        file << "module instance: " << miname << std::endl;
        file << "port: " << pname << std::endl;
        file << "object: " << oname << std::endl;
        file << "----------" << std::endl;
        
        VertexDescriptor md,mi,p,o;
        //find the module definition
        find_iter = module_definition_vertex_map.find(mdname);
        if(find_iter == module_definition_vertex_map.end()){
        //not already added
          SegNode mdnode;          
          mdnode.label = mdname;
          md = boost::add_vertex(mdnode,port_mapping_graph);
          module_definition_vertex_map[mdname] = md;
        }else{
        //already added
          md = module_definition_vertex_map[mdname];
        }
        
        //find the module instance
        find_iter = module_instance_vertex_map.find(miname);
        if(find_iter == module_instance_vertex_map.end()){
        //not already added
          SegNode minode;          
          minode.label = miname;
          mi = boost::add_vertex(minode,port_mapping_graph);
          module_instance_vertex_map[miname] = mi;
        }else{
        //already added
          mi = module_instance_vertex_map[miname];
        }

        find_iter = port_vertex_map.find(pname);
        if(find_iter == port_vertex_map.end()){
          SegNode pnode;
          pnode.label = pname;
          p = boost::add_vertex(pnode, port_mapping_graph);
          port_vertex_map[pname] = p;
        }else{
          p = port_vertex_map[pname];
        }
        
        find_iter = object_vertex_map.find(oname);
        if(find_iter == object_vertex_map.end()){
          SegNode onode;
          onode.label = oname;
          o = boost::add_vertex(onode, port_mapping_graph);
          port_vertex_map[oname] = o;
        }else{
          o = port_vertex_map[oname];
        }
        // now we have the vertexdescriptor for
        // module instance -- mi
        // port -- p
        // object -- o
        // then, connect them
        if(boost::edge(md, mi,port_mapping_graph).second==false)
          boost::add_edge(md,mi,port_mapping_graph);
        if(boost::edge(md,o,port_mapping_graph).second==false)
          boost::add_edge(md,o,port_mapping_graph);
        if(boost::edge(mi,p,port_mapping_graph).second==false)
          boost::add_edge(mi,p,port_mapping_graph);
        if(boost::edge(p,o,port_mapping_graph).second==false)        
          boost::add_edge(p,o,port_mapping_graph);
        if(boost::edge(o,p,port_mapping_graph).second==false)
          boost::add_edge(o,p,port_mapping_graph);
      }
        
      
    }
    
  }
  
  if(flag) file << "*/";
  file.close();
  
}
void risc::Design::draw_port_mapping(){
  std::ofstream file;
  file.open("ppm.dot", std::ofstream::out | std::ofstream::trunc);

  boost::write_graphviz(
             file,
             port_mapping_graph,
             label_writer(port_mapping_graph)
            );
  file.close();

}
void risc::Design::load_port_mapping(std::string filename)
{
  std::ifstream in(filename);
  std::string s;
  
  Port* port = NULL;
  Object* obj = NULL;
  ModuleInstance* moduleinstance = NULL;
  std::string port_name;
  std::string obj_name;
  std::string module_name;
  while (getline(in, s))
  {
    bool flag = false;
    //one mapping resolved
    if(s=="----------") {
      //std::cout << "found ----------" << std::endl;
      if(port==NULL) {
        std::cout << "Warning: " << port_name << " not found in Internal Representation" << std::endl;
        exit(1);
      }
      if(obj==NULL) {
        std::cout << "Warning: " << obj_name << " not found in Internal Representation" << std::endl;
        exit(1);
      }
      if(moduleinstance==NULL) {
        std::cout << "Warning: " << module_name << " not found in Internal Representation" << std::endl;
        exit(1);
      }
      
      moduleinstance->add_mapping(port, obj);
      //port->bounded = true;
      loaded_module_instances.push_back(moduleinstance);
      
      port = NULL;
      obj = NULL;
      moduleinstance = NULL;
    }
    
    //find port
    
    if(s.substr(0,4)=="port"){
      port_name = s.substr(6,s.size()-6);
      for(ModuleVector::iterator m_iter = module_definitions_.begin();
        m_iter != module_definitions_.end();
        m_iter ++)
      {
        if(flag == true) break;
        for(risc::InPortVector::iterator ip_iter = (*m_iter)->inports_.begin();
        ip_iter != (*m_iter)->inports_.end();
        ip_iter ++){
          if((*ip_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==port_name)
          {
            //std::cout << port_name << " found" << std::endl;
            port = *ip_iter;
            flag = true;
            break;
          }
        }
        
        if(flag == true) break;
        for(risc::InOutPortVector::iterator iop_iter = (*m_iter)->inoutports_.begin();
        iop_iter != (*m_iter)->inoutports_.end();
        iop_iter ++){
          if((*iop_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==port_name)
          {
            //std::cout << port_name << " found" << std::endl;
            port = *iop_iter;
            flag = true;
            break;
          }
        }
        
        if(flag == true) break;
        for(risc::OutPortVector::iterator op_iter = (*m_iter)->outports_.begin();
        op_iter != (*m_iter)->outports_.end();
        op_iter ++){
          if((*op_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==port_name)
          {
            //std::cout << port_name << " found" << std::endl;
            port = *op_iter;
            flag = true;
            break;
          }
        }
        
      }
    }
    
    //find channel
    
    if(s.substr(0,6)=="object"){
      obj_name = s.substr(8,s.size()-8);
      for(ModuleVector::iterator m_iter = module_definitions_.begin();
        m_iter != module_definitions_.end();
        m_iter ++)
      {
        if(flag == true) break;
        for(risc::InPortVector::iterator ip_iter = (*m_iter)->inports_.begin();
        ip_iter != (*m_iter)->inports_.end();
        ip_iter ++){
          if((*ip_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==obj_name)
          {
            //std::cout << obj_name << " found" << std::endl;
            obj = *ip_iter;
            flag = true;
            break;
          }
        }
        
        if(flag == true) break;
        for(risc::InOutPortVector::iterator iop_iter = (*m_iter)->inoutports_.begin();
        iop_iter != (*m_iter)->inoutports_.end();
        iop_iter ++){
          if((*iop_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==obj_name)
          {
            //std::cout << obj_name << " found" << std::endl;
            obj = *iop_iter;
            flag = true;
            break;
          }
        }
        
        if(flag == true) break;
        for(risc::OutPortVector::iterator op_iter = (*m_iter)->outports_.begin();
        op_iter != (*m_iter)->outports_.end();
        op_iter ++){
          if((*op_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==obj_name)
          {
            //std::cout << obj_name << " found" << std::endl;
            obj = *op_iter;
            flag = true;
            break;
          }
        }
        
        if(flag == true) break;
        for(risc::PrimitiveChannelInstanceVector::iterator pc_iter = (*m_iter)->sub_prim_channel_.begin();
        pc_iter != (*m_iter)->sub_prim_channel_.end();
        pc_iter ++){
          if((*pc_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==obj_name)
          {
            //std::cout << obj_name << " found" << std::endl;
            obj = *pc_iter;
            flag = true;
            break;
          }
        }
        
        if(flag == true) break;
        for(risc::HierarchicalChannelInstanceVector::iterator hc_iter = (*m_iter)->sub_hier_channel_.begin();
        hc_iter != (*m_iter)->sub_hier_channel_.end();
        hc_iter ++){
          if((*hc_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString()==obj_name)
          {
            //std::cout << obj_name << " found" << std::endl;
            obj = *hc_iter;
            flag = true;
            break;
          }
        }
        
      }
    }
    
    //find module instance
    if(s.substr(0,15)=="module instance"){
      module_name = s.substr(17,s.size()-17);
      for(ModuleVector::iterator m_iter = module_definitions_.begin();
        m_iter != module_definitions_.end();
        m_iter ++)
      {
         for(ModuleInstanceVector::iterator 
          sub_module_instance_iter = (*m_iter) -> sub_modules_.begin(); 
          sub_module_instance_iter != (*m_iter) -> sub_modules_.end();
          sub_module_instance_iter++)
        {
          if((*sub_module_instance_iter)->get_ast_node()->get_declaration()->get_decl_item(0)->get_qualified_name().getString() == module_name)
          {
            //std::cout << module_name << " found" << std::endl;
            moduleinstance = (*sub_module_instance_iter);
            break;
          }
        }
      }
    }
  }
}
/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
