#include "port_mapping_cache.h"

#include "design.h"
#include "hierarchical_channel.h"
#include "inoutport.h"
#include "inport.h"
#include "module.h"
#include "outport.h"
#include "port_mapping.h"

risc::PortMappingCache::PortMappingCache():
    is_initialized_(false)
{ }

void risc::PortMappingCache::initialize(
    PathInstanceMapper *path_instance_mapper)
{
  if(is_initialized_) {
    return;
  }

  if(PSG_MODE){
    is_initialized_=true;
    return;
  }
  
  Design *d = risc::get_design();

  if(!d->perform_static_analysis()) {
    return;
  }

  std::unordered_map<int,
                     std::unordered_map<SgVariableDefinition*,
                                        std::pair<risc::InstanceTree,
                                                  int> > >::iterator cache_iter;

  for(risc::ModuleVectorConstIter
       iter_module  = d->module_definitions_.begin();
       iter_module != d->module_definitions_.end();
       iter_module++) {

    // Strategy
    // First,  we get all instances of the port
    // Second, we determine the instance id of the port hosting module
    // Third,  we determine the mapped channel path
    // Forth,  we determine the instance id of the mapped channel
    // Sixth,  we add the pair of channel path
    //         and channel instance id to the table

    #define ANALYZE_PORT( PORT_TYPE, VAR_NAME )                           \
    for(risc:: PORT_TYPE                                                  \
         iter_port  = (*iter_module)-> VAR_NAME .begin();                 \
         iter_port != (*iter_module)-> VAR_NAME .end();                   \
         iter_port++) {                                                   \
      SgVariableDefinition *port_var = (*iter_port)->get_ast_node();      \
      std::list<risc::InstanceTree> ports                                 \
        = d->get_root().find_all_instances_of(port_var);                  \
      for(std::list<risc::InstanceTree>::iterator                         \
          port_path_iter  = ports.begin();                                \
          port_path_iter != ports.end();                                  \
          port_path_iter++) {                                             \
        int port_id                                                       \
          = path_instance_mapper->get_id(*port_path_iter);                \
        InstanceTree channel_path                                         \
          = risc::get_mapped_channel_instance(*port_path_iter);           \
        int channel_id                                                    \
          = path_instance_mapper->get_id(channel_path);                   \
        mapping_[port_id][port_var]                                       \
          = std::pair<risc::InstanceTree, int>(channel_path, channel_id); \
      }                                                                   \
    }
    ANALYZE_PORT( InPortVectorIter , inports_ );
    ANALYZE_PORT( InOutPortVectorIter , inoutports_ );
    ANALYZE_PORT( OutPortVectorIter , outports_ );
    #undef ANALYZE_PORT
  }

  for(risc::HierarchicalChannelVectorConstIter
       iter_hier_channel  = d->hierarchical_channel_.begin();
       iter_hier_channel != d->hierarchical_channel_.end();
       iter_hier_channel++) {

    #define ANALYZE_PORT( PORT_TYPE, VAR_NAME )                           \
    for(risc:: PORT_TYPE                                                  \
         iter_port  = (*iter_hier_channel)-> VAR_NAME .begin();           \
         iter_port != (*iter_hier_channel)-> VAR_NAME .end();             \
         iter_port++) {                                                   \
      SgVariableDefinition *port_var = (*iter_port)->get_ast_node();      \
      std::list<risc::InstanceTree> ports                                 \
        = d->get_root().find_all_instances_of(port_var);                  \
      for(std::list<risc::InstanceTree>::iterator                         \
          port_path_iter  = ports.begin();                                \
          port_path_iter != ports.end();                                  \
          port_path_iter++) {                                             \
        int port_id                                                       \
          = path_instance_mapper->get_id(*port_path_iter);                \
        InstanceTree channel_path                                         \
          = risc::get_mapped_channel_instance(*port_path_iter);           \
        int channel_id                                                    \
          = path_instance_mapper->get_id(channel_path);                   \
        mapping_[port_id][port_var]                                       \
          = std::pair<risc::InstanceTree, int>(channel_path, channel_id); \
      }                                                                   \
    }
    ANALYZE_PORT( InPortVectorIter , inports_ );
    ANALYZE_PORT( InOutPortVectorIter , inoutports_ );
    ANALYZE_PORT( OutPortVectorIter , outports_ );
    #undef ANALYZE_PORT
  }

  is_initialized_ = false;
}


void risc::PortMappingCache::print()
{
  std::unordered_map<int,
                     std::unordered_map<SgVariableDefinition*,
                                        std::pair<risc::InstanceTree,
                                                  int> > >::iterator iter;

  std::unordered_map<SgVariableDefinition*,
                     std::pair<risc::InstanceTree, int> >::iterator iter2;

  for(iter  = mapping_.begin();
      iter != mapping_.end();
      iter++) {

    std::unordered_map<SgVariableDefinition*,
                                        std::pair<risc::InstanceTree,
                                                  int> > &smap = (*iter).second;

    for(iter2  = smap.begin();
        iter2 != smap.end();
        iter2++) {

      std::cout << iter->first << " | "
                << iter2->first->get_declaration()->unparseToString() << ":"
                << iter2->first->get_declaration()->get_file_info()->get_line()
                << " | ";
      iter2->second.first.print_instance_tree();
      std::cout << ":" << iter2->second.second
                << std::endl;
    }
  }
}

std::list<std::pair<int, risc::InstanceTree>>
risc::PortMappingCache::get_mapped_module(
    int channel_id,
    SgClassDefinition *channel_definition)
{
  std::list<std::pair<int, risc::InstanceTree>> result;
  for(std::unordered_map<int,
                         std::unordered_map<SgVariableDefinition*,
                                            std::pair<risc::InstanceTree,
                                                      int> > >::iterator
      iter  = mapping_.begin();
      iter != mapping_.end();
      iter++) {

    for(std::unordered_map<SgVariableDefinition*,
                           std::pair<risc::InstanceTree,int> >::iterator
        iter2  = iter->second.begin();
        iter2 != iter->second.end();
        iter2++) {

      if(iter2->second.second == channel_id) {

        SgClassType *path_channel_type
          = isSgClassType(iter2->second.first.get_object_symbol()->get_type());

        SgClassDeclaration *path_channel_decl
          = isSgClassDeclaration(path_channel_type->get_declaration());

        path_channel_decl
          = isSgClassDeclaration(path_channel_decl->get_definingDeclaration());


        if(path_channel_decl->get_definition() == channel_definition) {

          std::list<InstanceTree> temp2
            = risc::get_design()->get_root()
                                    .find_all_instances_of(iter2->first);

          InstanceTree path
            = risc::get_design()->path_instance_mapper_
                                    ->get_instance_tree(iter->first, temp2);

          path.pop_back();

          std::pair<int, risc::InstanceTree> temp;
          temp.first = iter->first;
          temp.second = path;

          result.push_back(temp);
        }
      }
    }
  }

  return result;
}

