#ifndef DATA_CONFLICT_TABLE_H_INCLUDED_
#define DATA_CONFLICT_TABLE_H_INCLUDED_

#include<vector>
#include<utility>

#include "rose.h"

#include "segment_graph.h" // needed for VertexDescriptor
#include "conflict_table.h"
#include "mapped_variable.h" // needed for MappedVariable
#include "data_conflict_ast_attributes.h" //needed for DataConflictAstAttributes

namespace risc {

namespace sg {

/**
 * \class DataConflictTable
 * \brief This class represents the data conflict table among the individual
 *        segments. Two segments have a conflict if a least one variable is
 *        by the first segment and read by the second segment. A two
 *        dimensional array represents this array.
 */
class DataConflictTable: public ConflictTable {

public:

  /**
   * \brief    Default constructor for the DataConflictTable
   */
  DataConflictTable(
      SegmentGraph &graph,
      PathInstanceMapper *path_instance_mapper,
      bool debugging);

  /**
   * \fn       determine_conflict_table
   * \brief    This function determines conflicts among the segments
   */
  virtual void determine_conflict_table();

  /**
   * \fn       color_members_of_class
   * \brief    This function checks if the type of the given variable symbol is
   *           is a class or struct. If so, all the member variables which are
   *           not static will be set to the given color.
   */
  void
  color_members_of_class(const MappedVariable &variable,
      DataConflictAstAttributes::Status color);

  /**
   * \fn       check_members_of_class
   * \brief    This function checks if the type of the given variable symbol is
   *           is a class or struct. If so, it will check if at least on member
   *           variable which is not static has the given color.
   */
  bool check_members_of_class(const MappedVariable &mapped_variable,
           DataConflictAstAttributes::Status color,
           std::set<Conflict> &conflicts);

  /**
   * \fn       data_conflict_between_statically
   * \brief    This function checks if there is a read write conflict between
   *           two segments. The boolean flag defines if we perform a static
   *           or dynamic analysis
   */
  bool
  data_conflict_between(
      Segment &first_seg,
      Segment &second_seg,
      int row, int column,
      int inst_id_first, int inst_id_second,
      bool dynamic_analysis);

  /**
   * \fn       is_conflict_free_variable
   * \brief    This function returns true if the variable declaration of
   *           of the symbol 'symbol' has a pragma statement with
   *           "RISC variable is conflict free" before.
   */
  bool is_conflict_free_variable(SgVariableSymbol *symbol);

  /**
   * \fn    color_symbol
   * \brief Colors the symbol with color and instance id.
   *        Also considers member variables
   */
  void color_symbol(
      const MappedVariable &variable, DataConflictAstAttributes::Status color);
};


}; // end of namesapce sg

}; // end of namespace risc

#endif /* DATA_CONFLICT_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
