#ifndef GRAPH_HELPER_H_INCLUDED_
#define GRAPH_HELPER_H_INCLUDED_

#include<set>
#include<utility>
#include<list>

#include "rose.h"

#include "boost/graph/adjacency_list.hpp"
#include "boost/graph/breadth_first_search.hpp"

#include "segment_graph.h" // needed for vertex_descriptor

namespace risc {

namespace sg {

class LeafNodeVisitor: public boost::default_bfs_visitor {

public:

  LeafNodeVisitor(std::set<Graph::vertex_descriptor>& leaf_nodes):
    leaf_nodes_(leaf_nodes)
  { }

  LeafNodeVisitor(const LeafNodeVisitor &other):
    leaf_nodes_(other.leaf_nodes_)
  { }


  void discover_vertex(const Graph::vertex_descriptor vertex,
           const Graph &graph)
  {
    if(boost::out_degree(vertex, graph) == 0) {
      leaf_nodes_.insert(vertex);
    }
  }

  std::set<Graph::vertex_descriptor>& leaf_nodes_;
};


class BreakStmtVisitor: public boost::default_bfs_visitor {

public:

  BreakStmtVisitor(std::set<Graph::vertex_descriptor>& segment_with_break):
    segment_with_break_(segment_with_break)
  { }

  BreakStmtVisitor(const BreakStmtVisitor &other):
    segment_with_break_(other.segment_with_break_)
  { }


  void discover_vertex(const Graph::vertex_descriptor vertex,
           const Graph &graph)
  {
    for(std::list<NodeWithPath>::const_iterator
        iter  = graph[vertex].expressions_.begin();
        iter != graph[vertex].expressions_.end();
        iter++) {

      if(isSgBreakStmt(iter->node_)) {
        segment_with_break_.insert(vertex);
        return;
      }
    }
  }

  std::set<Graph::vertex_descriptor>& segment_with_break_;
};


class ContinueStmtVisitor: public boost::default_bfs_visitor {

public:

  ContinueStmtVisitor(std::set<Graph::vertex_descriptor>& segment_with_break):
    segment_with_break_(segment_with_break)
  { }

  ContinueStmtVisitor(const ContinueStmtVisitor &other):
    segment_with_break_(other.segment_with_break_)
  { }


  void discover_vertex(const Graph::vertex_descriptor vertex,
           const Graph &graph)
  {
    for(std::list<NodeWithPath>::const_iterator
        iter  = graph[vertex].expressions_.begin();
        iter != graph[vertex].expressions_.end();
        iter++) {

      if(isSgContinueStmt(iter->node_)) {
        segment_with_break_.insert(vertex);
        return;
      }
    }
  }

  std::set<Graph::vertex_descriptor>& segment_with_break_;
};

/**
 * \fn       get_all_reachable_segments
 * \brief    This function returns which segments are reachable from the
 *           given segment
 */
std::list<int>
get_all_reachable_segments(SegmentGraph &segment_graph, int starting_id);


/**
 * \fn       get_all_reachable_segments
 * \brief    This function returns which segments are reachable from the
 *           given segment
 */
std::vector<VertexDescriptor>
get_all_reachable_segments(
    SegmentGraph &segment_graph,
    VertexDescriptor start_vertex);

/**
 * \fn    clone_graph
 * \brief clones a subgraph
 */
std::pair<risc::sg::VertexDescriptor, risc::sg::SegmentGraph::SegmentSet>
clone_graph(risc::sg::VertexDescriptor start_vertex,
    risc::sg::SegmentGraph::SegmentSet leaving_vertices,
    risc::sg::SegmentGraph &sg,
    bool channel_segments);

void
add_pcp_to_graph(
    risc::sg::VertexDescriptor start_vertex,
    risc::sg::SegmentGraph &sg,
    SgFunctionCallExp *func_call_exp);

} // end of namespace sg

} // end of namespace risc

#endif /* GRAPH_HELPER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
