#include "function_helper.h"

#include "class_is_derived_from.h"
#include "file_classifier.h"
#include "../tools/rose_nodes.h"
#include "../tools/systemc_definitions.h"
#include "../segment_graph/function_annotation_attributes.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/hierarchical_channel_instance.h"
#include "../internal_representation/interface.h"
#include "../internal_representation/path_instance_mapper.h"
#include "../internal_representation/primitive_channel.h"
#include "../internal_representation/primitive_channel_instance.h"
#include "../internal_representation/port.h"

extern DynamicModule *ast_root;

SgFunctionDeclaration*
risc::tools::get_function_declaration_of_channel_call_statically(
    SgFunctionCallExp *func_call)
{
  // here we are getting the hosting module of the port
  SgFunctionDeclaration *func_decl
    = SageInterface::getEnclosingFunctionDeclaration(func_call);

  SgMemberFunctionDeclaration *mem_func_decl
    = isSgMemberFunctionDeclaration(func_decl);

  SgClassDefinition *class_def = mem_func_decl->get_class_scope();

  Design *design = risc::get_design();
  risc::Module *module = design->find_module_definition(class_def);

  // getting the port variable
  std::vector<SgNode*> var_refs
    = NodeQuery::querySubTree(func_call->get_function(), V_SgVarRefExp);

  SgVariableDefinition *port_def = NULL;
  for(std::vector<SgNode*>::iterator
      iter  = var_refs.begin();
      iter != var_refs.end();
      iter++) {

    SgVarRefExp *var_ref = isSgVarRefExp(*iter);

    if(tools::is_derived_from_sc_port_base(var_ref)) {

      port_def = isSgVariableDefinition(
                    var_ref->get_symbol()->get_declaration()->get_definition());
      break;
    }
  }

  Port *port = module->find_port(port_def);

  SgBinaryOp *binary = isSgBinaryOp(func_call->get_function());

  SgMemberFunctionSymbol *virtual_func_symbol
    = isSgMemberFunctionRefExp(binary->get_rhs_operand())->get_symbol();

  SgMemberFunctionDeclaration *virtual_func_decl
    = virtual_func_symbol->get_declaration();

  // getting all potentially mapped channel types
  std::set<risc::Class*> channel_types;
  port->get_mapped_channel_types(channel_types);

  // iterating over all channels which are potentially mapped
  for(std::set<risc::Class*>::iterator
      iter  = channel_types.begin();
      iter != channel_types.end();
      iter++) {

    std::vector<SgNode*> func_decls_channel
      = NodeQuery::querySubTree(
            (*iter)->get_ast_node(),
            V_SgFunctionDeclaration);

    for(std::vector<SgNode*>::iterator
        iter  = func_decls_channel.begin();
        iter != func_decls_channel.end();
        iter++) {

      SgFunctionDeclaration *channel_func_decl
        = isSgFunctionDeclaration(*iter);

      // we can only check some certain properties
      // probably, I student can write an optimized strategy for this one
      // - function name
      // - number of parameters
      // - is const function
      //
      // we cannot check for
      // - return type
      // - function parameter types

      if(channel_func_decl->get_name() != virtual_func_decl->get_name()) {
        continue;
      }

      if(isSgMemberFunctionType(channel_func_decl->get_type())->isConstFunc()
          != isSgMemberFunctionType(
                 virtual_func_decl->get_type())->isConstFunc()) {
        continue;
      }

      SgInitializedNamePtrList &parameters_func_1
        = channel_func_decl->get_parameterList()->get_args();

      SgInitializedNamePtrList &parameters_func_2
        = virtual_func_decl->get_parameterList()->get_args();

      if(parameters_func_1.size() != parameters_func_2.size()) {
        continue;
      }

      SgExprListExp *args =func_call->get_args();
      SgInitializedNamePtrList::iterator iter1 = parameters_func_1.begin();
      SgExpressionPtrList::iterator arg_iter
        = func_call->get_args()->get_expressions().begin();

      assert(args->get_expressions().size() <= parameters_func_1.size());

      for(; arg_iter != args->get_expressions().end()
          && iter1 != parameters_func_1.end(); arg_iter++, iter1++) {

        if(!SageInterface::checkTypesAreEqual(
              (*iter1)->get_type(),
              (*arg_iter)->get_type())) {
          continue;
        }
      }

      if(channel_func_decl->get_orig_return_type()
          != virtual_func_decl->get_orig_return_type()) {

        // TODO This one needs more intensive tests
        //      This should be later covered through test cases of
        //      test_channel_analysis

        //same = false;
      }

      return channel_func_decl;
    }
  }

  assert(0 && "Cannot identify mapped channel");
}

bool
risc::tools::function_call_belongs_to_a_port(SgFunctionCallExp *func_call_exp)
{
  if(func_call_exp == NULL) {
    return false;
  }

  // here we are checking if it is a sc_fifo_in or sc_fifo_out port
  SgDotExp *dot_exp = isSgDotExp(func_call_exp->get_function());
  SgArrowExp *arrow_exp = NULL;

  if(dot_exp) {

    arrow_exp = isSgArrowExp(dot_exp->get_lhs_operand());
    if(arrow_exp) {

      SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
      if(var_ref) {
        if(tools::is_sc_fifo_in(var_ref) || tools::is_sc_fifo_out(var_ref)) {
          return true;
        }
      }
    }
  }

  // The function must be in an arrow operator like some_var->foo();
  arrow_exp = isSgArrowExp(func_call_exp->get_function());
  dot_exp = isSgDotExp(func_call_exp->get_function());

  if(!arrow_exp) {

    if(dot_exp) {

      SgArrowExp *temp = isSgArrowExp(dot_exp->get_lhs_operand());
      if(temp) {
        arrow_exp = temp;

      } else {
        return false;
      }
    }

    return false;
  }

  SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_lhs_operand());
  if(var_ref) {

    SgClassType *class_type = isSgClassType(var_ref->get_type());

    if(class_type == NULL) {
      return false;
    }

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    if(tools::is_derived_from_sc_port_base(class_decl->get_definition()) ) {
      return true;
    } else {
      return false;
    }
  }

  // here we are checking for a custom designed channel like port->write
  // This is in a new scope because I don't want to mess up the function scope
  {
    SgFunctionCallExp *inner_func_call_exp
      = isSgFunctionCallExp(arrow_exp->get_lhs_operand());

    if(inner_func_call_exp) {

      SgDotExp *dot_exp = isSgDotExp(inner_func_call_exp->get_function());

      if(dot_exp) {

        SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());

        if(cast_exp) {

          SgArrowExp *arrow_exp = isSgArrowExp(cast_exp->get_operand());

          if(arrow_exp) {
            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());

            if(var_ref) {

              if(tools::is_derived_from_sc_port_base(var_ref)) {
                return true;
              }
            }
          }
        }
      }
    }
  }

  // this is the test case for fifo->read(...) and fifo_write(...)
  arrow_exp = isSgArrowExp(func_call_exp->get_function());
  if(arrow_exp) {

    SgCastExp *cast_exp = isSgCastExp(arrow_exp->get_lhs_operand());

    if(cast_exp) {
      SgFunctionCallExp *func_call_exp
        = isSgFunctionCallExp(cast_exp->get_operand());

      if(func_call_exp) {

        SgDotExp *dot_exp = isSgDotExp(func_call_exp->get_function());

        if(dot_exp) {

          SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());

          if(cast_exp) {

            SgCastExp *cast_exp2 = isSgCastExp(cast_exp->get_operand());

            if(cast_exp2) {

              SgArrowExp *arrow_exp = isSgArrowExp(cast_exp2->get_operand());

              if(arrow_exp) {

                SgVarRefExp *var_ref
                  = isSgVarRefExp(arrow_exp->get_rhs_operand());

                if(var_ref) {
                  if(tools::is_derived_from_sc_port_base(var_ref)) {
                    return true;
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  return false;
}

bool
risc::tools::function_call_is_in_user_domain(SgFunctionCallExp *func_call_exp)
{
  SgFunctionDeclaration *func_decl1 =
    func_call_exp->getAssociatedFunctionDeclaration();
  if(func_decl1) {

    func_decl1 = isSgFunctionDeclaration(func_decl1->get_definingDeclaration());

    if(func_decl1) {

      Rose::StringUtility::FileNameLibrary test
        = risc::tools::get_library(func_decl1->get_definition()->
                                       get_file_info()->get_file_id());

      if(test != "User") {

        return false;
      }
    }
  }

  return true;
}

std::vector<SgFunctionDeclaration*>
risc::tools::get_all_other_function_declarations(
    SgFunctionDeclaration *func_decl)
{
  std::vector<SgFunctionDeclaration*> func_decls;

  std::vector<SgNode*> temp = RoseNodes::get_func_decls();

  for(std::vector<SgFunctionDeclaration*>::iterator
      iter  = func_decls.begin();
      iter != func_decls.end();
      iter++) {

    SgFunctionDeclaration *func_temp = isSgFunctionDeclaration(*iter);

    if(func_decl->get_firstNondefiningDeclaration()
        == func_temp->get_firstNondefiningDeclaration()) {

      temp.push_back(func_temp);
    }
  }

  return func_decls;
}


SgFunctionDeclaration*
risc::tools::get_function_declaration_of_channel_call_dynamically(
    SgFunctionCallExp *func_call)
{
  // Getting all modules instances where the port is declared
  std::list<DynamicModule*> modules;

  SgFunctionDefinition *func_def
    = SageInterface::getEnclosingFunctionDefinition(func_call);

  SgMemberFunctionDeclaration *mem_decl
    = isSgMemberFunctionDeclaration(func_def->get_declaration());

  DynamicModule *root
    = risc::get_design()->path_instance_mapper_->get_dynamic_tree_root();

  get_all_components_of_type(
    root,
    SageInterface::get_name(mem_decl->get_class_scope()),
    modules);


  // Port name
  SgVariableDefinition *port_var_def = get_port_variable(func_call);
  std::string port_name
    = SageInterface::get_name(port_var_def->get_declaration());
  port_name.erase(0, 22);


  // Getting the channel
  // We iterated over all modules to see if a port is bound to different
  // channel types. If so, we are not able to generated the segment graph.
  DynamicObject *channel = NULL;
  DynamicModule *host_module = NULL;

  for(std::list<DynamicModule*>::iterator
      module_iter  = modules.begin();
      module_iter != modules.end();
      module_iter++) {

    DynamicModule *module = *module_iter;

    for(std::list<DynamicPort*>::iterator
        iter  = module->ports_.begin();
        iter != module->ports_.end();
        iter++) {

      if((*iter)->name_ == port_name) {

        DynamicObject *result = find_channel(root, (*iter)->address_);
        if(channel == NULL) {
          channel = result;
          host_module = get_host_module_of_member(ast_root, channel->address_);

        } else {
          if(channel->name_ != result->name_) {
            std::cout << "Error: A module port is bound to two different"
                      << " channel types."
                      << std::endl;
            std::cout << "The segment graph cannot be generated." << std::endl;
            exit(1);
          }
        }
      }
    }
  }

  assert(channel);

  SgClassDefinition *channel_class_def = NULL;

  // We are doing this to get the class type of the channel!
  for(ModuleVectorIter
      iter  = risc::get_design()->module_definitions_.begin();
      iter != risc::get_design()->module_definitions_.end();
      iter++) {

    Module *module = *iter;

    if (module->get_name() == host_module->name_) {

      SgDeclarationStatementPtrList &temp
        = module->get_ast_node()->get_members();
      for(std::vector<SgDeclarationStatement*>::iterator
          iter_decl  = temp.begin();
          iter_decl != temp.end();
          iter_decl++) {

        SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter_decl);
        if(var_decl) {

          std::string var_name = SageInterface::get_name(*iter_decl);
          var_name.erase(0, 22);

          SgClassType *type
            = isSgClassType(var_decl->get_definition()
                ->get_vardefn()->get_type());
          SgPointerType *ptr_type
            = isSgPointerType(var_decl->get_definition()
                ->get_vardefn()->get_type());

          if (ptr_type) {
            SgPointerType *ptr_type_2
              = isSgPointerType(ptr_type->get_base_type());
            if (ptr_type_2) {

              if (isSgClassType(ptr_type_2->get_base_type())) {
                type = isSgClassType(ptr_type_2->get_base_type());
              }
            }
          }

          if (type) {

            SgClassDeclaration *class_decl
              = isSgClassDeclaration(type->get_declaration());
            class_decl
              = isSgClassDeclaration(class_decl->get_definingDeclaration());

            if (channel->name_ == SageInterface::get_name(class_decl)) {
              channel_class_def = class_decl->get_definition();
            }
          }
        }
      }
    }
  }

  if(channel_class_def) {

    for(PrimitiveChannelVectorIter
        iter  = risc::get_design()->primitive_channel_.begin();
        iter != risc::get_design()->primitive_channel_.end();
        iter++) {

      if((*iter)->get_name() == channel->name_) {
        channel_class_def = (*iter)->get_ast_node();
        break;
      }
    }
  }

  // Caution this code is copied from the function
  // risc::tools::get_function_declaration_of_channel_call_statically(
  SgBinaryOp *binary = isSgBinaryOp(func_call->get_function());

  SgMemberFunctionSymbol *virtual_func_symbol
    = isSgMemberFunctionRefExp(binary->get_rhs_operand())->get_symbol();

  SgMemberFunctionDeclaration *virtual_func_decl
    = virtual_func_symbol->get_declaration();

  std::vector<SgNode*> func_decls_channel
    = NodeQuery::querySubTree(channel_class_def, V_SgFunctionDeclaration);

  for(std::vector<SgNode*>::iterator
      iter  = func_decls_channel.begin();
      iter != func_decls_channel.end();
      iter++) {

    SgFunctionDeclaration *channel_func_decl
      = isSgFunctionDeclaration(*iter);

    // we can only check some certain properties
    // probably, I student can write an optimized strategy for this one
    // - function name
    // - number of parameters
    // - is const function
    //
    // we cannot check for
    // - return type
    // - function parameter types

    if(channel_func_decl->get_name() != virtual_func_decl->get_name()) {
      continue;
    }

    if(isSgMemberFunctionType(channel_func_decl->get_type())->isConstFunc()
        != isSgMemberFunctionType(
               virtual_func_decl->get_type())->isConstFunc()) {
      continue;
    }

    SgInitializedNamePtrList &parameters_func_1
      = channel_func_decl->get_parameterList()->get_args();

    SgInitializedNamePtrList &parameters_func_2
      = virtual_func_decl->get_parameterList()->get_args();

    if(parameters_func_1.size() != parameters_func_2.size()) {
      continue;
    }

    SgExprListExp *args =func_call->get_args();
    SgInitializedNamePtrList::iterator iter1 = parameters_func_1.begin();
    SgExpressionPtrList::iterator arg_iter
      = func_call->get_args()->get_expressions().begin();

    assert(args->get_expressions().size() <= parameters_func_1.size());

    for(; arg_iter != args->get_expressions().end()
        && iter1 != parameters_func_1.end(); arg_iter++, iter1++) {

      if(!SageInterface::checkTypesAreEqual(
            (*iter1)->get_type(),
            (*arg_iter)->get_type())) {
        continue;
      }
    }

    if(channel_func_decl->get_orig_return_type()
        != virtual_func_decl->get_orig_return_type()) {

      // TODO This one needs more intensive tests
      //      This should be later covered through test cases of
      //      test_channel_analysis

      //same = false;
    }

    return channel_func_decl;
  }

  return NULL;
}


std::list<SgClassDefinition*>
risc::tools::get_channel_interfaces(SgClassDefinition *channel_def)
{
  std::list<SgClassDefinition*> interfaces;

  for(SgBaseClassPtrList::iterator
      base_iter  = channel_def->get_inheritances().begin();
      base_iter != channel_def->get_inheritances().end();
      base_iter++) {

    SgClassDeclaration *base_class_decl = (*base_iter)->get_base_class();
    base_class_decl
      = isSgClassDeclaration(base_class_decl->get_definingDeclaration());

    if(base_class_decl) {
      SgClassDefinition *base_class_def = base_class_decl->get_definition();

      if(base_class_def) {

        for(SgBaseClassPtrList::iterator
            base_base_iter  = base_class_def->get_inheritances().begin();
            base_base_iter != base_class_def->get_inheritances().end();
            base_base_iter++) {

          SgClassDeclaration *base_base_class_decl
            = (*base_base_iter)->get_base_class();

          base_base_class_decl
            = isSgClassDeclaration(
                  base_base_class_decl->get_definingDeclaration());

          if(base_base_class_decl) {
            SgClassDefinition *base_base_class_def
              = base_base_class_decl->get_definition();

            if(base_base_class_def
                == tools::SystemCDefinitions::get_sc_interface()) {

              interfaces.push_back(base_class_def);
            }
          }
        }
      }
    }
  }

  return interfaces;
}


std::list<SgClassDefinition*>
risc::tools::get_all_used_interfaces(bool only_in_user_domain)
{
  risc::Design *design = risc::get_design();

  std::list<SgClassDefinition*> interfaces;

#if 0
  #define ANALYZE_CHANNEL(TYPE, VARIABLE)                                                        \
  for(risc:: #TYPE ::iterator                                                                    \
      channel_iter  = design->#VARIABLE.begin();                                                 \
      channel_iter != design->#VARIABLE.end();                                                   \
      channel_iter++) {                                                                          \
    if(only_in_user_domain) {                                                                    \
      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() == 0) {                 \
        continue;                                                                                \
      }                                                                                          \
    }                                                                                            \
    SgClassDefinition *channel_def = (*channel_iter)->get_ast_node();                            \
    std::list<SgClassDefinition*> channel_interfaces = get_channel_interfaces(channel_def);      \
    if(only_in_user_domain) {                                                                    \
      for(std::list<SgClassDefinition*>::iterator                                                \
          interface_iter  = channel_interfaces.begin();                                          \
          interface_iter != channel_interfaces.end();                                            \
          interface_iter++) {                                                                    \
        if((*interface_iter)->get_file_info()->get_file_id() == 0) {                             \
          interfaces.push_back(*interface_iter);                                                 \
        }                                                                                        \
      }                                                                                          \
    } else {                                                                                     \
      interfaces.insert(interfaces.end(), channel_interfaces.begin(), channel_interfaces.end()); \
    }                                                                                            \
  }

  ANALYZE_CHANNEL(PrimitiveChannelVector, primitive_channel_);
  ANALYZE_CHANNEL(HierarchicalChannelVector, hierarchical_channel_);

  #undef ANALYZE_CHANNEL


#endif

#if 1
  for(risc::PrimitiveChannelVector::iterator
      channel_iter  = design->primitive_channel_.begin();
      channel_iter != design->primitive_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() == 0) {
        continue;
      }
    }

    SgClassDefinition *channel_def = (*channel_iter)->get_ast_node();
    std::list<SgClassDefinition*> channel_interfaces
      = get_channel_interfaces(channel_def);

    if(only_in_user_domain) {
      for(std::list<SgClassDefinition*>::iterator
          interface_iter  = channel_interfaces.begin();
          interface_iter != channel_interfaces.end();
          interface_iter++) {

        if((*interface_iter)->get_file_info()->get_file_id() == 0) {

          interfaces.push_back(*interface_iter);
        }
      }
    } else {

      interfaces.insert(
          interfaces.end(),
          channel_interfaces.begin(),
          channel_interfaces.end());
    }
  }

  for(risc::HierarchicalChannelVector::iterator
      channel_iter  = design->hierarchical_channel_.begin();
      channel_iter != design->hierarchical_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() != 0) {
        continue;
      }
    }

    SgClassDefinition *channel_def = (*channel_iter)->get_ast_node();
    std::list<SgClassDefinition*> channel_interfaces
      = get_channel_interfaces(channel_def);

    if(only_in_user_domain) {

      for(std::list<SgClassDefinition*>::iterator
          interface_iter  = channel_interfaces.begin();
          interface_iter != channel_interfaces.end();
          interface_iter++) {

        if((*interface_iter)->get_file_info()->get_file_id() == 0) {

          interfaces.push_back(*interface_iter);
        }
      }
    } else {

      interfaces.insert(
          interfaces.end(),
          channel_interfaces.begin(),
          channel_interfaces.end());
    }
  }

#endif
  return interfaces;
}

std::list<SgFunctionDeclaration*>
risc::tools::get_interface_functions(
    SgClassDefinition *channel_def,
    bool only_annotated)
{
  std::list<SgFunctionDeclaration*> channel_interface_func_decls;

  std::list<SgClassDefinition*> interfaces
    = get_channel_interfaces(channel_def);

  std::vector<SgNode*> channel_func_decls
    = NodeQuery::querySubTree(channel_def, V_SgFunctionDeclaration);

  for(std::list<SgClassDefinition*>::iterator
      interface_iter  = interfaces.begin();
      interface_iter != interfaces.end();
      interface_iter++) {

    std::vector<SgNode*> interface_func_decls
      = NodeQuery::querySubTree(*interface_iter, V_SgFunctionDeclaration);

    for(std::vector<SgNode*>::iterator
        interface_func_decl_iter  = interface_func_decls.begin();
        interface_func_decl_iter != interface_func_decls.end();
        interface_func_decl_iter++) {

      SgFunctionDeclaration *interface_func_decl
        = isSgFunctionDeclaration(*interface_func_decl_iter);

      for(std::vector<SgNode*>::iterator
          channel_func_decl_iter  = channel_func_decls.begin();
          channel_func_decl_iter != channel_func_decls.end();
          channel_func_decl_iter++) {

        SgFunctionDeclaration *channel_func_decl
          = isSgFunctionDeclaration(*channel_func_decl_iter);

        if(SageInterface::get_name(interface_func_decl)
            == SageInterface::get_name(channel_func_decl)) {

          if(only_annotated) {

            if(is_annotated_function(channel_func_decl)) {
              channel_interface_func_decls.push_back(channel_func_decl);
            }
          } else {
            channel_interface_func_decls.push_back(channel_func_decl);
          }
        }
      }
    }
  }

  return channel_interface_func_decls;
}

std::list<SgFunctionCallExp*>
risc::tools::get_all_port_calls_to_annotated_functions(
    risc::sg::SegmentGraph &graph)
{
  std::list<SgFunctionCallExp*> func_decls;

  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(graph.graph_);
      vp.first != vp.second;
      ++vp.first) {

    if(graph.graph_[*vp.first].inlined_segment_) {
      SgNode *boundary
        = const_cast<SgNode*>(graph.graph_[*vp.first].boundary_node_);
      std::cout << " boundary" << boundary  << std::endl;
      SgFunctionCallExp *func_call_exp
        = isSgFunctionCallExp(boundary);

      if(func_call_exp) {
        std::cout << " func_call_exp" << func_call_exp->unparseToString() << std::endl;
        func_decls.push_back(func_call_exp);
      }
    }
  }

  return func_decls;
}


std::list<SgFunctionDeclaration*>
risc::tools::get_all_implemented_interface_functions(
    bool only_in_user_domain,
    bool only_annotated)
{
  std::list<SgFunctionDeclaration*> channel_interface_functions;

  for(risc::PrimitiveChannelVector::iterator
      channel_iter  = risc::get_design()->primitive_channel_.begin();
      channel_iter != risc::get_design()->primitive_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() != 0) {
        continue;
      }
    }
    std::list<SgFunctionDeclaration*> channel_interfaces
      = get_interface_functions(
            (*channel_iter)->get_ast_node(),
            only_annotated);

    channel_interface_functions.insert(
        channel_interface_functions.end(),
        channel_interfaces.begin(),
        channel_interfaces.end());
  }

  for(risc::HierarchicalChannelVector::iterator
      channel_iter  = risc::get_design()->hierarchical_channel_.begin();
      channel_iter != risc::get_design()->hierarchical_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() != 0) {
        continue;
      }
    }

    std::list<SgFunctionDeclaration*> channel_interfaces
      = get_interface_functions(
            (*channel_iter)->get_ast_node(),
            only_annotated);

    channel_interface_functions.insert(
        channel_interface_functions.end(),
        channel_interfaces.begin(),
        channel_interfaces.end());
  }

  return channel_interface_functions;
}

bool risc::tools::is_annotated_function(SgFunctionDeclaration *func_decl)
{
  func_decl
    = isSgFunctionDeclaration(func_decl->get_firstNondefiningDeclaration());

  sg::FunctionAnnotationAttributes *function_annotation_attribute =
    dynamic_cast<sg::FunctionAnnotationAttributes*>(
        func_decl->getAttribute("function_annotation"));

  if(function_annotation_attribute) {
    return true;
  } else {
    return false;
  }
}

bool risc::tools::has_reference_parameter(SgFunctionDeclaration *func_decl)
{
  for(SgInitializedNamePtrList::iterator
      iter  = func_decl->get_args().begin();
      iter != func_decl->get_args().end();
      iter++) {

    if(isSgReferenceType((*iter)->get_type())) {
      return true;
    }
  }
  return false;
}

SgFunctionDeclaration*
risc::tools::get_first_non_defining_function_declaration(
    SgFunctionCallExp *func_call_exp)
{
  SgFunctionDeclaration *func_decl = NULL;
  if(!PSG_MODE && tools::function_call_belongs_to_a_port(func_call_exp)) {
    if(risc::get_design()->perform_static_analysis()) {

      func_decl = tools::get_function_declaration_of_channel_call_statically(
                             func_call_exp);
    } else {

      func_decl = tools::get_function_declaration_of_channel_call_dynamically(
                             func_call_exp);
    }

    func_decl = isSgFunctionDeclaration(
                    func_decl->get_firstNondefiningDeclaration());

  } else {

    func_decl = func_call_exp->getAssociatedFunctionDeclaration();
    if(func_decl) {

      func_decl = isSgFunctionDeclaration(
                      func_decl->get_firstNondefiningDeclaration());

      if(is_overloaded_operator(func_decl)) {
        return NULL;
      }
    }
  }
  return func_decl;
}

bool risc::tools::is_overloaded_operator(SgFunctionDeclaration *func_decl)
{
  if(func_decl) {

    if(isSgMemberFunctionDeclaration(func_decl)) {

      if(SageInterface::get_name(func_decl) == "operator="

          // arithmetic
          || SageInterface::get_name(func_decl) == "operator+"
          || SageInterface::get_name(func_decl) == "operator-"
          || SageInterface::get_name(func_decl) == "operator*"
          || SageInterface::get_name(func_decl) == "operator/"
          || SageInterface::get_name(func_decl) == "operator%"

          || SageInterface::get_name(func_decl) == "operator+="
          || SageInterface::get_name(func_decl) == "operator-="
          || SageInterface::get_name(func_decl) == "operator*="
          || SageInterface::get_name(func_decl) == "operator/="
          || SageInterface::get_name(func_decl) == "operator%="
          || SageInterface::get_name(func_decl) == "operator&="
          || SageInterface::get_name(func_decl) == "operator|="
          || SageInterface::get_name(func_decl) == "operator~="

          || SageInterface::get_name(func_decl) == "operator&"
          || SageInterface::get_name(func_decl) == "operator|"
          || SageInterface::get_name(func_decl) == "operator^"

          || SageInterface::get_name(func_decl) == "operator&&"
          || SageInterface::get_name(func_decl) == "operator||"

          // compare
          || SageInterface::get_name(func_decl) == "operator!="
          || SageInterface::get_name(func_decl) == "operator=="
          || SageInterface::get_name(func_decl) == "operator<"
          || SageInterface::get_name(func_decl) == "operator<="
          || SageInterface::get_name(func_decl) == "operator>"
          || SageInterface::get_name(func_decl) == "operator>="

          // unary
          || SageInterface::get_name(func_decl) == "operator+"
          || SageInterface::get_name(func_decl) == "operator-"
          || SageInterface::get_name(func_decl) == "operator!"
          || SageInterface::get_name(func_decl) == "operator~"
          || SageInterface::get_name(func_decl) == "operator++"
          || SageInterface::get_name(func_decl) == "operator--") {

        return true;
      }
    }
  }
  return false;
}

SgVariableDefinition*
risc::tools::get_port_variable(SgFunctionCallExp *func_call_exp)
{
  if(func_call_exp == NULL) {
    return NULL;
  }

  // here we are checking if it is a sc_fifo_in or sc_fifo_out port
  SgDotExp *dot_exp = isSgDotExp(func_call_exp->get_function());
  SgArrowExp *arrow_exp = NULL;

  if(dot_exp) {

    arrow_exp = isSgArrowExp(dot_exp->get_lhs_operand());
    if(arrow_exp) {

      SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
      if(var_ref) {
        if(tools::is_sc_fifo_in(var_ref) || tools::is_sc_fifo_out(var_ref)) {
          return isSgVariableDefinition(
                    var_ref->get_symbol()->get_declaration()->get_definition());
        }
      }
    }
  }

  // The function must be in an arrow operator like some_var->foo();
  arrow_exp = isSgArrowExp(func_call_exp->get_function());
  dot_exp = isSgDotExp(func_call_exp->get_function());

  if(!arrow_exp) {

    if(dot_exp) {

      SgArrowExp *temp = isSgArrowExp(dot_exp->get_lhs_operand());
      if(temp) {
        arrow_exp = temp;

      } else {
        return NULL;
      }
    }

    return NULL;
  }

  SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_lhs_operand());
  if(var_ref) {

    SgClassType *class_type = isSgClassType(var_ref->get_type());

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    if(tools::is_derived_from_sc_port_base(class_decl->get_definition()) ) {

      return isSgVariableDefinition(
                var_ref->get_symbol()->get_declaration()->get_definition());

    } else {
      return NULL;
    }
  }

  // here we are checking for a custom designed channel like port->write
  // This is in a new scope because I don't want to mess up the function scope
  {
    SgFunctionCallExp *inner_func_call_exp
      = isSgFunctionCallExp(arrow_exp->get_lhs_operand());

    if(inner_func_call_exp) {

      SgDotExp *dot_exp = isSgDotExp(inner_func_call_exp->get_function());

      if(dot_exp) {

        SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());

        if(cast_exp) {

          SgArrowExp *arrow_exp = isSgArrowExp(cast_exp->get_operand());

          if(arrow_exp) {
            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());

            if(var_ref) {

              if(tools::is_derived_from_sc_port_base(var_ref)) {
                return isSgVariableDefinition(
                          var_ref->get_symbol()->get_declaration()
                              ->get_definition());
              }
            }
          }
        }
      }
    }
  }

  // this is the test case for fifo->read(...) and fifo_write(...)
  arrow_exp = isSgArrowExp(func_call_exp->get_function());
  if(arrow_exp) {

    SgCastExp *cast_exp = isSgCastExp(arrow_exp->get_lhs_operand());

    if(cast_exp) {
      SgFunctionCallExp *func_call_exp
        = isSgFunctionCallExp(cast_exp->get_operand());

      if(func_call_exp) {

        SgDotExp *dot_exp = isSgDotExp(func_call_exp->get_function());

        if(dot_exp) {

          SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());

          if(cast_exp) {

            SgCastExp *cast_exp2 = isSgCastExp(cast_exp->get_operand());

            if(cast_exp2) {

              SgArrowExp *arrow_exp = isSgArrowExp(cast_exp2->get_operand());

              if(arrow_exp) {

                SgVarRefExp *var_ref
                  = isSgVarRefExp(arrow_exp->get_rhs_operand());

                if(var_ref) {
                  if(tools::is_derived_from_sc_port_base(var_ref)) {
                    return isSgVariableDefinition(
                              var_ref->get_symbol()->get_declaration()
                                  ->get_definition());
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  return NULL;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
